#include "main.h"

GtkWidget *group_list, *node_tree;

static void
int_handle_child(int num)
{
   pid_t               pid;
   int                 status;
   
   num = 0;
   while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
     {
	Plugin *p;
	
	p = flim_plugin_find_pid(pid);
	if (p) flim_plugin_mark_dead(p);
     }
   return;
}

static void
int_handle_exit(int num)
{
   num = 0;
   exit(0);
}

int
main (int argc, char *argv[])
{
   GtkWidget *base;
   struct sigaction    sa;
   
   /* setup signal handler to handle child process returns to avoid zombies */
   sa.sa_handler = int_handle_child; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGCHLD, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGINT, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGQUIT, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGTERM, &sa, (struct sigaction *)0);
   
#ifdef ENABLE_NLS
   bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
   textdomain (PACKAGE);
#endif
   
   gtk_set_locale ();
   gtk_init (&argc, &argv);
   
   add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
   add_pixmap_directory (PACKAGE_SOURCE_DIR "/pixmaps");
   
   base = create_base ();
   gtk_widget_show (base);
   
   flim_config_load();
   gtk_widget_realize(node_tree);
   flim_nexxus_build_tree(node_tree);
   flim_group_build_list(group_list);
   
   gtk_main ();
   flim_config_save();
   return 0;
}

