/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Notes: This is now a stand-alone run-time linkable module for VACM
 *
 */

#include <stdio.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <netdb.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <vacmclient_api.h>

extern int	vacm_errno;
void				*handle;
struct termios *orig_term;

void split_response(char *response, char *module, unsigned int *job_id, char *msg);
static void usage(void);

int main(argc, argv)
int		argc;
char	**argv;
{
	struct termios	term;
	char						line[255];
	char						*response;
  int   					resp_len;
 	char						msg[255];
	int							fd;
  char  					*tmp;
  int   					len;
	int							rc;
	int							con_handle = -1, port = -1;
	struct in_addr	addr;
	char						otp[255];
	char						*nexxus=NULL,*user=NULL,*password=NULL;
	char						*node_id=NULL,*mode=NULL;
	struct sockaddr_in	sockaddr;
	char						buff[255];

  orig_term = (struct termios *)malloc(sizeof(struct termios));
  tcgetattr(fileno(stdin), orig_term);

	while((rc = getopt(argc, argv, "n:u:p:i:m:"))!=EOF)
		{
		if ((char)rc == 'n')
			nexxus = strdup(optarg);
		else if ((char)rc == 'u')
			user = strdup(optarg);
		else if ((char)rc == 'p')
			password = strdup(optarg);
		else if ((char)rc == 'i')
			node_id = strdup(optarg);
		else if ((char)rc == 'm')
			{
			mode = strdup(optarg);
			if ((strcasecmp(mode,"RO"))&&(strcasecmp(mode,"RW")))
				{
				printf("Mode must be either 'RO' or 'RW'\n");
				sleep(3);
				exit(1);
				}
			}
		else
			usage();
		}
	if ((!nexxus) || (!user) || (!password) || (!node_id))
		usage();

  if (api_nexxus_connect(nexxus, user, password, &handle)<0)
    {
    printf("Unable to connect to '%s' (API errcode %d)\n",
            argv[1],
            vacm_errno);
    exit(1);
    }
  if (api_nexxus_recv_ipc(handle, &tmp, &len)<0)
    {
    api_nexxus_perror("sercon_terminal");
    exit(-1);
    }
  if (strncmp(tmp, "NEXXUS_READY",12))
    {
    printf("sercon_terminal: Unknown nexxus init response '%s'\n",tmp);
    exit(-1);
    }
  free(tmp);


	sprintf(msg,"SERCON:CONNECT:%s:%s",node_id,mode);
	api_nexxus_send_ipc(handle, &msg[0], (strlen(msg)+1));
	while(1)
		{
		char				module[255];
		__uint32_t	job_id;

		if ((rc=api_nexxus_wait_for_data(handle, &response, &resp_len, 10))<0)
			{
			printf("Nexxus timed out on reply\n");
			exit(-1);
			}
		split_response(response, module, &job_id, line);
		if (!strcmp(line, "JOB_STARTED"))
			continue;
		if (!strcmp(line, "JOB_COMPLETED"))
			break;
		if (!strncmp(line, "JOB_ERROR",9))
			{
			printf("SERCON returned 'JOB_ERROR' (%s)\n",line);
			exit(2);
			}
		if (!strncmp(line, "CONNECT", 7))
			{
			strtok(line, ":");
			con_handle = atoi(strtok(NULL, ":"));
			addr.s_addr = inet_addr(strtok(NULL, ":"));
			port = atoi(strtok(NULL, ":"));
			strcpy(otp,strtok(NULL, ":"));
			}
		free(response);
		}
printf("con = %d, addr = %s, port = %d, otp = %s\n",
	con_handle,inet_ntoa(addr), port, otp);

	if ((fd = socket(AF_INET, SOCK_STREAM, 0))<0)
		{
		perror("socket");
		exit(1);
		}
	memset(&sockaddr, 0, sizeof(sockaddr));
	sockaddr.sin_family = AF_INET;
	sockaddr.sin_addr = addr;
	sockaddr.sin_port = htons(port);
	if (connect(fd,
							(struct sockaddr *) &sockaddr,
							sizeof(struct sockaddr)) <0)
		{
		perror("connect");
		exit(1);
		}
	rc = read(fd, &buff, sizeof(buff));
	buff[rc] = 0;
	printf("buff is <%s>\n",buff);

	write(fd, &otp[0], (strlen(otp)+1));
	write(fd, "\n", 1);

	system("clear");
  tcgetattr(fileno(stdin), &term);
  cfmakeraw(&term);
	term.c_iflag &=~ISIG;
  tcsetattr(fileno(stdin), TCSANOW, &term);
	signal(SIGINT, SIG_DFL);

	while(1)
		{
		struct timeval	to;
		fd_set					read_fds;
		int							max;

		FD_ZERO(&read_fds);
		max = 0;
		to.tv_sec = 1;
		to.tv_usec = 0;


		FD_SET(fd, &read_fds);
		FD_SET(fileno(stdin), &read_fds);
		if (fd > fileno(stdin))
			max = fd;
		else
			max = fileno(stdin);
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to))<0)
			{
			perror("select");
			exit(-1);
			}
		if (!rc)
			continue;
		if (FD_ISSET(fileno(stdin), &read_fds))
			{
			char	byte;

			if (read(fileno(stdin), &byte, 1)<0)
				{
				perror("user input");
				exit(-1);
				}
			write(fd, &byte, 1);
			}
		else if (FD_ISSET(fd, &read_fds))
			{
			char	byte;

			if ((rc = read(fd, &byte, 1))<0)
				{
				perror("read");
				exit(1);
				}
      else if (rc == 0)
        {
        tcsetattr(fileno(stdin), TCSANOW, orig_term);
        printf("\nConnection Reset by Peer\n");
        exit(-1);
        }
			printf("%c",byte);
			fflush(stdout);
			}
		else
			{
			printf("select() triggered by unknown entity!\n");
			exit(-1);
			}
		}
}

void split_response(char *response,
										char *module,
										unsigned int *job_id,
										char *msg)
{
	char	*bword;
	char	tmp[255];

	strcpy(tmp, response);
	if (!(bword = strtok(tmp,":")))
		{
		printf("Malformatted response pkt (1)(%s)\n",response);
		exit(-1);
		}
	strcpy(module,bword);
	if (!(bword = strtok(NULL,":")))
		{
		printf("Malformatted response pkt (2)(%s) (0x%.2x)\n",response,response[0]);
		exit(-1);
		}
	*job_id = atoi(bword);
	if (!(bword = strtok(NULL,":")))
		{
		printf("Malformatted response pkt (3)(%s)\n",response);
		exit(-1);
		}

	strcpy(msg,&response[(bword-tmp)]);
//	printf("Module <%s>, Job id <%d>, msg <%s>\n",module,*job_id,msg);
}

static void usage()
{
	printf("Usage: program <-n nexxus> <-u username> <-p password> <-i node_id> <-m mode>\n");
	printf("       Where <mode> is either 'RO' or 'RW'\n");
	exit(1);
}
