/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Apr 27, 2000: Added Ari Jort's modification to make variables case 
 * insensitive
 */

#include <config.h>

#include <stdio.h>
#include <ctype.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <nexxus.h>
#include <module_harness.h>
#include <prototypes.h>

static void config_error(int line, char *msg);

static int variable_set(NODE	*node,
												char						*module,
												char						*variable,
												char						*value,
												int							notify);
static void variable_destroy(NODE *node, VARIABLE_LIST *var);
NODE	*nodelist_entry_create(char *node_id);	// nexxus.c

extern int 						nexxus_debug;
extern NODEGROUP			*nodegroup_head;
extern NODEGROUP			*nodegroup_tail;
extern NEXXUS_USER		*user_head;
extern NEXXUS_USER		*user_tail;
extern bool_t					nexxus_continuous_config_backups;
extern int						errno;

static void config_error(int line, char *msg);
static int	variable_get(NODE *node,
								 char						*module,
								 char						*variable,
								 char						*buffer,
								 int						maxbuffer);

static FILE	*emergency_mode = NULL;

/*
 * Open's and reads the configuration file
 */
int configuration_load()
{
	NODE	*node = NULL;
	NODEGROUP				*group = NULL;
	NEXXUS_USER			*user = NULL;
	FILE						*Tfilep_ptr;
	int							line_count = 0;
	int							flag_nodedefine = 0;
	char						current_node[255];
	char						line_buffer[255];
	char						saved_line_buffer[255];
	char						*bword;
	char						*last = (char *) 0x00;
	struct stat			stat_buffer;
	int							rc;
	int							implicit_group_add = 0;

	/* Verify the file is owned by root and can only be read/written by root */
	rc =stat(PKGLIBDIR "/vacm_configuration", &stat_buffer);
	if (rc <0)
		{
		if (errno == ENOENT)
			{
			printf("\n");
			printf("[Nexxus] Configuration file path %s/vacm_configuration does not exist\n",
							PKGLIBDIR);
			printf("[Nexxus] If you legitimately do not have a configuration file at this time\n");
			printf("[Nexxus] create a blank one by typing 'touch %s/vacm_configuration'\n",PKGLIBDIR);
			printf("\n");
			return(-1);
			}
		nexxus_log_printf(NULL,"Unable to stat configuration file (%m)\n");
		return(-1);
		}
	if (stat_buffer.st_uid != 0)
		{
		nexxus_log_printf(NULL,
													"SECURITY ALERT! Configuration file not owned by root (uid 0)\n");
		return(-1);
		}
	if ((stat_buffer.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO))!=(S_IRUSR | S_IWUSR))
		{
		nexxus_log_printf(NULL,
													"SECURITY ALERT! Configuration file not mode 600!\n");
		return(-1);
		}
	if (!(Tfilep_ptr = fopen(PKGLIBDIR "/vacm_configuration", "r")))
		{
		nexxus_log_printf(NULL,"Unable to open configuration file (%m)\n");
		return(-1);
		}

	while(fgets(line_buffer, sizeof(line_buffer), Tfilep_ptr))
		{
		line_count++;
		line_buffer[strlen(line_buffer)-1] = 0;
		if ((line_buffer[0] == (char) '#') || (line_buffer[0] == (char ) NULL))
			continue;
		strcpy(saved_line_buffer, line_buffer);
		/* Process the line */
		if (!(bword = strtok_r(line_buffer, ":",&last)))
			{
			config_error(line_count,"Malformatted line");
			return(-1);
			}
		if (!strcasecmp(bword, "GROUP"))
			{
			if ((group) && (!user))
				{
				config_error(line_count, "Cannot process nested groups");
				return(-1);
				}
			else if ((!group) && (user))
				{
				/* group definition for a user */
				char	*grp_names;
	
				while((grp_names = strtok_r(NULL,":",&last)))
					user_group_member_add(user, grp_names);
				}
			else
				{
				char	*grp_name;
				/* new group definition */
				
				if (!(grp_name = strtok_r(NULL,":",&last)))
					{
					config_error(line_count,"GROUP specified with no group name");
					return(-1);
					}
				if (!(group=nodegroup_lookup_by_name(grp_name)))
					group = nodegroup_create(grp_name);
				}
			}
		else if (!strcasecmp(bword, "GROUP_END"))
			{
			if (!group)
				{
					config_error(line_count,"GROUP_END specified when not within group");
					return(-1);
				}
			group = NULL;
			}
		else if (!strcasecmp(bword, "NODE"))
			{
			if (flag_nodedefine)
				{
				config_error(line_count, "Nested node definitions not supported");
				continue;
				}
			if (!(bword = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "NODE keyword requres a node name");
				continue;
				}
			strcpy(current_node, bword);
			flag_nodedefine = 1;
			if (nexxus_debug)
				nexxus_log_printf(NULL, "Processing definition for node '%s'\n",current_node);
			node = nodelist_entry_create(current_node);
	
			if (!group)
				{
				nexxus_log_printf(NULL,
															"Node %s is not in a predefined group\n",
															current_node);
				nexxus_log_printf(NULL,"Adding to 'DEFAULT' group\n");

				if (!(group=nodegroup_lookup_by_name("default")))
					group = nodegroup_create("default");
				implicit_group_add = 1;
				}
			nodegroup_add_node(group, node);
			}
		else if (!strcasecmp(bword, "END_NODE"))
			{
			if (!flag_nodedefine)
				{
				config_error(line_count, "END_NODE appears without matching NODE");
				continue;
				}
			flag_nodedefine = 0;
			if (implicit_group_add)
				group = NULL;
			if (nexxus_debug)
				nexxus_log_printf(NULL,"Processing of node '%s' complete\n",current_node);
			}
		else if (!strcasecmp(bword, "USER"))
			{
			char	*user_name;

			if (user)
				{
				config_error(line_count, "Nested user definitions not supported");
				continue;
				}
			if (!(user_name = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "User definition contains empty name");
				continue;
				}
			if (user_exists(user_name))
				{
				config_error(line_count, "Duplicate username definition");
				continue;
				}
			if (!(user = user_create(user_name)))
				{
				config_error(line_count, "Insufficient memory to define user");
				continue;
				}
			}
		else if (!strcasecmp(bword, "END_USER"))
			{
			if (!user)
				{
				config_error(line_count, "END_USER appears without matching USER");
				continue;
				}
			if (nexxus_debug)
				nexxus_log_printf(NULL,
															"Processing of user '%s' complete\n",
															user->user_name);
			user = NULL;
			}
		else if (!strcasecmp(bword, "PASSWORD"))
			{
			char	*password;

			if (!user)
				{
				config_error(line_count, "Skipping 'PASSWORD' field");
				continue;
				}
			if (!(password = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "NULL 'PASSWORD' field specified");
				continue;
				}
			strcpy(user->password, password);
			}
		else if (!strncasecmp(bword, "DEFAULT_MODULE_POLICY",21))
			{
			char	*policy;

			if (!user)
				{
				config_error(line_count, "Skipping 'DEFAULT_MODULE_POLICY' field");
				continue;
				}
			if (!(policy = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "NULL 'DEFAULT_MODULE_POLICY' field specified");
				user->default_module_acl_policy = POLICY_DENY; // be safe
				continue;
				}
			if (!strcasecmp(policy,"ALLOW"))
				user->default_module_acl_policy = POLICY_ALLOW;
			else
				user->default_module_acl_policy = POLICY_DENY; // be safe
			}
		else if (!strcasecmp(bword, "DEFAULT_ADDRESS_POLICY"))
			{
			char	*policy;

			if (!user)
				{
				config_error(line_count, "Skipping 'DEFAULT_ADDRESS_POLICY' field");
				continue;
				}
			if (!(policy = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "NULL 'DEFAULT_ADDRESS_POLICY' field specified");
				user->default_address_acl_policy = POLICY_DENY; // be safe
				continue;
				}
			if (!strcasecmp(policy,"ALLOW"))
				user->default_address_acl_policy = POLICY_ALLOW;
			else
				user->default_address_acl_policy = POLICY_DENY; // be safe
			}
		else if (!strcasecmp(bword, "MOD_ACL"))
			{
			USER_MODULE_ACL	*acl;
			int							policy;
			char						module[16];
			char						cmd[255];
			char						*p;

			if (!user)
				{
				config_error(line_count, "Skipping 'MOD_ACL' field");
				continue;
				}

			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			if (!strcasecmp(p,"ALLOW"))
				policy = POLICY_ALLOW;
			else
				policy = POLICY_DENY;

			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			strncpy(module,p,sizeof(module));

			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			strncpy(cmd,p,sizeof(cmd));

			/* 
			 * Add this ACL entry into our list
			 */
			/* First check for existance of the MODULE_ACL */
			if (!(acl = user_module_acl_lookup(user, module)))
				acl = user_module_acl_create(user, module);
			/* Now add the policy entry to the list */
			user_module_acl_rule_create(acl, policy, cmd);
			}
		else if (!strcasecmp(bword, "ADDR_ACL"))
			{
			USER_ADDRESS_RULE	*acl;
			struct in_addr		addr;
			int								sigbits;
			char							*p;
			int								policy;

			if (!user)
				{
				config_error(line_count, "Skipping 'ADDR_ACL' field");
				continue;
				}
			memset(&addr, 0, sizeof(addr));
			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			if (!strcasecmp(p,"ALLOW"))
				policy = POLICY_ALLOW;
			else
				policy = POLICY_DENY;
			
			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			addr.s_addr = inet_addr(p);
			if (!(p = strtok_r(NULL,":",&last)))
				{
				config_error(line_count, "Syntax Error");
				continue;
				}
			sigbits = atoi(p);
			if(!(acl = user_address_acl_rule_create(user,
																				 &addr,
																				 sigbits,
																				 policy)))
				{
				config_error(line_count, "Unable to create address ACL entry");
				continue;
				} 
			}
		else if (!strcasecmp(bword, "VAR"))
			{
			char	*p, *q;
			char	varname[255];
			char	var_module[255];
			char	value[255];

			if (!(p = strtok_r(NULL, ":",&last)))
				{
				config_error(line_count, "Variable definition corrupt");
				continue;
				}
			strncpy(var_module,p,sizeof(varname));
			if (!(p = strtok_r(NULL, ":",&last)))
				{
				config_error(line_count, "Variable definition corrupt");
				continue;
				}
			strncpy(varname,p,sizeof(varname));

			/* BEGIN San Mod 07/12/00 */
		  q = &saved_line_buffer[((p - line_buffer)+(strlen(p)+1))];
			strncpy(value,q,(sizeof(value)-1));
			/* END San Mod 07/12/00 */
// printf("variable %s value = %s\n",varname,value);
			if (!strncasecmp(var_module, "GLOBAL", 6))
				variable_set(node, NULL, varname, value, 0);
			else
				variable_set(node,var_module, varname, value, 0);
			}
		else
			{
			config_error(line_count,"Syntax Error");
			}
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL,"Configuration file read complete\n");
	fclose(Tfilep_ptr);
	configuration_save();
	return(0);
}

int configuration_save()
{
	NODE		*node;
	NODEGROUP					*group;
	VARIABLE_LIST			*variable;
	NEXXUS_USER				*user;
	USER_MODULE_ACL		*mod_acl;
	USER_ADDRESS_RULE	*addr_rule;
	MODULE_RULE				*mod_rule;
	USER_GROUP				*user_group;
	char							new_path[255];
	char							groupstr[255];
	FILE							*Tfilep_ptr;
	time_t						mytime;
	mode_t						mask;

	mask = 0x1ff;
	mask &= ~S_IRUSR;
	mask &= ~S_IWUSR;
	umask(mask);
	mytime = time(NULL);
	if (!nexxus_continuous_config_backups)
		{
		sprintf(new_path, PKGLIBDIR "/vacm_configuration_backup");
		if (rename(PKGLIBDIR "/vacm_configuration", new_path) < 0)
			nexxus_log_printf(NULL,"Unable to backup configuration file (%m)\n");
		}
	else
		{
		sprintf(new_path, PKGLIBDIR "/vacm_configuration_backup.0x%x",(int)mytime);
		if (rename(PKGLIBDIR "/vacm_configuration", new_path) < 0)
			nexxus_log_printf(NULL,"Unable to backup configuration file (%m)\n");
		}
	unlink(PKGLIBDIR "/vacm_configuration");
	if (!emergency_mode)
		{
		if (!(Tfilep_ptr = fopen(PKGLIBDIR "/vacm_configuration","w")))
			{
    	if (errno == EMFILE)
				{
      	printf("[Nexxus] Out of open files. Config system in emergency mode\n");
      	fclose(stderr);
      	if (!(emergency_mode = fopen(PKGLIBDIR "/vacm_configuration","w")))
					{
        	printf("[Nexxus] Unable to enable emergency mode (%m)\n");
					return(-1);
					}
      	Tfilep_ptr = emergency_mode;
				}
			else
				{
				nexxus_log_printf(NULL,"Unable to create configuration file\n");
				return(-1);
				}
			}
		}
	else
		{
		/* Close the FD and reopen it */
		fclose(emergency_mode);
   	if (!(emergency_mode = fopen(PKGLIBDIR "/vacm_configuration","w")))
			{
     	printf("[Nexxus] Unable to enable emergency mode (%m)\n");
			return(-1);
			}
   	Tfilep_ptr = emergency_mode;
		}
	fprintf(Tfilep_ptr, "# VACM Configuration file\n");
	fprintf(Tfilep_ptr, "# Autogenerated by Nexxus - %s\n",ctime(&mytime));
	
	/* Save out all the nodes and groups */
	group = nodegroup_head;
	while(group)
		{
		fprintf(Tfilep_ptr, "GROUP:%s\n",group->group_name);
		node = group->nodelist;
		while(node)
			{
			fprintf(Tfilep_ptr, "#Definition for node '%s'\n",node->node_id);
			fprintf(Tfilep_ptr, "NODE:%s\n",node->node_id);
			variable = node->vars;
			while(variable)
				{
				fprintf(Tfilep_ptr, "VAR:%s:%s:%s\n",
							((variable->module) ? variable->module : "GLOBAL"),
							variable->variable_name,
							variable->variable_value);
				variable = variable->next;
				}
			fprintf(Tfilep_ptr, "END_NODE\n");
			if (node->next)
				fprintf(Tfilep_ptr, "\n");
			node = node->next;
			}
		fprintf(Tfilep_ptr,"GROUP_END\n");
		if (group->next)
			fprintf(Tfilep_ptr,"\n");
		group = group->next;
		}
	fprintf(Tfilep_ptr,"\n#End of node definitions\n\n");
	/* Save out all the users */
	user = user_head;
	while(user)
		{
		fprintf(Tfilep_ptr, "#Definition for user '%s'\n",user->user_name);
		fprintf(Tfilep_ptr, "USER:%s\n",user->user_name);
		fprintf(Tfilep_ptr, "PASSWORD:%s\n",user->password);

		user_group = user->groups;
		strcpy(groupstr,"");
		while(user_group)
			{
			strcat(groupstr,user_group->group_name);
			if (user_group->next)
				strcat(groupstr,":");
			user_group = user_group->next;
			}
		fprintf(Tfilep_ptr,"GROUP:%s\n",groupstr);

		fprintf(Tfilep_ptr,
						"DEFAULT_MODULE_POLICY:%s\n",
						(user->default_module_acl_policy == POLICY_ALLOW ? "ALLOW" : "DENY"));
		fprintf(Tfilep_ptr,
						"DEFAULT_ADDRESS_POLICY:%s\n",
						(user->default_address_acl_policy == POLICY_ALLOW ? "ALLOW" : "DENY"));

		mod_acl = user->mod_acl;
		while(mod_acl)
			{
			mod_rule = mod_acl->rules;
			while(mod_rule)
				{
				fprintf(Tfilep_ptr,
								"MOD_ACL:%s:%s:%s\n",
								(mod_rule->policy == POLICY_ALLOW ? "ALLOW" : "DENY"),
								mod_acl->module_ipc_tag,
								mod_rule->ipc_glob);
				mod_rule = mod_rule->next;
				}
			mod_acl = mod_acl->next;
			}
		addr_rule = user->addr_acl;
		while(addr_rule)
			{
			fprintf(Tfilep_ptr,
							"ADDR_ACL:%s:%s:%d\n",
							(addr_rule->policy == POLICY_ALLOW ? "ALLOW" : "DENY"),
							inet_ntoa(addr_rule->ipv4_address),
							addr_rule->ipv4_sigbits);
			addr_rule = addr_rule->next;
			}
		fprintf(Tfilep_ptr, "END_USER\n\n");
		user = user->next;
		}
	if (!emergency_mode)
		fclose(Tfilep_ptr);
	return(0);
}

/*
 * Runs through the configuration list and sends data only to the module
 * specified by 'ipc_tag' This is for loose modules which may connect
 * to nexxus *after* nexxus has done the main configuration file read
 */
int configuration_dispatch_to_loose_module(char *ipc_tag)
{
	NODE	*node;
	NODEGROUP				*group;
	VARIABLE_LIST		*variable;
	char						event_string[255];

	group = nodegroup_head;
	while(group)
		{
		node = group->nodelist;
		while(node)
			{
			/* Tell modules we're beginning a node definition */
			sprintf(event_string,"NODE:%s",node->node_id);
			dispatch_config_line_to_loose_module(ipc_tag, NULL, event_string);
	
			/* Send variables */
			variable = node->vars;
			while(variable)
				{
				if (((variable->module) &&(!strcmp(variable->module, ipc_tag))) || (!variable->module))
					{
					// Variable is either for our module or is a global
					sprintf(event_string,"%s:%s:%s:%s",
								((variable->module) ? variable->module : "GLOBAL"),
								node->node_id,
								variable->variable_name,
								variable->variable_value);
					if (dispatch_config_line_to_loose_module(ipc_tag,
																				 			 	variable->module,
																				 			 	event_string) <0)
						{
						if (nexxus_debug)
							nexxus_log_printf(NULL,"Warning: Module %s is not available\n",
										variable->module);
						}
					}
				variable = variable->next;
				}
			/* Tell the module the node definition is complete */
			sprintf(event_string,"END_NODE:%s",node->node_id);
			dispatch_config_line_to_loose_module(ipc_tag, NULL, event_string);
			node = node->next;
			}
		group = group->next;
		}
	return(0);
}

/*
 * Runs through the configuration list and sends appropriate data to the
 * modules that care about it
 */
int configuration_dispatch()
{
	NODE	*node;
	NODEGROUP				*group;
	VARIABLE_LIST		*variable;
	char						event_string[255];

	group = nodegroup_head;
	while(group)
		{
		node = group->nodelist;
		while(node)
			{
			/* Tell modules we're beginning a node definition */
			sprintf(event_string,"NODE:%s",node->node_id);
			dispatch_config_line(NULL, event_string);
	
			/* Send variables */
			variable = node->vars;
			while(variable)
				{
				sprintf(event_string,"%s:%s:%s:%s",
							((variable->module) ? variable->module : "GLOBAL"),
							node->node_id,
							variable->variable_name,
							variable->variable_value);
				if (dispatch_config_line(variable->module, event_string) <0)
					{
					if (nexxus_debug)
						nexxus_log_printf(NULL,"Warning: Module %s is not available\n",
									variable->module);
					}
				variable = variable->next;
				}
			/* Tell the module the node definition is complete */
			sprintf(event_string,"END_NODE:%s",node->node_id);
			dispatch_config_line(NULL, event_string);
			node = node->next;
			}
		group = group->next;
		}
	return(0);
}

static void config_error(int line, char *msg)
{
	nexxus_log_printf(NULL,"Syntax Error on line %d of configfile\n",line);
	nexxus_log_printf(NULL,"-- %s\n",msg);
}

static int	variable_get(NODE *node,
								 char						*module,
								 char						*variable,
								 char						*buffer,
								 int						maxbuffer)
{
	VARIABLE_LIST	*scan;

	if (nexxus_debug)
		nexxus_log_printf(NULL,"Variable get for '%s' in module '%s', node '%s'  requested\n",
					 variable, module, node->node_id);
	scan = node->vars;
	while(scan)
		{
		if ((module) && (scan->module))
			{
			if ((!strcasecmp(scan->variable_name, variable)) &&
					(!strcasecmp(scan->module, module)))
				{
				strncpy(buffer, scan->variable_value, maxbuffer);
				return(0);
				}
			}
		else if ((!module) && (!scan->module))
			{
			if (!strcasecmp(scan->variable_name, variable))
				{
				strncpy(buffer, scan->variable_value, maxbuffer);
				return(0);
				}
			}
		scan = scan->next;
		}
	return(-1);
}

static int variable_set(NODE	*node,
								 char						*module,
								 char						*variable,
								 char						*value,
								 int						notify)
{
	VARIABLE_LIST	*new = NULL;
	VARIABLE_LIST	*scan;
	char					tmp_str[255];
	
	sprintf(tmp_str,"variable <%s> value",variable);
	if (nexxus_debug)
		nexxus_log_printf(NULL,
						"Variable '%s' set to '%s' for module '%s' on node '%s'\n",
						variable, value, module, node->node_id);
	scan = node->vars;
	while(scan)
		{
		if ((module) && (scan->module))
			{
			if ((!strcasecmp(scan->variable_name, variable)) &&
					(!strcasecmp(scan->module, module)))
				{
				/* BEGIN San mods 07/12/00 */
				if (!value)
					{
					/* Erase variable */
					variable_destroy(node, scan);
					return(0);
					}
				/* END San mods 07/12/00 */

				nettmcs_free(scan->variable_value);
				scan->variable_value = (char *) nettmcs_alloc ((strlen(value)+1),
																											 tmp_str);
				strcpy(scan->variable_value, value);
				new = scan;
				break;
				}
			}
		else if ((!module) && (!scan->module))
			{
			if (!strcasecmp(scan->variable_name, variable))
				{
				/* BEGIN San mods 07/12/00 */
				if (!value)
					{
					/* Erase variable */
					variable_destroy(node, scan);
					return(0);
					}
				/* END San mods 07/12/00 */
				nettmcs_free(scan->variable_value);
				scan->variable_value = (char *) nettmcs_alloc ((strlen(value)+1),
																											 tmp_str);
				strcpy(scan->variable_value, value);
				new = scan;
				break;
				}
			}
		scan = scan->next;
		}
	if (!scan)
		{
		/* BEGIN San mods 07/12/00 */
		if (!value)
			return(0);
		/* END San mods 07/12/00 */
		sprintf(tmp_str,"variable <%s> structure",variable);
		if (!(new = (VARIABLE_LIST *) nettmcs_alloc (sizeof(VARIABLE_LIST),
																							 	tmp_str)))
			{
			nexxus_log_printf(NULL,"Unable to allocate memory for variable entry\n");
			return(-1);
			}
		memset(new, 0, sizeof(VARIABLE_LIST));
		if (!node->vars)
			node->vars = new;
		else
			{
			scan = node->vars;
			while (scan->next)
				scan = scan->next;
			scan->next = new;
			new->prev = scan;
			}
		if (module)
			{
			sprintf(tmp_str,"variable <%s> module name",variable);
			new->module = (char *) nettmcs_alloc((strlen(module)+1), tmp_str);
			strcpy(new->module, module);
			}
		sprintf(tmp_str,"variable <%s> name",variable);
		new->variable_name = (char *) nettmcs_alloc((strlen(variable)+1), tmp_str);
		strcpy(new->variable_name, variable);
		sprintf(tmp_str,"variable <%s> value",variable);
		new->variable_value = (char *) nettmcs_alloc ((strlen(value)+1), tmp_str);
		strcpy(new->variable_value, value);
		}
	// Now see if we have to notify modules of the change
	if (notify)
		{
		if (!module)
			{
			char	event_string[255];
			// Only global's need be resent
			sprintf(event_string,
							"GLOBAL:%s:%s:%s",
							node->node_id,
							new->variable_name,
							new->variable_value);
			dispatch_config_line(NULL, event_string);
			}
		}
	return(0);
}

/*
 * Function added 07/12/00 by San 
 */
static void variable_destroy(NODE *node, VARIABLE_LIST *var)
{
  if (nexxus_debug)
    nexxus_log_printf(NULL,"Destroying variable %s\n",var->variable_name);
  if (var->prev == NULL)
    {
    if (var != node->vars)
      {
      printf("NEXXUS ASSERT FAILURE!!!!\n");
      abort();
      }
		node->vars = var->next;
    if (node->vars)
			node->vars->prev = NULL;
    } 
  else if (var->next == NULL)
    {
    var->prev->next = NULL;
    }
  else
    {
    var->next->prev = var->prev;
    var->prev->next = var->next;
    }
	if (var->module)
		nettmcs_free(var->module);
	if (var->variable_name)
		nettmcs_free(var->variable_name);
	if (var->variable_value)
		nettmcs_free(var->variable_value);
  nettmcs_free(var);
}

/* 
 * END modifications by San 07/12/00
 */

int nexxus_variable_set(char	*node_id,
								 		 		char	*module,
								 		 		char	*variable,
								 		 		char	*value)
{
	NODE	*node;
	NODEGROUP				*group;
	int							rc;
	int							i;
	char	*myVar;

	myVar = strdup(variable);
	for(i=0; i< (int) strlen(myVar); i++)
		{
		if(isalpha(myVar[i]))
			myVar[i] = toupper(myVar[i]);
		}

	group = nodegroup_head;
	while(group)
		{
		node = group->nodelist;
		while(node)
			{
			if (!strcmp(node_id, node->node_id))
				{
				rc = variable_set(node, module, myVar, value, 1);
				configuration_save();
				free(myVar);
				return(rc);
				}
			node = node->next;
			}
		group = group->next;
		}
	free(myVar);
	return(-1);
}

int	nexxus_variable_get(char	*node_id,
								 		 		char	*module,
								 		 		char	*variable,
								 		 		char	*buffer,
								 		 		int		maxbuffer)
{
	NODE	*node;
	NODEGROUP				*group;

	group = nodegroup_head;
	while(group)
		{
		node = group->nodelist;
		while(node)
			{
			if (!strcmp(node_id, node->node_id))
				return(variable_get(node, module, variable, buffer, maxbuffer));
			node = node->next;
			}
		group = group->next;
		}
	return(-1);
}
