/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#ifndef _IPC_CONNECTION_H
#define _IPC_CONNECTION_H

#include <config.h>
#include <nexxus_users.h>
#include <nexxus.h>
#include <nexxus_nodes.h>

#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

typedef struct ipc_connection
	{
	int 						I_fd;
	short						chatty;	/* send back info on jobstart */
#ifdef OPENSSL
	SSL							*ssl;
	X509						*client_cert;
#endif
	NEXXUS_USER			*user;
	struct in_addr	conn_address;
  time_t          idle_timer;

	struct ipc_connection	*next;
	struct ipc_connection *prev;
	} IPC_CONNECTION;

struct ipc_commands
	{
	char	ipc_cmd[128]; 				/* command string */
	int		args;         				/* number of args after command string */
	int		first_is_node;				/* Set if the first argument is a node */
	int		requires_auth;				/* Cmd requires user be authenticated? */
	char 	*(*function)(IPC_CONNECTION *,
										 NEXXUS_USER 		*,
										 NODE	*,
										 __uint32_t			 ,
										 char 					**); /* callback function */
	};
#endif
