
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Notes: This is now a stand-alone run-time linkable module for VACM
 * Symbols beginning with 'api' denote a module harness required
 * symbol which should not be renamed. Renaming an 'api' symbol
 * will cause undesired operation, and program failure.
 *
 */

#ifndef _PROTOTYPES_H
#define _PROTOTYPES_H

#include <nexxus.h>
#include <nexxus_users.h>
#include <nexxus_nodes.h>
#include <module_harness.h>
#include <ipc_connection.h>
#include <stdarg.h>
#include <sys/types.h>
#include <config.h>

#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

/*
 * nexxus.c
 */
int send_socket_ipc(IPC_CONNECTION *ipc, char *msg, __uint32_t len);
int send_msg_to_client_by_fd(int I_fd, char *msg, __uint32_t len);
int send_msg_to_all_clients(char *msg, __uint32_t len);
void destroy_ipc_connection(IPC_CONNECTION *ipc);
void nexxus_timer_destroy(void *arg);
void *nexxus_timer_add(unsigned int seconds_till_alarm,
											void (*cb) (void *),
											void *arg);
void *nexxus_fd_watch(int		fd,
											void	(*cb) (int, void *),
											void	*arg);
void nexxus_fd_unwatch(int fd);
int nexxus_log_printf(char *module, char *fmt, ...);
#ifdef OPENSSL
SSL	*lookup_ssl_by_fd(int fd);
#endif

/*
 * configuration_harness.c
 */
int configuration_load(void);
int configuration_save(void);
int configuration_dispatch(void);
int configuration_dispatch_to_loose_module(char *ipc_tag);
int nexxus_variable_set(char	*node_id,
												char	*module,
												char	*variable,
												char	*value);
int	nexxus_variable_get(char	*node_id,
												char	*module,
												char	*variable,
												char	*buffer,
												int		maxbuffer);

/*
 * module_harness.c
 */
int	start_loose_module(char *filename);
int identify_modules(void);
int load_modules(char *filename);
int lookup_ipc_tag(char *key, char *out);
void notify_modules_of_disconnection(int fd);
int dispatch_ipc_to_module(int				I_fd,
													 char				*key,
													 char				*msg,
													 char				*tag,
													 char				*node_id,
													 __uint32_t	job_id);
int dispatch_config_line(char *module, char *msg);
int alloc_and_accept_loose_module_connection(int I_fd);
void add_loose_modules(int *max, fd_set *read_fds);
void dispatch_loose_module_ipc(fd_set *read_fds);
int return_num_modules(void);
void shutdown_modules(int severity);
int shutdown_module(char *ipc_tag);
int dispatch_config_line_to_loose_module(char *ipc_tag, char *module, char *msg);
void nexxus_return_module_error_string(int err, char *buf);
void *nexxus_lookup_backchannel_function_for_module(char *ipc_tag);
char *return_module_state(VACM_MODULE_LIST *mlist);
void handle_child_exit(pid_t pid, int status);

/*
 * locks.c
 */
int client_fd_lock_create(int fd);
void client_fd_lock_destroy(int fd);
void nexxus_do_lock(int fd);
void nexxus_do_unlock(int fd);

/*
 * nettmcs.c
 */
void *nettmcs_alloc(unsigned int size, char *label);
void nettmcs_free(void *ptr);
void nettmcs_dump(void);

/*
 * logging_harness.c
 */
int do_nexxus_log_printf(char *module, char *fmt, va_list ap);

/*
 * nexxus_users.c
 */
NEXXUS_USER	*user_create(char	*user_name);
void				user_destroy(NEXXUS_USER	*user);
int					user_exists(char *user_name);
NEXXUS_USER	*user_lookup(char *user_name);
USER_GROUP	*user_group_member_add(NEXXUS_USER *user, char *group_name);
void user_group_member_remove(NEXXUS_USER *user, char *group_name);
USER_GROUP	*user_group_user_is_member(NEXXUS_USER *user, char *group_name);

/* module acl stuff */
USER_MODULE_ACL	*user_module_acl_lookup(NEXXUS_USER	*user,
																				char				*module);
void						user_module_acl_destroy(NEXXUS_USER			*user,
																				USER_MODULE_ACL	*acl);
USER_MODULE_ACL	*user_module_acl_create(NEXXUS_USER	*user,
																			  char				*module);
MODULE_RULE			*user_module_acl_rule_create(USER_MODULE_ACL	*acl,
																						 int							policy,
																						 char							*glob);
MODULE_RULE			*user_module_acl_rule_lookup(NEXXUS_USER	*user,
												 										 char					*module,
												 										 int					policy,
												 										 char					*glob);
void user_module_acl_rule_destroy(USER_MODULE_ACL	*acl,
																	MODULE_RULE			*rule);
/* address acl stuff */
USER_ADDRESS_RULE	*user_address_acl_rule_create(NEXXUS_USER			*user,
																					 			struct in_addr	*addr,
																					 			int							sigbits,
																					 			int							policy);
USER_ADDRESS_RULE *user_address_acl_rule_lookup(NEXXUS_USER		*user,
														 										int							policy,
														 										struct in_addr	*addr,
														 										int							sigbits);
void user_address_acl_rule_destroy(NEXXUS_USER				*user,
																	 USER_ADDRESS_RULE	*rule);
/* permissions stuff */
int authenticate_user(char *user_name, char *password);
int verify_ipc_permissions(NEXXUS_USER *user, char *module, char *cmd_request);
int verify_address_permissions(NEXXUS_USER *user, struct in_addr *addr);
void verify_users(void);
void verify_groups(void);

/* nexxus_nodes.c */
NODEGROUP *nodegroup_create(char *group_name);
void nodegroup_destroy(NODEGROUP *group);
NODEGROUP *nodegroup_lookup_by_name(char *group_name);
NODEGROUP *nodegroup_lookup_by_node(NODE *node);
void nodegroup_add_node(NODEGROUP *group, NODE *node);
void nodegroup_remove_node(NODEGROUP *group, NODE *node);
NODE *nodelist_entry_create(char *node_id);
NODE *nodegroup_node_lookup(NODEGROUP *group, char *node_name);
NODE *node_lookup_by_glob(char *glob, NODEGROUP **last_gr, NODE **last_n);
int nodelist_entry_rename(NEXXUS_USER *user, char *old_name, char *new_name);
int nodelist_entry_delete(NEXXUS_USER *user, char *node_id);

/* nexxus_ipc.c */
char	*ipc_nexxus_debug(IPC_CONNECTION	*ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_nexxus_modules(IPC_CONNECTION	*ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_add(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_del(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_rename(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_list(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_var_list(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_get_var(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_set_var(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_version(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_auth(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_echo(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_add(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_del(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_rename(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_list(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_list_groups(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_add_mod_acl_rule(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_del_mod_acl_rule(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_list_mod_acl_rules(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_add_addr_acl_rule(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_del_addr_acl_rule(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_list_addr_acl_rules(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_list_default_acl_policy(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_chg_default_acl_policy(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_chg_password(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_admin_set_password(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_add(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_del(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_list(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_rename(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_add_admin(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_group_del_admin(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_node_set_group(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_set_chatty(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_module_unload(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_module_load(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_list_clients(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
char	*ipc_wall(IPC_CONNECTION *ipc, NEXXUS_USER *user, NODE *node, __uint32_t job_id, char **args);
#endif
