/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#include <config.h>

#include <stdio.h>
#include <sys/socket.h>
#include <signal.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/resource.h>
#include <nexxus.h>
#include <module_harness.h>
#include <ipc_connection.h>
#include <nexxus_users.h>
#include <prototypes.h>
#include <config.h>
#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define SSL_HOME PKGLIBDIR
#define CERTF SSL_HOME "/vacm.cert"
#define KEYF SSL_HOME "/vacm.key"
#endif

/* Private nexxus.c function prototypes */
static void 					signal_handler(int I_signo);
static IPC_CONNECTION	*alloc_and_accept_ipc_connection(int);
static void						nexxus_fd_watch_destroy(NEXXUS_FD_WATCHER *ptr);
static void						destroy_sockets(void);
static void						nexxus_shutdown(int exit_code);
static void						nexxus_ipc_handler(IPC_CONNECTION *ipc);
static int						process_nexxus_cmd(IPC_CONNECTION *ipc, char *msg);
static void						create_sockets(void);
static void						nexxus_main_loop(void);
static void						nexxus_dispatch_timers(void);
static __uint32_t			get_new_job_id(void);
static void						free_ipc_connection(IPC_CONNECTION *ipc);
static void handle_responder(int fd);


bool_t										nexxus_debug = BOOL_INIT;
bool_t										nexxus_stdout_logging = BOOL_INIT;
bool_t										nexxus_continuous_config_backups = BOOL_INIT;

static bool_t							signal_semaphore = BOOL_INIT;

static int								I_inetsocket = -1;		
static int								I_respondersocket = -1;
static int								I_loose_doorbell_fd = -1;
static IPC_CONNECTION 		*ipc_head = NULL;
static IPC_CONNECTION 		*ipc_tail = NULL;
static NEXXUS_FD_WATCHER	*watcher_head = NULL;
static NEXXUS_FD_WATCHER	*watcher_tail = NULL;
static NEXXUS_TIMER				*timer_head = NULL;
static NEXXUS_TIMER				*timer_tail = NULL;
static __uint32_t					job_id_cnt =1;
#ifdef OPENSSL
static SSL_CTX						*ctx;
static SSL_METHOD					*meth;
#endif

extern int								errno;
extern int								nexxus_api_version_major;
extern int								nexxus_api_version_minor;

extern NEXXUS_USER	*user_head;
extern NEXXUS_USER	*user_tail;

/*
 * Nexxus IPC command structure
 */
static struct ipc_commands ipc_table[]={
	{"NEXXUS_DEBUG",		  						1,0,1,ipc_nexxus_debug  								},
	{"NEXXUS_MODULES",								0,0,1,ipc_nexxus_modules								},
	{"NODE_ADD", 											2,0,1,ipc_node_add 											},
	{"NODE_DEL",											1,1,1,ipc_node_del 											},
	{"NODE_RENAME",										2,1,1,ipc_node_rename										},
	{"NODE_LIST",											0,0,1,ipc_node_list											},
	{"NODE_VAR_LIST",									1,0,1,ipc_node_var_list									},
	{"GET_VAR",												2,1,1,ipc_get_var												},
	{"SET_VAR",												3,1,1,ipc_set_var												},
	{"VERSION",												0,0,1,ipc_version												},
	{"AUTH",													2,0,0,ipc_auth													},
	{"ECHO",													1,0,0,ipc_echo													},
	{"ADMIN_ADD",											3,0,1,ipc_admin_add											},
	{"ADMIN_DEL",											1,0,1,ipc_admin_del											},
	{"ADMIN_RENAME",									2,0,1,ipc_admin_rename									},
	{"ADMIN_LIST",										0,0,1,ipc_admin_list										},
	{"ADMIN_LIST_GROUPS",							1,0,1,ipc_admin_list_groups							},
	{"ADMIN_ADD_MOD_ACL_RULE",				4,0,1,ipc_admin_add_mod_acl_rule				},
	{"ADMIN_DEL_MOD_ACL_RULE",				4,0,1,ipc_admin_del_mod_acl_rule				},
	{"ADMIN_LIST_MOD_ACL_RULES",			1,0,1,ipc_admin_list_mod_acl_rules			},
	{"ADMIN_ADD_ADDR_ACL_RULE",				4,0,1,ipc_admin_add_addr_acl_rule				},
	{"ADMIN_DEL_ADDR_ACL_RULE",				4,0,1,ipc_admin_del_addr_acl_rule				},
	{"ADMIN_LIST_ADDR_ACL_RULES",			1,0,1,ipc_admin_list_addr_acl_rules			},
	{"ADMIN_LIST_DEFAULT_ACL_POLICY",	2,0,1,ipc_admin_list_default_acl_policy	},
	{"ADMIN_CHG_DEFAULT_ACL_POLICY",	3,0,1,ipc_admin_chg_default_acl_policy	},
	{"ADMIN_CHG_PASSWORD",						2,0,1,ipc_admin_chg_password						},
	{"ADMIN_SET_PASSWORD",						2,0,1,ipc_admin_set_password						},
	{"GROUP_ADD",											1,0,1,ipc_group_add											},
	{"GROUP_DEL",											1,0,1,ipc_group_del											},
	{"GROUP_LIST",										0,0,1,ipc_group_list										},
	{"GROUP_RENAME",									2,0,1,ipc_group_rename									},
	{"GROUP_ADD_ADMIN",								2,0,1,ipc_group_add_admin								},
	{"GROUP_DEL_ADMIN",								2,0,1,ipc_group_del_admin								},
	{"NODE_SET_GROUP",								2,1,1,ipc_node_set_group								},
	{"MODULE_UNLOAD",                 1,0,1,ipc_module_unload                 },
	{"MODULE_LOAD",            	     	1,0,1,ipc_module_load      		          },
	{"WALL",											 		1,0,1,ipc_wall													},
	{"IPC_CHATTY",								 		1,0,0,ipc_set_chatty										},
	{"LIST_CLIENTS",    					 		0,0,1,ipc_list_clients									},
	{"",														-1,-1,-1,NULL}};

int main(int argc,char **argv)
{
	int						num_modules;
	int						rc, old_logging;
	struct rlimit	limit;

	signal (SIGINT, signal_handler);
	signal (SIGTERM, signal_handler);
	signal (SIGHUP, signal_handler);
	signal (SIGKILL, signal_handler);
	signal (SIGUSR1, signal_handler);
	signal (SIGPIPE, SIG_IGN);

	if (getuid() != 0)
		{
		printf("[Nexxus] You must be ROOT to run Nexxus\n");
		exit(-1);
		}

	memset(&limit, 0, sizeof(limit));
	limit.rlim_cur = RLIM_INFINITY;
	if (setrlimit(RLIMIT_CORE, &limit)<0)
		{
		perror("setrlimit");
		exit(-1);
		}
	memset(&limit, 0, sizeof(limit));
	limit.rlim_cur = RLIM_INFINITY;
	if (setrlimit(RLIMIT_CORE, &limit)<0)
		{
		perror("setrlimit");
		exit(-1);
		}
	chdir ("/");
#if 0
	printf("[Nexxus] Maximum core size set to %d (max %d)\n",
				(int)limit.rlim_cur,
				(int)limit.rlim_max);
	if (getrlimit(RLIMIT_NOFILE, &limit)<0)
		{
		perror("getrlimit");
		exit(-1);
		}
	printf("[Nexxus] Maximum open files set to %d (max %d)\n",
				(int)limit.rlim_cur,
				(int)limit.rlim_max);
	exit(0);
#endif
	while ((rc = getopt(argc, argv, "dDbl"))!=EOF)
		{
		if ((char) rc == 'b')
			{
			nexxus_log_printf(NULL,"Continuous Configuration Backups Enabled\n");
			BOOL_SET(nexxus_continuous_config_backups);
			}
		if ((char) rc == 'd')
			{
			printf("[Nexxus] Debugging Enabled\n");
			BOOL_SET(nexxus_debug);
			}
		if ((char) rc == 'D')
			{
			printf("-D option is deprecated. Use & to background nexxus\n");
			exit(1);
#if 0
			BOOL_SET(nexxus_stdout_logging);
			nexxus_log_printf(NULL,"Daemon Mode Enabled\n");
			BOOL_UNSET(nexxus_stdout_logging);
#endif
			}
		if ((char) rc =='l')
			{
			BOOL_SET(nexxus_stdout_logging);
			}
		}
	old_logging = nexxus_stdout_logging;
	BOOL_SET(nexxus_stdout_logging);
	nexxus_log_printf("VACM","%s Nexxus daemon (Build %s %s)\n",
					VERSION,
					__DATE__,
					__TIME__);
	if (!old_logging)
		BOOL_UNSET(nexxus_stdout_logging);

	if (nexxus_stdout_logging)
		nexxus_log_printf(NULL,"Standard Out Logging Enabled\n");
	if (nexxus_debug)
		{
		BOOL_SET(nexxus_stdout_logging);
		}
	if (configuration_load() <0)
		{
		nexxus_log_printf(NULL,"Configuration process failed. Exiting\n");
		nexxus_shutdown(-1);
		}
	create_sockets();
	
#ifdef OPENSSL
	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
	meth = SSLv23_server_method();
	ctx = SSL_CTX_new(meth);
	if (!ctx)
		{
		ERR_print_errors_fp(stdout);
		nexxus_shutdown(-1);
		}
	if (SSL_CTX_use_certificate_file(ctx, CERTF, SSL_FILETYPE_PEM) <=0)
		{
		ERR_print_errors_fp(stdout);
		nexxus_shutdown(-1);
		}
	if (SSL_CTX_use_PrivateKey_file(ctx, KEYF, SSL_FILETYPE_PEM) < 0)
		{
		ERR_print_errors_fp(stdout);
		nexxus_shutdown(-1);
		}
	if (!SSL_CTX_check_private_key(ctx))
		{
		nexxus_log_printf(NULL,
													"Prv key does not match the cert public key\n");
		nexxus_shutdown(-1);
		}

	nexxus_log_printf(NULL,"OpenSSL encryption layer enabled\n");
	
#endif
	num_modules = identify_modules();
	nexxus_log_printf(NULL,"Identified %d modules\n",num_modules);

	/* verify that at least one user exists. if not create blum frub */
	verify_users();
	verify_groups();

	if (configuration_dispatch() <0)
		{
		nexxus_log_printf(NULL,"Exiting startup because of configuration error\n");
		nexxus_shutdown(-1);
		}
	nexxus_main_loop();
	nexxus_shutdown(0);
	exit(0); // duhhh
}

static void nexxus_main_loop()
{
	IPC_CONNECTION		*ipc;
	fd_set						read_fds;
	struct timeval		to;
	int								max, rc, status;
	pid_t							pid;
	NEXXUS_FD_WATCHER	*watcher;

	while(1)
		{
		max = 0;
		if (I_inetsocket > max)
			max = I_inetsocket;
		if (I_respondersocket > max)
			max = I_respondersocket;
		if (I_loose_doorbell_fd > max)
			max = I_loose_doorbell_fd;

		to.tv_usec = 0;
		to.tv_sec = 1;
		FD_ZERO(&read_fds);
		FD_SET(I_inetsocket, &read_fds);
		FD_SET(I_respondersocket, &read_fds);
		FD_SET(I_loose_doorbell_fd, &read_fds);

		while((pid = wait3(&status, WNOHANG, NULL)))
			{
			if (pid <0)
				{
				nexxus_log_printf(NULL,"wait3 failed (%m)\n");
				sleep(1);
				continue;
				}
			handle_child_exit(pid, status);
			}
		add_loose_modules(&max, &read_fds);
		ipc = ipc_head;
		while(ipc)
			{
			FD_SET(ipc->I_fd, &read_fds);
			if (ipc->I_fd > max)
				max = ipc->I_fd;
			ipc = ipc->next;
			}
		/* Add any watchers we may have */
		watcher = watcher_head;
		while(watcher)
			{
			if (watcher->fd != -1)
				{
				if (max < watcher->fd)
					max = watcher->fd;
				FD_SET(watcher->fd, &read_fds);
				}
			watcher = watcher->next;
			}
		rc = select((max + 1), &read_fds, NULL, NULL, &to);
		if (rc < 0)
			{
			if (errno == EINTR)
				continue;
			nexxus_log_printf(NULL,"select() failure (%m)\n");
			sleep(5);
			continue;
			}
		if (!rc)
			{
			nexxus_dispatch_timers();
			}
		else
			{
			if (FD_ISSET(I_inetsocket, &read_fds))
				{
				if (nexxus_debug)
					nexxus_log_printf(NULL,"Remote IPC connection request.\n");
				if (!(ipc = alloc_and_accept_ipc_connection(I_inetsocket)))
					continue;
				}
			if (FD_ISSET(I_loose_doorbell_fd, &read_fds))
				{
				if (nexxus_debug)
					nexxus_log_printf(NULL,"Loose module connection request.\n");
				alloc_and_accept_loose_module_connection(I_loose_doorbell_fd);
				}
			if (FD_ISSET(I_respondersocket, &read_fds))
				{
				if (nexxus_debug)
					nexxus_log_printf(NULL,"Remote responder message available.\n");
				handle_responder(I_respondersocket);
				}
			dispatch_loose_module_ipc(&read_fds);

			ipc = ipc_head;
			while(ipc)
				{
				if (FD_ISSET(ipc->I_fd, &read_fds))
					{
					if (nexxus_debug)
						nexxus_log_printf(NULL,
																	"IPC message received from client %d\n",
																	ipc->I_fd);
					nexxus_ipc_handler(ipc);
					}
				ipc = ipc->next;
				}
			watcher = watcher_head;
			while(watcher)
				{
				if (watcher->fd != -1)
					{
					if (FD_ISSET(watcher->fd, &read_fds))
						{
						watcher->cb(watcher->fd, watcher->arg);
						}
					}
				watcher = watcher->next;
				}
			/* Look for any watchers that may have to be nukated */
			watcher = watcher_head;
			while(watcher)
				{
				if (watcher->fd == -1)
					{
					nexxus_fd_watch_destroy(watcher);
					watcher = watcher_head; /* BUG 100141 fix */
					if (!watcher) break;
					continue;
					}
				watcher = watcher->next;
				}
			/**/
			nexxus_dispatch_timers();
			}
		}
}

static void handle_responder(int fd)
{
	struct sockaddr_in	addr_recv;
	struct sockaddr_in	addr_send;
	char								ping_msg[16];
	int									fromlen;

	fromlen = sizeof(struct sockaddr_in);
	if (recvfrom(fd,
							 (void *) &ping_msg[0],
							 sizeof(ping_msg),
							 0,
							 (struct sockaddr *) &addr_recv,
							&fromlen) < 0)
		{
		nexxus_log_printf(NULL,"Unable to read PING msg (%m)\n");
		return;
		}

	if (strncmp(ping_msg,"PING",4))
		{
		nexxus_log_printf(NULL,"Received NON 'PING' dgram on responder socket\n");
		return;
		}

	memset(&addr_send, 0, sizeof(addr_send));
	addr_send.sin_family = AF_INET;
	addr_send.sin_port = htons(6446);
	addr_send.sin_addr.s_addr = addr_recv.sin_addr.s_addr;

	strcpy(ping_msg,"PONG");
	if (sendto(fd,
						 (void *) &ping_msg[0],
						 (strlen(ping_msg)+1),
						 0,
						 (struct sockaddr *) &addr_send,
						 sizeof(addr_send))<0)
		{
		nexxus_log_printf(NULL,"Unable to send PONG (%m)\n");
		return;
		}
	return;
}

static void nexxus_dispatch_timers(void)
{
	NEXXUS_TIMER	*timers;
	unsigned int	current_time;

	if (!timer_head)
		return;
	current_time = time(NULL);

	timers = timer_head;

	while(timers)
		{
		if (current_time >= timers->alarm_time)
			{
			timers->cb(timers->arg);
			timers->alarm_time = 0;
			}
		timers = timers->next;
		}
	timers = timer_head;
	while(timers)
		{
		if (timers->alarm_time == 0)
			{
			nexxus_timer_destroy(timers);
			timers = timer_head;
			if (!timers)
				break;
			continue;
			}
		timers = timers->next;
		}
}

void nexxus_timer_destroy(void *arg)
{
	NEXXUS_TIMER *ptr = (NEXXUS_TIMER *) arg;

	if (ptr == timer_head)
		{
		timer_head = timer_head->next;
		if (timer_head)
			timer_head->prev = NULL;
		else
			timer_tail = NULL;
		}
	else if (ptr == timer_tail)
		{
		timer_tail = timer_tail->prev;
		if (timer_tail)
			timer_tail->next = NULL;
		else
			timer_head = NULL;
		}
	else
		{
		ptr->next->prev = ptr->prev;
		ptr->prev->next = ptr->next;
		}
	nettmcs_free(ptr);
}

void *nexxus_timer_add(unsigned int seconds_till_alarm,
											 void	(*cb) (void *),
											 void	*arg)
{
	NEXXUS_TIMER	*new;

	if (!(new = (NEXXUS_TIMER *) nettmcs_alloc (sizeof(NEXXUS_TIMER),
																							"nexxus timer")))
		{
		nexxus_log_printf(NULL,"Unable to allocate memory for Nexxus timer\n");
		return(NULL);
		}
	memset(new, 0, sizeof(NEXXUS_TIMER));
	
	if (!timer_tail)
		timer_tail = timer_head = new;
	else
		{
		timer_tail->next = new;
		new->prev = timer_tail;
		timer_tail = new;
		}
	timer_tail = new;
	new->cb = cb;
	new->alarm_time = time(NULL) + seconds_till_alarm;
	new->arg = arg;
	return((void *) new);
}

/*
 * Read the incomming IPC message and dispatch it to somebody
 */
static void nexxus_ipc_handler(IPC_CONNECTION *ipc)
{
	char				node_id[NODE_NAME_MAXLENGTH];
	char				module[255];
	char				*last = (char *) 0x00;
	__uint32_t	msglen = 0;
	char				*msg;
	char				*savedmsg;
	char				*bword;
	char				*offset;
	__uint32_t	brtr;
	int					rc;
	__uint32_t	job_id;
#ifdef OPENSSL
	int					err;
#endif


#ifdef OPENSSL
	if (nexxus_debug)
		nexxus_log_printf(NULL,
													"Calling SSL_read for fd %d (length)\n",
													ipc->I_fd);
	err = SSL_read(ipc->ssl, (char *) &msglen, sizeof(msglen));
	if ((!err) || (err <0))
		goto error_disconnect;

	msglen = ntohl(msglen);
	if (msglen > MAX_IPC_LENGTH)
		{
		nexxus_log_printf(NULL,"IPC message length greater than MAX_IPC_LEN. (fd %d)\n",ipc->I_fd);
		goto error_disconnect;
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL,
													"Allocating %d byte IPC buffer\n",
													msglen);
	if (!(msg = (char *) nettmcs_alloc((msglen+1), "message")))
		{
		nexxus_log_printf(NULL,
													"Unable to nettmcs_alloc msg of size %d bytes (%m)\n",
													msglen);
		return;
		}
	brtr = msglen;
	offset = msg;
	while(brtr)
		{
		err = SSL_read(ipc->ssl, offset, (int) brtr);
		if ((!err) || (err <0))
			{
			if (nexxus_debug)
				nexxus_log_printf(NULL,
															"Disconnecting client fd %d\n",
															ipc->I_fd);
			nettmcs_free(msg);
			goto error_disconnect;
			}
		brtr -= err;
		offset += err;
		}
#else
	rc = read(ipc->I_fd, &msglen, sizeof(msglen));
	if ((rc ==  0 ) || (rc < 0))
		{
		if (rc == 0)
			goto error_disconnect;
		if ((rc == -1) && ((errno !=ECONNRESET) && (errno != EPIPE)))
			{
			nexxus_log_printf(NULL,"Error while reading IPC msglen (%m)\n");
			}
		else
			{
			if (nexxus_debug)
				nexxus_log_printf(NULL,"Client hangup detected. (fd %d)\n",ipc->I_fd);
			}
		goto error_disconnect;
		}
	msglen = ntohl(msglen);
	if (rc != sizeof(msglen))
		{
		nexxus_log_printf(NULL,"Split packet for msglen detected!\n");
		goto error_disconnect;
		}
	if (msglen > MAX_IPC_LENGTH)
		{
		nexxus_log_printf(NULL,"IPC message length greater than MAX_IPC_LEN. (fd %d)\n",ipc->I_fd);
		goto error_disconnect;
		}
	if (!(msg = (char *) nettmcs_alloc((msglen+1), "message")))
		{
		nexxus_log_printf(NULL,"Unable to nettmcs_alloc msg of size %d bytes (%m)\n",msglen);
		sleep(2);
		return;
		}
	brtr = msglen;
	offset = msg;
	while(brtr)
		{
		rc = read(ipc->I_fd, offset, brtr);
		if ((rc ==  0 ) || (rc < 0))
			{
			if (rc == 0)
				return;
			if ((rc == -1) && ((errno !=ECONNRESET) || (errno != EPIPE)))
				nexxus_log_printf(NULL,"Error while reading IPC msg data (%m)\n");
			else
				{
				if (nexxus_debug)
					nexxus_log_printf(NULL,"Client hangup detected. (fd %d)\n",ipc->I_fd);
				}
			nettmcs_free(msg);
			goto error_disconnect;
			}
		brtr -= rc;
		offset += rc;
		}
#endif
	if (!(savedmsg = (char *) nettmcs_alloc(msglen+1, "saved message")))
		{
		nexxus_log_printf(NULL,"Unable to allocate memory for saved msg\n");
		nettmcs_free(msg);
		return;
		}
	memcpy(savedmsg, msg, msglen);
	savedmsg[msglen] = '\0';	/* null terminate message */
	msg[msglen] = '\0';	/* null terminate message */
	
	if (nexxus_debug)
		nexxus_log_printf(NULL,"Got msg [%s] from client %d\n",msg, ipc->I_fd);

  /* Reset our idle_timer */
  ipc->idle_timer = time(NULL);

	/* Now see which module the msg is for */
	bword = strtok_r(msg, ":",&last);
	strncpy(module,bword,(sizeof(module)-1));
	if (!strcasecmp(module,"NEXXUS"))
		process_nexxus_cmd(ipc, savedmsg);
	else
		{
		NODEGROUP		*group;
		NODE				*node;
		char				errstr[255];
		char				module_ipc_tag[32];
		int					batch_mode = 0;
		char				batch_glob[255];
		char				cmd[255];
		int					i;

		for (i=0;i<(int)strlen(module); i++)
			{
			if (isalpha(module[i]))
				module[i] = toupper(module[i]);
			}
		job_id = get_new_job_id();
		sprintf(errstr,"%s:%d:JOB_STARTED",module, job_id);
		if (ipc->chatty) 
			{
			strcat(errstr, ":");
			strcat(errstr, savedmsg);
			}
		if (send_socket_ipc(ipc, errstr,(strlen(errstr)+1))<0)
			{
			nettmcs_free(savedmsg);
			nettmcs_free(msg);
			return;
			}
		if (!lookup_ipc_tag(module, module_ipc_tag))
			{
			sprintf(errstr, "%s:%d:MODULE_NOT_FOUND",module,job_id);
			send_socket_ipc(ipc, errstr,(strlen(errstr)+1));
			nettmcs_free(savedmsg);
			nettmcs_free(msg);
			return;
			}
		if (!ipc->user)
			{
			sprintf(errstr, "%s:%d:JOB_ERROR:NEED_AUTH",module_ipc_tag,job_id);
			send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
			nettmcs_free(savedmsg);
			nettmcs_free(msg);
			return;
			}
		/* Pull out the cmd */
		if (!(bword = strtok_r(NULL,":",&last)))
			{
			sprintf(errstr,
							"%s:%d:JOB_ERROR:MALFORMATTED_MESSAGE",
							module_ipc_tag,
							job_id);
			send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
			nettmcs_free(savedmsg);
			nettmcs_free(msg);
			return;
			}
		strncpy(cmd, bword,(sizeof(cmd)-1));

		/* Pull out the target node or glob */
		if (!(bword = strtok_r(NULL,":",&last)))
			{
		// 	printf("Unable to qualify nodename for ipc msg\n");
			memset(&node_id[0], 0, sizeof(node_id));
			}
		else
			{
			strncpy(node_id, bword, (sizeof(node_id)-1));
			/* 
			 * Before doing anything here, figure out if we 
       * have to do wildcard / batch processing
			 */
			if((index(node_id,(int) '?')) ||
				 (index(node_id,(int) '*')) ||
				 (index(node_id,(int) '[')))
				{
				batch_mode = 1;
				strcpy(batch_glob, node_id);
				}
			else
				{
				/* 
		 	 	 * Check to make sure the user can see the node 
		 		 */
				if (!(node= nodegroup_node_lookup(NULL,node_id)))
					{
					sprintf(errstr,
								"%s:%d:JOB_ERROR:NODE_NOT_FOUND",
								module_ipc_tag,
								job_id);
					send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
					nettmcs_free(savedmsg);
					nettmcs_free(msg);
					return;
					}
				if (!(group= nodegroup_lookup_by_node(node)))
					{
					sprintf(errstr,
								"%s:%d:JOB_ERROR:INTERNAL_ERROR:NODE_HAS_NO_GROUP",
								module_ipc_tag,
								job_id);
					send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
					nettmcs_free(savedmsg);
					nettmcs_free(msg);
					return;
					}
				if (!(user_group_user_is_member(ipc->user, group->group_name)))
					{
					sprintf(errstr,
								"%s:%d:JOB_ERROR:NODE_NOT_FOUND",
								module_ipc_tag,
								job_id);
					send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
					nettmcs_free(savedmsg);
					nettmcs_free(msg);
					return;
					}
				}
			}
		if (!batch_mode)
			{
			/* Single shot command */
			if (!verify_ipc_permissions(ipc->user, module, cmd))
				{
				sprintf(errstr,
							"%s:%d:JOB_ERROR:PERMISSION_DENIED",
							module_ipc_tag,
							job_id);
				send_socket_ipc(ipc, errstr, (strlen(errstr)+1));
				nettmcs_free(savedmsg);
				nettmcs_free(msg);
				return;
				}
			rc=dispatch_ipc_to_module(ipc->I_fd,
																module,
																savedmsg,
																&module_ipc_tag[0],
																node_id,
																job_id);
			if (rc < 0)
				{
				if (rc == -ENOENT)
					sprintf(errstr,"%s:%d:MODULE_NOT_FOUND",module_ipc_tag, job_id);
				else
					sprintf(errstr,"%s:%d:JOB_ERROR:%s",module_ipc_tag, job_id,strerror(-rc));
				if (send_socket_ipc(ipc, errstr,(strlen(errstr)+1))<0)
					{
					nexxus_log_printf(NULL,"send_socket_ipc() failed. Cancelling batch\n");
					nettmcs_free(savedmsg);
					nettmcs_free(msg);
					return;
					}
				}
			}
		else
			{
			NODEGROUP	*group_last = NULL;
			NODE			*node_last = NULL;
			int				batch_num = 0;

			/* Batch mode command */
//			printf("Batch command processing beginning\n");
			while((node = node_lookup_by_glob(batch_glob, &group_last, &node_last)))
				{
//				printf("Node %s matched glob %s\n",node->node_id,batch_glob);
				group = nodegroup_lookup_by_node(node);
				if (!user_group_user_is_member(ipc->user, group->group_name))
					{

//					printf("User %s not a member of group %s. skipping node %s\n",
//									ipc->user->user_name,group->group_name,node->node_id);
					continue;
					}
				batch_num++;
				if (batch_num > 1)
					{
					job_id = get_new_job_id();
					sprintf(errstr,"%s:%d:JOB_STARTED",module_ipc_tag, job_id);
					if (ipc->chatty) 
						{
						strcat(errstr, ":");
						strcat(errstr, savedmsg);
						}
					if (send_socket_ipc(ipc, errstr,(strlen(errstr)+1))<0)
						{
						nexxus_log_printf(NULL,"send_socket_ipc() failed. Cancelling batch\n");
						nettmcs_free(savedmsg);
						nettmcs_free(msg);
						return;
						}
					}
				if (!verify_ipc_permissions(ipc->user, module, cmd))
					{
					sprintf(errstr,
							"%s:%d:JOB_ERROR:PERMISSION_DENIED",
							module_ipc_tag,
							job_id);
					if (send_socket_ipc(ipc, errstr, (strlen(errstr)+1))<0)
						{
						nexxus_log_printf(NULL,"send_socket_ipc() failed.\n");
						nettmcs_free(savedmsg);
						nettmcs_free(msg);
						return;
						}
					continue;
					}
				rc=dispatch_ipc_to_module(ipc->I_fd,
																	module,
																	savedmsg,
																	&module_ipc_tag[0],
																	node->node_id,
																	job_id);
				if (rc < 0)
					{
					if (rc == -ENOENT)
						sprintf(errstr,"%s:%d:MODULE_NOT_FOUND",module_ipc_tag, job_id);
					else
						sprintf(errstr,"%s:%d:JOB_ERROR:%s",module_ipc_tag, job_id,strerror(-rc));
					if (send_socket_ipc(ipc, errstr,(strlen(errstr)+1))<0)
						{
						nettmcs_free(savedmsg);
						nettmcs_free(msg);
						return;
						}
					}
				}
			if (batch_num == 0)
				{
				sprintf(errstr,
						"%s:%d:JOB_ERROR:NODE_NOT_FOUND",
						module_ipc_tag,
						job_id);
				if (send_socket_ipc(ipc, errstr, (strlen(errstr)+1))<0)
					{
					nettmcs_free(savedmsg);
					nettmcs_free(msg);
					return;
					}
				}
//			printf("Batch processing complete (%d entries processed)\n",batch_num);
			}

		}
	nettmcs_free(savedmsg);
	nettmcs_free(msg);
	return;
error_disconnect:
	notify_modules_of_disconnection(ipc->I_fd);
	destroy_ipc_connection(ipc);
	return;
}

static __uint32_t get_new_job_id()
{
	__uint32_t tmp;

	tmp = job_id_cnt;
	job_id_cnt++;
	if (job_id_cnt == 0x00)
		job_id_cnt = 1;

	if (nexxus_debug)
		nexxus_log_printf(NULL,"Allocating new job id %d\n",tmp);
	return(tmp);
}

static int process_nexxus_cmd(IPC_CONNECTION *ipc, char *msg)
{
	char						*bword, *saved_msg;
	char						*last = (char *) 0x00;
	__uint32_t			job_id;
	int							i,j;
	char						clientmsg[512], module[32],cmd[32], node_id[64];
	char						**arguments = NULL;
	NODE	*node = NULL;
	NODEGROUP				*group = NULL;
	char						*respmsg = NULL;
	char						tmp_str[255];

	for (i=0;i<(int)strlen(module); i++)
		{
		if (isalpha(module[i]))
			module[i] = toupper(module[i]);
		}
	job_id = get_new_job_id();
	if (ipc->chatty) 
		sprintf(clientmsg, "NEXXUS:%d:JOB_STARTED:%s", job_id,msg);
	else
		sprintf(clientmsg, "NEXXUS:%d:JOB_STARTED", job_id);
	if (send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1))<0)
		{
		return(-1);
		}
	saved_msg = nettmcs_alloc((strlen(msg)+1),"saved_msg");
	strcpy(saved_msg, msg);

	/* Knock the 'module' name off */
	if(!(bword = strtok_r(msg, ":", &last)))
		{
		nettmcs_free(saved_msg);
		return(-1);
		}
	strncpy(module,bword,(sizeof(module)-1));

	/* Get the 'cmd' out */
	if(!(bword = strtok_r(NULL, ":", &last)))
		{
		sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:MALFORMATTED_MESSAGE",job_id);
		send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
		nettmcs_free(saved_msg);
		return(0);
		}
	strncpy(cmd,bword,(sizeof(module)-1));

	for (i=0;ipc_table[i].args != -1; i++)
		{
		if (!strcasecmp(ipc_table[i].ipc_cmd, cmd))
			{
			/* Get the 'node' out if the first argument is the host*/
			if (ipc_table[i].first_is_node)
				{
				if (!(bword = strtok_r(NULL,":",&last)))
					{
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:MALFORMATTED_MESSAGE",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				strncpy(node_id,bword,(sizeof(node_id)-1));
				}
			else
				memset(&node_id[0], 0, sizeof(node_id));

			/* Verify user credentials if the the cmd requires auth */
			if (ipc_table[i].requires_auth)
				{
				if (!ipc->user)
					{ 
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:NEED_AUTH",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				if (!verify_ipc_permissions(ipc->user, "nexxus", cmd))
					{
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:PERMISSION_DENIED",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				}
			/* If node id is set, convert it to a node structure */
			if (node_id[0] !='\0')
				{
				if (!(node = nodegroup_node_lookup(NULL,node_id)))
					{
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:NODE_NOT_FOUND",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				if (!(group = nodegroup_lookup_by_node(node)))
					{
					nexxus_log_printf(NULL,"INTERNAL ERROR in nexxus IPC\n");
					nexxus_log_printf(NULL,"Node %s doesnt have a group\n",node_id);
					nettmcs_free(saved_msg);
					return(0);
					}
				if (!user_group_user_is_member(ipc->user, group->group_name))
					{
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:NODE_NOT_FOUND",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				}
			/* pull out the rest of the args */
			sprintf(tmp_str,"cmd <%s> arg list",cmd);
			arguments = nettmcs_alloc((sizeof(char *) * 65),tmp_str);
			memset(arguments, 0, (sizeof(char *) * 65));
		
			if (ipc_table[i].first_is_node)
				j=1;
			else
				j=0;
			for (; j < ipc_table[i].args; j++)
				{
				char	tmp_arg[255];

				if (!(bword =strtok_r(NULL,":",&last)))
					{
					sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:MALFORMATTED_MESSAGE",job_id);
					send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
					nettmcs_free(saved_msg);
					return(0);
					}
				strncpy(tmp_arg, bword, (sizeof(tmp_arg)-1));
				if (ipc_table[i].first_is_node)
					{
					sprintf(tmp_str,"cmd <%s> arg %d",cmd,(j-1));
					arguments[(j-1)] = (char *)nettmcs_alloc((strlen(tmp_arg))+1,tmp_str);
					strcpy(arguments[(j-1)], tmp_arg);
					}
				else
					{
					sprintf(tmp_str,"cmd <%s> arg %d",cmd,j);
					arguments[j] = (char *) nettmcs_alloc((strlen(tmp_arg))+1,tmp_str);
					strcpy(arguments[j], tmp_arg);
					}
				}
			arguments[j] = NULL;
			/* do callback */
			if ((respmsg = ipc_table[i].function(ipc,
																						ipc->user,
																						node,
																						job_id,
																						arguments)))
				{
				/* send response */
				sprintf(clientmsg, "NEXXUS:%d:%s",job_id,respmsg);
				send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
				}
			for (j=0; arguments[j]!=NULL; j++)
				{
				nettmcs_free(arguments[j]);
				}
			nettmcs_free(arguments);
			nettmcs_free(saved_msg);
			return(0);
			}
		}

	if (ipc_table[i].args == -1)
		{
		sprintf(clientmsg, "NEXXUS:%d:JOB_ERROR:UNSUPPORTED_MESSAGE", job_id);
		send_socket_ipc(ipc, clientmsg, (strlen(clientmsg)+1));
		}
	nettmcs_free(saved_msg);
	return(0);
}

int send_msg_to_all_clients(char *msg, __uint32_t len)
{
	IPC_CONNECTION	*ipc;
	int							count = 0;

	ipc = ipc_head;
	while(ipc)
		{
		if (send_socket_ipc(ipc, msg, len)<0)
			{
			}
		count ++;

		ipc = ipc->next;
		}
	return(count);
}

int send_msg_to_client_by_fd(int I_fd, char *msg, __uint32_t len)
{
	IPC_CONNECTION	*ipc;

	ipc = ipc_head;
	while(ipc)
		{
		if (ipc->I_fd == I_fd)
			{
			send_socket_ipc(ipc, msg, len);
			return(0);
			}
		ipc = ipc->next;
		}
	return(-1);
}

int send_socket_ipc(IPC_CONNECTION *ipc,
									 char						*msg,
									 __uint32_t			len)
{
	__uint32_t	network_len;
#ifdef OPENSSL
	int					err;
#endif

	if (nexxus_debug)
		nexxus_log_printf(NULL,"send_socket_ipc() msg %s to fd %d\n",msg,ipc->I_fd);
	nexxus_do_lock(ipc->I_fd);
	
	network_len = htonl(len);
#ifdef OPENSSL
	if ((err = SSL_write(ipc->ssl, (char *) &network_len, sizeof(network_len)))<0)
		{
	// 	nexxus_log_printf(NULL,"Error while sending SSL length\n");
		destroy_ipc_connection(ipc);
		return(-1);
		}
	if ((err = SSL_write(ipc->ssl, msg, (int) len))<0)
		{
		// nexxus_log_printf(NULL,"Error while sending SSL data\n");
		destroy_ipc_connection(ipc);
		return(-1);
		}
#else
  if (write(ipc->I_fd, &network_len, sizeof(network_len))<0)
    {
	// 	nexxus_log_printf(NULL,"Communications error detected(1). (fd %d) (%m)\n",ipc->I_fd);
		destroy_ipc_connection(ipc);
    return(-1);
    }
  if (write(ipc->I_fd, msg, len)<0)
    {
	// 	nexxus_log_printf(NULL,"Communications error detected(2). (fd %d) (%m)\n",ipc->I_fd);
		destroy_ipc_connection(ipc);
    return(-1);
    }
#endif
	nexxus_do_unlock(ipc->I_fd);
  return(0);
}

static void create_sockets()
{
	struct sockaddr_in	sock_in;
	struct sockaddr_un	sock_un;
	int									I_optval;

	/* Create the remote IPC socket */
  if ((I_inetsocket = socket(AF_INET, SOCK_STREAM,0))<0)
    {
    nexxus_log_printf(NULL,"Unable to create remote IPC socket (%m)\n");
		nexxus_shutdown(-1);
    }
	I_optval =1;
  if (setsockopt(I_inetsocket,
     SOL_SOCKET,
     SO_REUSEADDR,
     &I_optval,
     sizeof(I_optval)) < 0) 
			{
    	nexxus_log_printf(NULL,"setsockopt failure (%m)\n");
			nexxus_shutdown(-1);
  		}
  memset(&sock_in,0,sizeof(sock_in));
  sock_in.sin_family= AF_INET;
  sock_in.sin_addr.s_addr= htonl(INADDR_ANY);
  sock_in.sin_port= htons(6442);
  if (bind(I_inetsocket,
          (struct sockaddr *) &sock_in,
          sizeof(sock_in))<0)
    {
    nexxus_log_printf(NULL,"Unable to bind remote IPC socket (%m)\n");
		nexxus_shutdown(-1);
    }
  if (listen(I_inetsocket, 5)<0)
    {
    nexxus_log_printf(NULL,"Unable to listen on remote IPC socket (%m)\n");
		nexxus_shutdown(-1);
    }

	/* Create the remote ping responder socket */
  if ((I_respondersocket = socket(AF_INET, SOCK_DGRAM,0))<0)
    {
    nexxus_log_printf(NULL,"Unable to create remote responder socket (%m)\n");
		nexxus_shutdown(-1);
    }
	I_optval =1;
  if (setsockopt(I_respondersocket,
     SOL_SOCKET,
     SO_REUSEADDR,
     &I_optval,
     sizeof(I_optval)) < 0) 
			{
    	nexxus_log_printf(NULL,"Unable to set socket options (%m)\n");
			nexxus_shutdown(-1);
  		}
  memset(&sock_in,0,sizeof(sock_in));
  sock_in.sin_family= AF_INET;
  sock_in.sin_addr.s_addr= htonl(INADDR_ANY);
  sock_in.sin_port= htons(6445);
  if (bind(I_respondersocket,
          (struct sockaddr *) &sock_in,
          sizeof(sock_in))<0)
    {
    nexxus_log_printf(NULL,"Unable to bind remote responder socket (%m)\n");
		nexxus_shutdown(-1);
    }
  memset(&sock_un,0,sizeof(struct sockaddr));
  if ((I_loose_doorbell_fd=socket(AF_UNIX,SOCK_STREAM,0))<0)
    {
    nexxus_log_printf(NULL,"Unable to create loose module doorbell socket (%m)\n");
		nexxus_shutdown(-1);
    }
	if (nexxus_debug)
		{
 		nexxus_log_printf(NULL,"Loose module doorbell socket created. (%d)\n",
    			 I_loose_doorbell_fd);
		}
  sock_un.sun_family = AF_UNIX;
  sprintf(sock_un.sun_path,"/tmp/vacm");
  if (bind(I_loose_doorbell_fd,
          (struct sockaddr *)&sock_un,
          sizeof(struct sockaddr))<0)
    {     
    fprintf(stderr,"Unable to bind loose module doorbell socket (%m)\n");
		fprintf(stderr,"Either there is a Nexxus already running, or\n");
		fprintf(stderr,"there is a stale '/tmp/vacm' socket. If Nexxus\n");
		fprintf(stderr,"is not running, please delete this file.\n");
		exit(-1);
    }
  if (listen(I_loose_doorbell_fd,5)<0)
    {
    nexxus_log_printf(NULL,"Unable to listen on loose module doorbell socket (%m)\n");
		nexxus_shutdown(-1);
    }
}

static void nexxus_shutdown(int exit_code)
{
	nexxus_log_printf(NULL,"Commencing shutdown sequence\n");
	shutdown_modules(1);
	destroy_sockets();
#ifdef OPENSSL
	SSL_CTX_free(ctx);
#endif
	if (nexxus_debug)
		nettmcs_dump();
	exit(exit_code);
}

static void destroy_sockets()
{
	close(I_inetsocket);				/* Remote inet IPC socket */
	close(I_respondersocket);	/* Remote network ping responder socket */
	close(I_loose_doorbell_fd);
	unlink("/tmp/vacm");
}

static void signal_handler(int I_signo)
{
	if (!signal_semaphore)
		signal_semaphore ++;
	else
		return;
	if ((I_signo==SIGINT)||(I_signo==SIGKILL)||(I_signo==SIGHUP)||(I_signo==SIGTERM))
		{
		BOOL_SET(nexxus_stdout_logging);
		nexxus_log_printf(NULL,"Caught a signal\n");
		nexxus_shutdown(-(I_signo));
		}
	if (I_signo==SIGUSR1)
		nettmcs_dump();
	signal_semaphore --;
}

static void free_ipc_connection(IPC_CONNECTION *ipc)
{
	if (ipc == ipc_head)
		{
		ipc_head = ipc->next;
		if (ipc_head)
			ipc_head->prev = NULL;
		else
			ipc_tail = NULL;
		}
	else if (ipc == ipc_tail)
		{
		ipc_tail = ipc->prev;
		if (ipc_tail)
			ipc_tail->next = NULL;
		else
			ipc_head = NULL;
		}
	else
		{
		if (ipc->prev)
			ipc->prev->next = ipc->next;
		if (ipc->next)
			ipc->next->prev = ipc->prev;
		}
	nettmcs_free(ipc);
}

void destroy_ipc_connection(IPC_CONNECTION *ipc)
{
#if 1 // BUG 168 fix
	nexxus_log_printf(NULL,
												"Logging user %s out on client %d from %s\n",
												((ipc->user) ? ipc->user->user_name : "unknown"),
												ipc->I_fd,
												inet_ntoa(ipc->conn_address));
#endif
#ifdef OPENSSL
	SSL_free(ipc->ssl);
#endif
	client_fd_lock_destroy(ipc->I_fd);
	close(ipc->I_fd);
	free_ipc_connection(ipc);
	return;
}

static IPC_CONNECTION *alloc_and_accept_ipc_connection(int I_socket)
{
	IPC_CONNECTION			*ipc_new;
	struct sockaddr_in	addr_in;
	int									len;
	char								readymsg[64];
#ifdef OPENSSL
	int									err;
	char								*str;
#endif

	if (!(ipc_new = (IPC_CONNECTION *) nettmcs_alloc (sizeof(IPC_CONNECTION),
																										"ipc structure")))
		{
		nexxus_log_printf(NULL,"Unable to allocate memory for ipc structure\n");
		return(NULL);
		}
	memset(ipc_new, 0, sizeof(IPC_CONNECTION));

  /* Initialize our idle_timer */
  ipc_new->idle_timer = time(NULL);

	if (!ipc_head)
		ipc_head = ipc_tail = ipc_new;
	else
		{
		ipc_tail->next = ipc_new;
		ipc_new->prev = ipc_tail;
		ipc_tail = ipc_new;
		}

	len = sizeof(struct sockaddr_in);
	if ((ipc_new->I_fd = accept(I_socket, 
															(struct sockaddr *)&addr_in,
															&len))<0)
		{
		nexxus_log_printf(NULL,
													"Unable to accept remote client connection (%m)\n");
		free_ipc_connection(ipc_new);
		return(NULL);
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL,"Client connection accepted (fd %d)\n",ipc_new->I_fd);
	memcpy(&ipc_new->conn_address, &addr_in.sin_addr, sizeof(struct in_addr));
	client_fd_lock_create(ipc_new->I_fd);

#ifdef OPENSSL
	if (nexxus_debug)
		nexxus_log_printf(NULL,"Creating SSL context for connection\n");
	ipc_new->ssl = SSL_new(ctx);
	if (!ipc_new->ssl)
		{
		nexxus_log_printf(NULL,"SSL context creation failed\n");
		goto ssl_skip;
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL,"Binding context to fd %d\n",ipc_new->I_fd);
	SSL_set_fd(ipc_new->ssl, ipc_new->I_fd);
	err = SSL_accept(ipc_new->ssl);
	if (err <0)
		{
		nexxus_log_printf(NULL,
											"SSL_accept() failed. (%s)\n",
											ERR_error_string(err, NULL));
		nexxus_do_unlock(ipc_new->I_fd);
		destroy_ipc_connection(ipc_new);
		return(NULL);
		}
	if (!strcmp(SSL_get_cipher(ipc_new->ssl), "(NONE)"))
		{
		__uint32_t	network_len;

		nexxus_log_printf(NULL,
													"Non-SSL client detected. Disconnecting.\n");
		SSL_free(ipc_new->ssl);
		strcpy(readymsg, "NEXXUS_REQUIRES_ENCRYPTION");
		network_len = htonl((strlen(readymsg)+1));
		nexxus_do_lock(ipc_new->I_fd);
 		if (write(ipc_new->I_fd, &network_len, sizeof(network_len))<0)
			{
			}
 		if (write(ipc_new->I_fd, readymsg, len)<0)
			{
			}
		nexxus_do_unlock(ipc_new->I_fd);
		destroy_ipc_connection(ipc_new);
		return(NULL);
		}
	if (nexxus_debug)
		{
		nexxus_log_printf(NULL,
													"SSL connection using %s\n",
													SSL_get_cipher(ipc_new->ssl));
		}
	if (!(ipc_new->client_cert = SSL_get_peer_certificate(ipc_new->ssl)))
		{
		if (nexxus_debug)
			nexxus_log_printf(NULL,"Note: Client does not have a certificate\n");
		}
	else
		{
		nexxus_log_printf(NULL,"Client certificate:\n");
		if (!(str = X509_NAME_oneline(X509_get_subject_name(ipc_new->client_cert),
																	0,
																	0)))
			{
			nexxus_log_printf(NULL,"Client certificate contains no subject\n");
			SSL_free(ipc_new->ssl);
			goto ssl_skip;
			}
		nexxus_log_printf(NULL,"\t subject: %s\n",str);
		free(str);
		if (!(str = X509_NAME_oneline(X509_get_issuer_name(ipc_new->client_cert),
																	0,
																	0)))
			{
			nexxus_log_printf(NULL,"Client certificate contains no issuer\n");
			SSL_free(ipc_new->ssl);
			goto ssl_skip;
			}
		nexxus_log_printf(NULL,"\t issuer: %s\n",str);
		free(str);
		X509_free(ipc_new->client_cert);
		}
ssl_skip:
#endif
	strcpy(readymsg, "NEXXUS_READY");
	if (send_socket_ipc(ipc_new,&readymsg[0],(strlen(readymsg)+1))<0)
		{
		return(NULL);
		}

	return(ipc_new);
}

void *nexxus_fd_watch(int		fd,
											void	(*cb) (int, void *),
											void	*arg)
{
	NEXXUS_FD_WATCHER	*new;
	char							tmp_str[255];

	sprintf(tmp_str,"nexxus fd %d watcher",fd);
	if (!(new = (NEXXUS_FD_WATCHER *) nettmcs_alloc (sizeof(NEXXUS_FD_WATCHER),
																									 tmp_str)))
		{
		nexxus_log_printf(NULL,"Unable to allocate memory for Nexxus FD watcher\n");
		return(NULL);
		}
	memset(new, 0, sizeof(NEXXUS_FD_WATCHER));
	
	if (!watcher_tail)
		watcher_tail = watcher_head = new;
	else
		{
		watcher_tail->next = new;
		new->prev = watcher_tail;
		}
	watcher_tail = new;
	new->cb = cb;
	new->fd = fd;
	new->arg = arg;
	return((void *) new);
}

void nexxus_fd_unwatch(int fd)
{
	NEXXUS_FD_WATCHER *ptr = watcher_head;
	
	while(ptr)
		{
		if (ptr->fd == fd)
			{
			ptr->fd = -1;
			return;
			}
		ptr = ptr->next;
		}
	nexxus_log_printf(NULL,"Warning! Attempt to unwatch an fd we weren't watching!\n");
}

static void nexxus_fd_watch_destroy(NEXXUS_FD_WATCHER *ptr)
{
	if (ptr == watcher_head)
		{
		watcher_head = watcher_head->next;
		if (watcher_head)
			watcher_head->prev = NULL;
		else
			watcher_tail = NULL;
		}
	else if (ptr == watcher_tail)
		{
		watcher_tail = watcher_tail->prev;
		if (watcher_tail)
			watcher_tail->next = NULL;
		else
			watcher_head = NULL;
		}
	else
		{
		ptr->next->prev = ptr->prev;
		ptr->prev->next = ptr->next;
		}
	nettmcs_free(ptr);
}

#ifdef OPENSSL
SSL	*lookup_ssl_by_fd(int fd)
{
	IPC_CONNECTION	*ipc;

	ipc = ipc_head;
	while(ipc)
		{
		if (ipc->I_fd == fd)
			return(ipc->ssl);
		ipc = ipc->next;
		}
	return(NULL);
}
#endif

int nexxus_log_printf(char *module, char *fmt, ...)
{
  va_list ap;
  int     rc;

  va_start(ap, fmt);
  rc = do_nexxus_log_printf(module, fmt, ap);
  va_end(ap);
  return(rc);
}

