/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <pthread.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <emp_module_harness.h>
#include <emp_slave_context.h>
#include <prototypes.h>
#include <emp_slave_context.h>
#include <pthread.h>
#include <emp.h>
#include <emp_sdr.h>
#include <emp_sel.h>
#include <emp_sensor_events.h>

//#define SENSOR_CONNECT_DEBUG
//#define SENSOR_DEBUG

static void sensor_thread_cleanup(void *);
static int update_sensor(CONTEXT *ctx, SENSOR *sensor, int id_mode, int flg);
static int extentd_connect(CONTEXT *ctx);

typedef struct sensord_req
  {
  unsigned char sensor_number;
  unsigned char identify_cmd;
  } EXTENTD_REQ;

typedef struct sensord_21_rep
	{
  __uint32_t    rc;
  __uint32_t    datalen;
  unsigned char buff[12];
	} EXTENTD_21_REP; /* length 20 bytes */

typedef struct sensord_31_rep
  {
  __uint32_t    rc;
  __uint32_t    datalen;
	unsigned char	cc;
  unsigned char buff[4]; /* length */
  } EXTENTD_31_REP;

typedef struct sensord_40_rep
  {
	char					key;
	char					ext_major_ver;
	char					ext_minor_ver;
	char					drv_major_ver;
	char					drv_minor_ver;
  __uint32_t    rc;
  __uint32_t    datalen;
	unsigned char	cc;
  unsigned char buff[4];
  } EXTENTD_40_REP;


/*
 * Sensor polling thread
 */
void *sensor_thread_start(void *data)
{
	CONTEXT *ctx = (CONTEXT *) 			data;
	struct _pthread_cleanup_buffer 	_buff;
	SENSOR													*sensor;
	struct timeval									to;
	fd_set													read_fds;
	int															rc, max;
	char														log_msg[255];

#ifdef THREAD_DEBUG
  sprintf(log_msg,
          "0:0:LOG:Thread %s driver starting (pid %d)",
          ctx->node_id,
          getpid());
  send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	ctx->refcount++;
	_pthread_cleanup_push(&_buff, sensor_thread_cleanup, (void *) ctx);

	while(1)
		{
		if (!ctx->negotiation_complete)
			sleep(5);
		else
			break;
		pthread_testcancel();
		}
	
	while((rc = extentd_connect(ctx)) != 0)
		{
#ifdef SENSOR_CONNECT_DEBUG
		printf("[%s] Connection to extentd rc = %d\n",ctx->node_id,rc);
#endif
		sleep(5);
		pthread_testcancel();
		}
#ifdef SENSOR_CONNECT_DEBUG
	printf("[%s] Connection to extentd rc = %d\n",ctx->node_id,rc);
#endif

	while(1)
		{
		pthread_testcancel();
		FD_ZERO(&read_fds);
		max = 0;
		if (ctx->sensor_ipc_read != -1)
			{
			FD_SET(ctx->sensor_ipc_read, &read_fds);
			max = ctx->sensor_ipc_read;
			}
		to.tv_sec = 5;
		to.tv_usec = 0;
		if ((rc=select((max+1), &read_fds, NULL, NULL, &to))<0)
			{
			sprintf(log_msg,"0:0:LOG:Node %s sensor thread select failure",
							ctx->node_id);
    	send_msg(ctx,ctx->thread_ipc_write, log_msg);
			continue;
			}
		if ((rc) && (FD_ISSET(ctx->sensor_ipc_read, &read_fds)))
			{
			char				*msg_buffer;
			__uint32_t	len;

			if ((rc = recv_msg(ctx->sensor_ipc_read, &msg_buffer, &len))<0)
				{
				sprintf(log_msg,"0:0:LOG:Node %s sensor thread recv_msg failure",
								ctx->node_id);
    		send_msg(ctx,ctx->thread_ipc_write, log_msg);
				continue;
				}
			if (msg_buffer[3] == '0')
				update_sensor(ctx, NULL, 1, 0);
			else
				update_sensor(ctx, NULL, 1, 1);
			free(msg_buffer);
			}
		if (!ctx->negotiation_complete)
			continue;

		sensor = ctx->sensor_list;
		while(sensor)
			{
			if ((sensor->next_poll <= time(NULL))||(sensor->next_poll == 0))
				{
				if (lookup_sensor_number(sensor) < 0)
					{
					sensor = sensor->next;
					continue;
					}
				update_sensor(ctx, sensor, 0, 0);
				sensor->next_poll = time(NULL)+10;
				}
			sensor = sensor->next;
			}
		}
}

static int extentd_connect(CONTEXT *ctx)
{
	struct sockaddr_in	sockaddr;
	char								log_msg[255];

	if (!ctx->ip_address)
		{
		ctx->extentd_connect_status = ENOPKG;
		return(-1);
		}
	if (ctx->emp_extentd_fd != -1)
		{
		close(ctx->emp_extentd_fd);
		ctx->emp_extentd_fd = -1;
		}
	if ((ctx->emp_extentd_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
#ifdef SENSOR_CONNECT_DEBUG
		printf("[%s] Unable to create socket (%m)\n",ctx->node_id);
#endif
    sprintf(log_msg,
            "0:0:LOG:Thread %s unable to create emp extentd socket (%m)",
            ctx->node_id);
    send_msg(ctx,ctx->thread_ipc_write, log_msg);
    ctx->emp_extentd_fd = -1;
    return(-1);
    }
  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = inet_addr(ctx->ip_address);
  sockaddr.sin_port = htons(9352);
  if (connect(ctx->emp_extentd_fd,
              (struct sockaddr *) &sockaddr,
              sizeof(struct sockaddr)) <0)
    {
#ifdef SENSOR_CONNECT_DEBUG
		printf("[%s] Unable to connect (%m)\n",ctx->node_id);
#endif
    close(ctx->emp_extentd_fd);
    ctx->emp_extentd_fd = -1;
		ctx->extentd_connect_status = errno;
		return(-1);
    }
	ctx->extentd_connect_status = 0;
	return(0);
}

static int update_sensor(CONTEXT *ctx, SENSOR *sensor, int id_mode, int flag)
{
	struct timeval	to;
	int							rc;
	fd_set					read_fds;
	EXTENTD_REQ			request_pkt;
	char						response_pkt[64];
	int							retries = 0;
	unsigned int		rep_rc, rep_dlen;
	unsigned char		rep_cc, rep_data[16];

	pthread_mutex_lock(&ctx->extentd_lock);
	while(1)
		{
		memset(&request_pkt, 0, sizeof(request_pkt));
		if (!id_mode)
			{
  		request_pkt.identify_cmd = 0;
  		request_pkt.sensor_number = lookup_sensor_number(sensor);
			}
		else
			{
  		request_pkt.identify_cmd = 1;
  		request_pkt.sensor_number = flag;
			}

#ifdef SENSOR_DEBUG
		if (!id_mode)
			printf("[%-7s] Sensor req %.2d\n",ctx->node_id,lookup_sensor_number(sensor));
		else
			printf("[%-7s] ID req %d\n",ctx->node_id,flag);
#endif
		if (write(ctx->emp_extentd_fd,
							(void *) &request_pkt,
							sizeof(request_pkt)) < 0)
			{
			retries++;
			if (retries < 5)
				{
#ifdef SENSOR_DEBUG
				printf("[%-7s] Reconnecting..\n",ctx->node_id);
#endif
				if (extentd_connect(ctx) <0)
					{
					pthread_mutex_unlock(&ctx->extentd_lock);
#ifdef SENSOR_DEBUG
					printf("[%-7s] Reconnect error\n",ctx->node_id);
#endif
					if (!id_mode)
						sensor->rc = ENODATA;
					return(-errno);
					}
				continue;
				}
			pthread_mutex_unlock(&ctx->extentd_lock);
#ifdef SENSOR_DEBUG
			printf("[%-7s] Reconnect times exceeded\n",ctx->node_id);
#endif
			if (!id_mode)
				sensor->rc = ENODATA;
			return(-errno);
			}

		FD_ZERO(&read_fds);
		to.tv_sec = 3;
		to.tv_usec = 0;
		FD_SET(ctx->emp_extentd_fd, &read_fds);
		if ((rc = select((ctx->emp_extentd_fd+1), &read_fds, NULL, NULL, &to))<0)
			{
			printf("Select failure (%m)\n");
			pthread_mutex_unlock(&ctx->extentd_lock);
#ifdef SENSOR_DEBUG
			printf("[%-7s] select failure (%m)\n",ctx->node_id);
#endif
			if (!id_mode)
				sensor->rc = ENODATA;
			return(-1);
			}
		if (!rc)
			{
#ifdef SENSOR_DEBUG
			printf("[%-7s] select timeout\n",ctx->node_id);
#endif
			if (!id_mode)
				sensor->rc = ENODATA;
			continue;
			}
		if (FD_ISSET(ctx->emp_extentd_fd, &read_fds))
			{
			if ((rc = read(ctx->emp_extentd_fd,
							(void *) &response_pkt[0],
							sizeof(response_pkt))) <0)
				{
				pthread_mutex_unlock(&ctx->extentd_lock);
#ifdef SENSOR_DEBUG
				printf("[%-7s] read failed (%m)\n",ctx->node_id);
#endif
				if (!id_mode)
					sensor->rc = ENODATA;
				return(-1);
      	}
			if (rc == sizeof(EXTENTD_21_REP))
				{
				EXTENTD_21_REP	*rep21 = (EXTENTD_21_REP *) &response_pkt[0];

				rep21->datalen = ntohl(rep21->datalen);
				rep21->rc = ntohl(rep21->rc);
#ifdef SENSOR_DEBUG
				printf("[%-7s] Extentd Reply format 2.1, rc %d, datalen %d\n",ctx->node_id,rep21->rc,rep21->datalen);
#endif
				rep_rc = rep21->rc;
				rep_dlen = rep21->datalen;
				memcpy(&rep_data, &rep21->buff, sizeof(rep21->buff));

				/* We dont like 2.1 packets anymore.. toss it */
				if (sensor == ctx->sensor_list)
					{
					printf("[%-7s] Old version (2.1) emp_extentd detected on node! Please replace!\n",ctx->node_id);
					}
				sensor->rc = ENODATA;
				pthread_mutex_unlock(&ctx->extentd_lock);
				return(-1);
				}
			else if (rc == sizeof(EXTENTD_31_REP))
				{
				EXTENTD_31_REP	*rep31 = (EXTENTD_31_REP *) &response_pkt[0];

				rep31->datalen = ntohl(rep31->datalen);
				rep31->rc = ntohl(rep31->rc);
#ifdef SENSOR_DEBUG
				printf("[%-7s] Extentd Reply format 3.1, rc %d, datalen %d, cc 0x%.2x\n",ctx->node_id,rep31->rc,rep31->datalen,rep31->cc);
#endif
				rep_rc = rep31->rc;
				rep_cc = rep31->cc;
				rep_dlen = rep31->datalen;
				memcpy(&rep_data, &rep31->buff, sizeof(rep31->buff));
				if (sensor == ctx->sensor_list)
					{
					printf("[%-7s] Old version (3.1) emp_extentd detected on node! Please replace!\n",ctx->node_id);
					}
				}
			else if (rc == sizeof(EXTENTD_40_REP))
				{
				EXTENTD_40_REP	*rep40 = (EXTENTD_40_REP *) &response_pkt[0];

				rep40->datalen = ntohl(rep40->datalen);
				rep40->rc = ntohl(rep40->rc);
#ifdef SENSOR_DEBUG
				printf("[%-7s] Extentd Reply. Pkt format %d.%d, Dvr ver %d.%d rc %d, dlen %d, cc 0x%.2x\n",
								ctx->node_id,
								rep40->ext_major_ver,
								rep40->ext_minor_ver,
								rep40->dvr_major_ver,
								rep40->dvr_minor_ver,
								rep40->rc,
								rep40->datalen,
								rep40->cc);
#endif
				rep_rc = rep40->rc;
				rep_cc = rep40->cc;
				rep_dlen = rep40->datalen;
				memcpy(&rep_data, &rep40->buff, sizeof(rep40->buff));
				}
			else
				{
				printf("[%-7s] Unknown or BAD extentd reply from this node... Check node side software components!!!!\n",ctx->node_id);
				sensor->rc = ENODATA;
				pthread_mutex_unlock(&ctx->extentd_lock);
				return(-1);
				}
			if (id_mode)
				{
				pthread_mutex_unlock(&ctx->extentd_lock);
				return(0);
				}
			if (rep_cc != 0x00)
				{
				sensor->rc = ENODATA;
				if ((rep_cc != 0xce) && (rep_cc != 0xc1))
					printf("Got CC 0x%.2x from sensor request\n",rep_cc);
				pthread_mutex_unlock(&ctx->extentd_lock);
				return(-1);
				}
			if (rep_dlen == 0)
				{
				printf("Got 0 length data field on sensor reply 8(\n");
				pthread_mutex_unlock(&ctx->extentd_lock);
				sensor->rc = ENODATA;
				return(-1);
				}
			if (rep_dlen > 4)
				{
				printf("Sensor response datalength invalid (%d)\n",
								rep_dlen);
				sensor->rc = ENODATA;
				pthread_mutex_unlock(&ctx->extentd_lock);
				return(-1);
				}
			if ((int) rep_rc != 0)
				sensor->rc = ENODATA;
			else
				sensor->rc = 0;
			memcpy(&sensor->cached_value,&rep_data,rep_dlen);
			pthread_mutex_unlock(&ctx->extentd_lock);
			return(0);
			}
		else
			{
			printf("[%s] Select triggered for unknown entity!\n",ctx->node_id);
			sensor->rc = ENODATA;
			pthread_mutex_unlock(&ctx->extentd_lock);
			return(-1);
			}
		}
}

static void sensor_thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	ctx->refcount--;
}
