
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <pthread.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <emp_module_harness.h>
#include <emp_slave_context.h>
#include <prototypes.h>
#include <emp_slave_context.h>
#include <pthread.h>
#include <iana_list.h>
#include <emp.h>
#include <emp_sdr.h>
#include <emp_sel.h>
#include <emp_sensor_events.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static void	process_emp_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static void	set_serial_port(CONTEXT *, char *);
static void	set_password(CONTEXT *, char *);
static void set_ipaddress(CONTEXT *ctx, char *ip);
static int	emp_engage(CONTEXT *);
static void	emp_disengage(CONTEXT *);
static int	thread_detect_protocol(CONTEXT *);
static int	do_emp_transaction(CONTEXT			*,
											 				unsigned char,
											 				unsigned char,
											 				unsigned char,
											 				unsigned char,
											 				unsigned char,
											 				unsigned char *,
											 				int,
															unsigned char	**,
															int *);
static int do_emp_submitpass(CONTEXT *);
static void validate_transaction(CONTEXT				*ctx,
																 __uint32_t			job_id,
																 int						client_fd,
																 int						rc,
																 unsigned char	*response);
static int do_emp_downloadfru(CONTEXT *ctx);
static int do_emp_downloadsdr(CONTEXT *ctx);
static int do_emp_downloadsel(CONTEXT *ctx);
static void do_negotiation(CONTEXT *ctx);
static void parse_fru(CONTEXT *ctx);
static void decode_string(unsigned char type,
                   				unsigned char language_code,
                   				unsigned char *source,
                   				char					*target,
                   				int						size);
static int do_emp_set_at(CONTEXT *ctx, char *tag);
static void process_sdr_record(CONTEXT *ctx, unsigned char *buffer);
int return_device_type_entry(unsigned char type_code,
                             unsigned char type_modifier,
                             char          *buffer);
double linearize_sensor_reading(SENSOR *slist, unsigned char reading);
static int return_num_processed_sensors(CONTEXT *ctx);
static void sensor_free(CONTEXT *ctx, SENSOR *sensor);
static SENSOR	*sensor_alloc(CONTEXT *ctx);
static void check_timers(CONTEXT *ctx);
static int do_emp_skipreinit(CONTEXT *ctx);
static void fru_free(CONTEXT *ctx);
static int do_emp_getdeviceid(CONTEXT *ctx);
static int do_emp_getempconfiguration(CONTEXT *ctx);
static int do_emp_restart(CONTEXT			*ctx,
													__uint32_t	job_id,
													int					client_fd);
static char *parse_sel_entry(CONTEXT *ctx, SEL_RESP *entry);
char *sensor_get_canned_event_desc(unsigned char sensor_type_code,
                            unsigned char sensor_specific_offset);
char *sensor_get_type(unsigned char sensor_type_code);
char *sensor_get_trigger_description(unsigned char sensor_reading_type_code,
                                   unsigned char offset);
int sensor_get_class(unsigned char sensor_reading_type_code);
char *sensor_get_unit_type(unsigned char sensor_unit_type_code);
static void process_msg_from_driver(CONTEXT *ctx);
static void close_sel(FILE	*Tfilep_ptr);
static FILE *open_sel(CONTEXT	*ctx);
void sensor_str_convert(CONTEXT				*ctx,
                        unsigned char	sensor_number,
                        unsigned char	*data,
                        char					*string_msg);
static int do_emp_rearm(CONTEXT *ctx);
static int do_emp_getchassiscapabilities(CONTEXT *ctx);
static char *iana_lookup(unsigned int iana_number);


#define EMP_HIGH_LEVEL_RETRY_DELAY 250000
#define EMP_HIGH_LEVEL_RETRY_MAX 32

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

/*
#define EMPHW_DEBUG
#define FRU_DEBUG
#define SEL_DEBUG
#define SDR_DEBUG
#define ENGAGE_DEBUG
#define SENSOR_DEBUG
#define ALL_DEBUG
#define SEL_DEBUG
#define REARM_DEBUG
#define DETECT_DEBUG
#define TRANS_DEBUG	
#define THREAD_DEBUG
#define NEGO_DEBUG
*/
#ifdef ALL_DEBUG
  #define DRIVER_DEBUG
  #define THREAD_DEBUG
	#define EMPHW_DEBUG
#endif

#ifdef THREAD_DEBUG
	#define EMPHW_DEBUG
//  #define SEL_DEBUG
//  #define NEGO_DEBUG
	#define ENGAGE_DEBUG
  #define SENSOR_DEBUG
#endif

#ifdef NEGO_DEBUG
  #define SDR_DEBUG
	#define FRU_DEBUG
	#define SEL_DEBUG
#endif

#ifdef DRIVER_DEBUG
  #define DETECT_DEBUG
  #define TRANS_DEBUG
  #define REARM_DEBUG
#endif

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
	pthread_cancel(ctx->driver_thread);
	pthread_cancel(ctx->sensor_thread);
	emp_disengage(ctx);
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];
	int															pipefd1[2];
	int															pipefd2[2];
	pthread_mutex_t									tmp_mutex = PTHREAD_MUTEX_INITIALIZER;

	memcpy(&ctx->ipc_writelock, &tmp_mutex, sizeof(pthread_mutex_t));
	pthread_mutex_init(&ctx->ipc_writelock, NULL);

	memcpy(&ctx->extentd_lock, &tmp_mutex, sizeof(pthread_mutex_t));
	pthread_mutex_init(&ctx->extentd_lock, NULL);

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;
	if ((pipe(pipefd1) < 0) || (pipe(pipefd2) < 0))
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to create driver pipe (%m)",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		ctx->driver_pwrite = -1;
		ctx->thread_pread = -1;
		ctx->driver_pread = -1;
		ctx->thread_pwrite = -1;
		}
	else
		{
		ctx->driver_pwrite = pipefd1[1];
		ctx->thread_pread = pipefd1[0];
		ctx->thread_pwrite = pipefd2[1];
		ctx->driver_pread = pipefd2[0];
		}
	if (pthread_create(&ctx->driver_thread,
										 NULL,
										 driver_thread_start,
										 (void *) ctx) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to create driver thread",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		}

	if (pipe(pipefd1) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to create sensor pipe (%m)",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		ctx->sensor_ipc_write = -1;
		ctx->sensor_ipc_read = -1;
		}
	else
		{
		ctx->sensor_ipc_write = pipefd1[1];
		ctx->sensor_ipc_read = pipefd1[0];
		}
	if (pthread_create(&ctx->sensor_thread,
										 NULL,
										 sensor_thread_start,
										 (void *) ctx) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to create sensor thread",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		}
	ctx->next_seldl = ctx->next_rearm = 0;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		FD_SET(ctx->thread_pread, &read_fds);
		if (ctx->thread_ipc_read > ctx->thread_pread)
			max = ctx->thread_ipc_read;
		else
			max = ctx->thread_pread;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			check_timers(ctx);
			}
		else if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		else if (FD_ISSET(ctx->thread_pread, &read_fds))
			{
			process_msg_from_driver(ctx);
			}
		pthread_testcancel();
		check_timers(ctx);
		}
}

static void process_msg_from_driver(CONTEXT *ctx)
{
	DRIVER_IPC	msg;
	int					rc;
	char				log_msg[255];

	if ((rc = read(ctx->thread_pread, &msg, sizeof(DRIVER_IPC))) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from driver (%s)",
					 ctx->node_id,
					 strerror(-rc));
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return;
		}
	switch(msg.request)
		{
		case DIPCREQ_NEG:
#ifdef NEGO_DEBUG
			sprintf(log_msg,
							"0:0:LOG:Thread %s calling do_negotiation on behalf of driver",
					 		ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
			do_negotiation(ctx);
			break;
		case DIPCREQ_CON:
			sprintf(log_msg,
							"0:0:LOG:Thread %s calling submitpass on behalf of driver",
							 ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			do_emp_submitpass(ctx);
			break;
		case DIPCREQ_CB:
			break;
		case DIPCREQ_RESP:
			printf("Thread %s response pkt received in main DIPCREQ handler\n",
							ctx->node_id);
			printf("Return code -> %d\n",msg.rc);
			printf("cmd         -> 0x%.2x\n",msg.cmd);
			break;
		default:
			printf("Unknown msg from driver (%d)\n",msg.request);
			break;
		}
}

static void check_timers(CONTEXT *ctx)
{
	time_t	current_time;
	int			rc;

	if (!ctx->session_active)
		return;
	if (ctx->console_mode)
		return;
	if (!ctx->protocol_detected)
		return;
	current_time = time(NULL);

	if ((ctx->capabilities.has_sel)&&(!ctx->bios_cfg.restricted_mode))
		{
		if (current_time >= ctx->next_seldl)
			{
			rc = do_emp_downloadsel(ctx);
			ctx->next_seldl = time(NULL) + 5;
			}
		}
	if (current_time >= ctx->next_rearm)
		{
		rc = do_emp_rearm(ctx);
		ctx->next_rearm = time(NULL) + 30;
		}
	if (current_time >= ctx->next_cstatus)
		{
		/* XXX: Do the next chassis status call here */
		ctx->next_cstatus = time(NULL) + 5;
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	char				*new_name;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		pthread_cancel(ctx->driver_thread);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",
				 ctx->node_id,
				 msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_emp_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;
		if (!(value = strtok_r(NULL, ":", &last)))
			goto out;
		if (!strcasecmp(variable,"SERIAL_PORT"))
			{
			int	drc;

			set_serial_port(ctx, value);
			ctx->enable_driver_checks = 0;
			emp_disengage(ctx);
			if (!strcasecmp(value, "NONE"))
				{
#if 0
				sprintf(log_msg,
								"0:0:LOG:Thread %s disengaged",
								ctx->node_id);
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
				}
			else
				{
				if ((rc = emp_engage(ctx)) < 0)
					{
#ifdef ENGAGE_DEBUG
					sprintf(log_msg,
									"0:0:LOG:Thread %s unable to engaging on %s",
									ctx->node_id,
								ctx->device_address);
					send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
					return;
					}
#ifdef DETECT_DEBUG
				printf("Thread %s calling detect_protocol as part of engage process\n",ctx->node_id);
#endif
				drc = thread_detect_protocol(ctx);
				ctx->enable_driver_checks = 1;

				if (!drc)
					do_negotiation(ctx);
				else if (drc == -EUNATCH)
					ctx->console_mode = 1;
				}
			}
		else if (!strcasecmp(variable,"PASSWORD"))
			{
			set_password(ctx, value);
			}
		else if (!strcasecmp(variable, "IP_ADDRESS"))
			{
			set_ipaddress(ctx, value);
			}
//		printf("Setting node %s variable %s to %s\n",node_id,variable,value);
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char	*fd;
		int		client_fd, I_c1;

		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
		for (I_c1 = 0; I_c1 < MAX_SEL_WATCHERS; I_c1++)
			{
			if (ctx->sel_watchers[I_c1] == client_fd)
				{
				ctx->sel_watchers[I_c1] = -1;
				break;
				}
			}
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	logfile_old[255];
		char	logfile_new[255];

		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
		/* Rename the logfile */
		sprintf(logfile_old,
						"%s/vacm_emp_sel_log.%s",
						LOGDIR,
						ctx->node_id);
		free(ctx->node_id);
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		sprintf(logfile_new,
						"%s/vacm_emp_sel_log.%s",
						LOGDIR,
						ctx->node_id);
		if (rename(logfile_old, logfile_new) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s unable to rename logfile (%m)",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			}
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Automatically do the auth, FRU, SDR, and SEL download
 */
static void do_negotiation(CONTEXT *ctx)
{
	int		rc;
	char	log_msg[255];

	// TEST TEST XXX:
ctx->console_switch_allow = 0;
	if (ctx->console_mode)
		{
		sprintf(log_msg,
					"0:0:LOG:Thread %s autonegotiation aborted. Console mode enabled.",
				 	 	ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return;
		}
	ctx->negotiation_complete = 0;
	if ((rc = do_emp_submitpass(ctx)) < 0)
		{
		if (rc == -EACCES)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s autonegotiation connect failed. Bad EMP password",
					 	 	ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			}
		else if (rc == -EUNATCH)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s autonegotiation connect failed. Console mode detected",
					 	 	ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			}
		return;
		}

	if (do_emp_getdeviceid(ctx) == -EUNATCH)
		return;
	if (do_emp_getchassiscapabilities(ctx) == -EUNATCH)
		return;
	if (do_emp_getempconfiguration(ctx) == -EUNATCH)
		return;

	if (ctx->bios_cfg.restricted_mode)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s hardware in restricted mode. ChassisCtrl & SEL Ops disabled",
					 	ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	if (!ctx->bios_cfg.direct_connect)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s hardware in MODEM mode. Attempting to fake a modem",
					 	ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	if (ctx->bios_cfg.activation_mode != ALWAYS_ACTIVE)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s hardware not ALWAYS_ACTIVE",
					 	ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	if (ctx->capabilities.has_fru)
		{
#ifdef NEGO_DEBUG
		printf("Thread %s downloading FRU\n",ctx->node_id);
#endif
		if (do_emp_downloadfru(ctx) == -EUNATCH)
			return;
		}
	if (ctx->capabilities.has_sdr)
		{
#ifdef NEGO_DEBUG
		printf("Thread %s downloading SDR\n",ctx->node_id);
#endif
		if (do_emp_downloadsdr(ctx) == -EUNATCH)
			return;
		}
	if ((ctx->capabilities.has_sel) && (!ctx->bios_cfg.restricted_mode))
		{
#ifdef NEGO_DEBUG
		printf("Thread %s downloading SEL\n",ctx->node_id);
#endif
		if ((rc = do_emp_downloadsel(ctx)) == -EUNATCH)
			return;
		}
#ifdef NEGO_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s negotiation complete",
					ctx->node_id);
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
	ctx->negotiation_complete = 1;
}

/*
 * Engage the serial port
 * Wait for protocol.
 */
static int emp_engage(CONTEXT *ctx)
{
	pthread_mutexattr_t		mutex_attr;
	struct termios				tty;
	char									*last = (char *) 0x00;
	char									log_msg[255];
#ifndef __sgi
	struct serial_struct	ser;
#endif

	if (ctx->emp_sock != -1)
		return(-EADDRINUSE);
		
	if (!ctx->device_address)
		return(-EFAULT);

	if (ctx->device_address[0] == '/')
		{
#ifdef ENGAGE_DEBUG
		sprintf(log_msg,
						"0:0:LOG:Thread %s engaging on serial port %s",
						ctx->node_id,
						ctx->device_address);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
		if ((ctx->emp_sock = open(ctx->device_address, O_RDWR)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
    tcgetattr(ctx->emp_sock,&tty);
    cfmakeraw(&tty);
    cfsetospeed(&tty,B19200);
    cfsetispeed(&tty,B19200);
    tcsetattr(ctx->emp_sock,TCSANOW,&tty);
#ifndef __sgi
    if (ioctl(ctx->emp_sock,TIOCGSERIAL,&ser)==0)
      {
      if (!(ser.flags & ASYNC_LOW_LATENCY))
        {
        ser.flags |= ASYNC_LOW_LATENCY;
        ioctl(ctx->emp_sock,TIOCSSERIAL,&ser);
        }
      } 
#endif  /*__sgi*/
		}
	else
		{
    struct sockaddr_in  sock;
    int                 I_optval;
    char                address[64];
    unsigned int        port;
    char                *bword;
    char                tmp[64];
    
    strcpy(tmp, ctx->device_address);
    
    if (!(bword = strtok_r(tmp, "-", &last)))
      return(-EFAULT);
    strcpy(address, bword);

    if (!(bword = strtok_r(NULL, "-", &last)))
			return(-EFAULT);

    port = atoi(bword);
#ifdef ENGAGE_DEBUG
		sprintf(log_msg,
						"0:0:LOG:Thread %s engaging on inet addr %s port %d",
						ctx->node_id,
						address,
						port);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
		if ((ctx->emp_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
      return(-errno);
			}
    I_optval = 1;
    if (setsockopt(ctx->emp_sock,
                   SOL_SOCKET,
                   SO_REUSEADDR,
                   &I_optval,
                   sizeof(I_optval)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-errno);
			}

    memset(&sock, 0, sizeof(sock));
    sock.sin_family = AF_INET;
    sock.sin_addr.s_addr = inet_addr(address);
    sock.sin_port = htons(port);
    if (connect(ctx->emp_sock,
                (struct sockaddr *) &sock,
                sizeof(sock)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
    }
	/* Initialize the lock mutex */
	pthread_mutexattr_init(&mutex_attr);
	pthread_mutexattr_setkind_np(&mutex_attr, PTHREAD_MUTEX_ERRORCHECK_NP);

	return(0);
}

/*
 * Disengage from the serial port
 */
static void emp_disengage(CONTEXT *ctx)
{
	if (ctx->emp_sock != -1)
		{
		int	tmp_sock;

		tmp_sock = ctx->emp_sock;
		ctx->emp_sock = -1;
		close(tmp_sock);
		}
	if(ctx->sensor_list)
		{
		while(ctx->sensor_list)
			sensor_free(ctx, ctx->sensor_list);
		}
	ctx->enable_driver_checks = 0;
	ctx->console_mode = 0;
	ctx->protocol_detected = 0;
	fru_free(ctx);
}

/*
 * Modify the serial port
 */
static void set_serial_port(CONTEXT *ctx, char *port)
{
	if (ctx->device_address)
		free(ctx->device_address);
	ctx->device_address = malloc((strlen(port)+1));
	strcpy(ctx->device_address, port);
}

/*
 * Modify the IP address
 */
static void set_ipaddress(CONTEXT *ctx, char *ip)
{
	pthread_mutex_lock(&ctx->extentd_lock);
	if (ctx->ip_address)
		free(ctx->ip_address);
	ctx->ip_address = malloc((strlen(ip)+1));
	strcpy(ctx->ip_address, ip);

	if (ctx->emp_extentd_fd != -1)
		{
		close(ctx->emp_extentd_fd);
		ctx->emp_extentd_fd = -1;
		}
	pthread_mutex_unlock(&ctx->extentd_lock);
	return;
}

/*
 * Modify the password
 */
static void set_password(CONTEXT *ctx, char *passwd)
{
	if (ctx->password)
		free(ctx->password);
	ctx->password = malloc((strlen(passwd)+1));
	strcpy(ctx->password, passwd);
}

/*
 * Process an IPC message
 */
static void process_emp_ipc(CONTEXT			*ctx,
														__uint32_t	job_id,
														int					client_fd,
														char				*ipc_cmd,
														char				*ipc_args)
{
	unsigned char	*response = NULL;
	int						response_len = 0;
	char					resp_msg[255];
	char					*last = (char *) 0x00;
	int						rc = 0;
	char					log_msg[255];

	if (!strcasecmp(ipc_cmd,"CONFIGURATION"))
		{
		char		*dev, *pass;
		CONTEXT	*ctx_scan;
		int			dev_changed = 0;
		int			pass_changed = 0;
		char		old_pass[16];

		if (!(dev = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(pass = strtok_r(NULL, ":", &last)))
			goto out_malformatted;
		if (strcasecmp(dev, "NONE"))
			{
			pthread_mutex_lock(&context_mutex);
			ctx_scan = context_head;
			while(ctx_scan)
				{
				if (ctx_scan->device_address)
					if ((!strcmp(ctx_scan->device_address, dev)) &&
							(ctx_scan != ctx))
						break;
				ctx_scan = ctx_scan->next;
				}
			pthread_mutex_unlock(&context_mutex);
			if (ctx_scan)
				{
				sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(EADDRINUSE));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			}

    // Do error checking on the password
		if (strlen(pass) > 8)
			{
			sprintf(resp_msg,
							"%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(EOVERFLOW));
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (strcasecmp(pass, "NONE"))
			{
			int chk;
			
			for (chk = 0; chk < (int) strlen(pass); chk++)
				if (!isdigit(pass[chk]))
					{
					sprintf(resp_msg,
									"%d:%d:FOR:JOB_ERROR:%s",
									job_id,
									client_fd,
									strerror(EINVAL));
					send_msg(ctx,ctx->thread_ipc_write, resp_msg);
					goto out_ok;
					}
			}

		// See if the serial port has changed 
		if ((ctx->device_address ? strcmp(ctx->device_address, dev) : 1))
			{
			sprintf(resp_msg,
							"0:0:SET:%s:EMP:SERIAL_PORT:%s",
							ctx->node_id,
							dev);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
      // If this is a new configuration... add the password
      if (!ctx->device_address)
        {
			  sprintf(resp_msg,
								"0:0:SET:%s:EMP:PASSWORD:%s",
								ctx->node_id,
								pass);
			  send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			  set_password(ctx, pass);
        }
			set_serial_port(ctx, dev);
			dev_changed = 1;
			}

		if (strcasecmp(ctx->password, pass))
			{
			strcpy(old_pass, ctx->password);
			set_password(ctx, pass);
			pass_changed = 1;
			}
		if ((dev_changed) || (pass_changed))
			{
			if (strcasecmp(dev,"NONE"))
				{
				int restartrc;
				restartrc = do_emp_restart(ctx, job_id, client_fd);
				if (restartrc != -EDQUOT)
					{
					sprintf(resp_msg,"0:0:SET:%s:EMP:PASSWORD:%s",ctx->node_id,pass);
					send_msg(ctx,ctx->thread_ipc_write, resp_msg);
					}
				else
					set_password(ctx, old_pass);
				}
			else if (dev_changed)
				{
#ifdef ENGAGE_DEBUG
				sprintf(log_msg,
								"0:0:LOG:Thread %s disengaging from port",
								ctx->node_id);
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
				emp_disengage(ctx);
				sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				}
			}
		else
			{
			sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			}
		goto out_ok;
		}
	if (ctx->emp_sock == -1)
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(ENOPKG));
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	if (!strcasecmp(ipc_cmd,"BMC_INFO"))
		{
		char	*iana_name;

		iana_name = iana_lookup(ctx->ver.manufacturer_id);
		sprintf(resp_msg,
						"%d:%d:FOR:BMCINFO:%x.%x:%s:%d.%d:%d",
						job_id,
						client_fd,
						ctx->ver.bmc_ver_major,
						ctx->ver.bmc_ver_minor,
						(iana_name != NULL ? iana_name : "Invalid IANA Number"),
						ctx->ver.ipmi_ver_major,
						ctx->ver.ipmi_ver_minor,
						ctx->ver.device_revision);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);

		sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd,"REFRESH"))
		{
		do_emp_restart(ctx, job_id, client_fd);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "INQUIRY"))
		{
		sprintf(resp_msg,
						"%d:%d:FOR:INQUIRY:%s:%s:%s:%s:%s",
						job_id,
						client_fd,
						ctx->node_id,
						(ctx->fru_data.board_serial ? ctx->fru_data.board_serial : "N/A"),
						ctx->device_address,
						"N/A",
						ctx->password);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "DOWNLOAD_LOG"))
		{
		FILE	*Tfilep_log;
		char	logfile[255];
		char	line[255];
		char	nexxus_line[255];

		sprintf(logfile,
						"%s/vacm_emp_sel_log.%s",
						LOGDIR,
						ctx->node_id);
		if (!(Tfilep_log = fopen(logfile,"r")))
			{
			if (errno == ENOENT)
				{
				validate_transaction(ctx, job_id, client_fd, 0, NULL);
				goto out_ok;
				}
			validate_transaction(ctx, job_id, client_fd, -errno, NULL);
			goto out_ok;
			}
		while(fgets(line, sizeof(line), Tfilep_log))
			{
			line[(strlen(line)-1)] = 0;
			sprintf(nexxus_line,
							"%d:%d:FOR:%s",
							job_id,
							client_fd,
							line);
			send_msg(ctx,ctx->thread_ipc_write, nexxus_line);
			}	
		fclose(Tfilep_log);
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "CLEAR_LOG"))
		{
		char	logfile[255];

		sprintf(logfile,
						"%s/vacm_emp_sel_log.%s",
						LOGDIR,
						ctx->node_id);
		if ((rc = unlink(logfile)) < 0)
			rc = -errno;
		validate_transaction(ctx, job_id, client_fd, rc, NULL);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "LOG_TAIL"))
		{
		char	*argument;
		int		slot;

		if (!(argument = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;

		rc = 0;
		if (!strcasecmp(argument, "ON"))
			{
			for (slot = 0; slot < MAX_SEL_WATCHERS; slot ++)
				if (ctx->sel_watchers[slot] == client_fd)
					break;
			if (slot != MAX_SEL_WATCHERS)
					rc = -EADDRINUSE;
			else
				{
				for (slot = 0; slot < MAX_SEL_WATCHERS; slot ++)
					if (ctx->sel_watchers[slot] == -1)
						break;
				if (slot == MAX_SEL_WATCHERS)
					rc = -EADDRNOTAVAIL;
				else
					ctx->sel_watchers[slot] = client_fd;
				}
			}
		else if (!strcasecmp(argument, "OFF"))
			{
			for (slot = 0; slot < MAX_SEL_WATCHERS; slot ++)
				if (ctx->sel_watchers[slot] == client_fd)
					break;
			if (slot == MAX_SEL_WATCHERS)
				rc = -ENOENT;
			else
				ctx->sel_watchers[slot] = -1;
			}
		else
			rc = -EINVAL;
		validate_transaction(ctx, job_id, client_fd, rc, NULL);
		goto out_ok;
		}

	if (ctx->console_mode)
		{
		validate_transaction(ctx, job_id, client_fd, -EUNATCH ,NULL);
		return;
		}
	if (!ctx->protocol_detected)
		{
		validate_transaction(ctx, job_id, client_fd, -ENOPROTOOPT ,NULL);
		return;
		}
	if (!ctx->negotiation_complete)
		{
		validate_transaction(ctx, job_id, client_fd, -EBUSY ,NULL);
		return;
		}

	if ((!strcasecmp(ipc_cmd, "READ_SENSOR")) ||
					(!strcasecmp(ipc_cmd, "IDENTIFY")))
		{
		char		*argument;
		SENSOR	*sensor;
		int			sensor_number;
		int			identify;
		char		sensor_msg[255];

		if (!(argument = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (ctx->emp_extentd_fd == -1)
			{
			validate_transaction(ctx, job_id, client_fd, -ENOTCONN, NULL);
			goto out_ok;
			}
		if (ctx->sensor_ipc_write == -1)
			{
			validate_transaction(ctx, job_id, client_fd, -ENOTCONN, NULL);
			goto out_ok;
			}
		if (!strcasecmp(ipc_cmd, "READ_SENSOR"))
			{
			identify = 0;
			if (!(sensor_number = atoi(argument)))
				{
				validate_transaction(ctx, job_id, client_fd, -EBADSLT, NULL);
				goto out_ok;
				}
			sensor = ctx->sensor_list;
			while(sensor)
				{
				if (lookup_sensor_number(sensor) == sensor_number)
					break;
				sensor = sensor -> next;
				}
			if (!sensor)
				{
				validate_transaction(ctx, job_id, client_fd, -EBADSLT, NULL);
				goto out_ok;
				}
			if (ctx->extentd_connect_status != 0)
				{
				sprintf(resp_msg,
								"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(ctx->extentd_connect_status));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			/* Pull sensor out of cache */
			if (sensor->rc != 0)
				{
				sprintf(resp_msg,
								"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(sensor->rc));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			else
				{
				sensor_str_convert(ctx,
													 sensor_number,
													 &sensor->cached_value[0],
													 sensor_msg);
				sprintf(resp_msg,
								"%d:%d:FOR:%s",
								job_id,
								client_fd,
								sensor_msg);
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				if (!strncmp(sensor_msg,"JOB_ERROR:",10))
					goto out_ok;
				}
			}
		else
			{
			if (!strcasecmp(argument, "ON"))
				send_msg(ctx, ctx->sensor_ipc_write, "ID:1");
			else if (!strcasecmp(argument, "OFF"))
				send_msg(ctx, ctx->sensor_ipc_write, "ID:0");
			else
				{
				validate_transaction(ctx, job_id, client_fd, -EINVAL, NULL);
				goto out_ok;
				}
			}
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd,"SET_ASSET_TAG"))
		{
		char	*tag;
		if (!(tag = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if ((rc = do_emp_set_at(ctx,tag)) < 0)
			{
			if (rc == -EOVERFLOW)
				{
				sprintf(resp_msg,
								"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(-rc));
				}
			else if (rc == -ENOSYS)
				{
				sprintf(resp_msg,
								"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(rc));
				}
			else
				{
				sprintf(resp_msg,
								"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(-rc));
				}
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		}
	else if (!strcasecmp(ipc_cmd,"POWER_OFF"))
		{
		unsigned char	arg = 0x00;
		
		if (ctx->bios_cfg.restricted_mode)
			{
			validate_transaction(ctx, job_id, client_fd, -EPERM, NULL);
			goto out_ok;
			}
		do_emp_skipreinit(ctx);
		rc = do_emp_transaction(ctx,
										 				CMD_CHASSIS_CONTROL,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				&arg,
										 				1,
														&response,
														&response_len);
		do_emp_submitpass(ctx);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd,"POWER_ON"))
		{
		unsigned char	arg = 0x01;

		if (ctx->bios_cfg.restricted_mode)
			{
			validate_transaction(ctx, job_id, client_fd, -EPERM, NULL);
			goto out_ok;
			}
		do_emp_skipreinit(ctx);
		rc = do_emp_transaction(ctx,
										 				CMD_CHASSIS_CONTROL,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				&arg,
										 				1,
														&response,
														&response_len);
		if (!rc)
			do_emp_submitpass(ctx);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd,"PULSE_FP_NMI"))
		{
		unsigned char	arg = 0x04;

		if (ctx->bios_cfg.restricted_mode)
			{
			validate_transaction(ctx, job_id, client_fd, -EPERM, NULL);
			goto out_ok;
			}
		rc = do_emp_transaction(ctx,
										 				CMD_CHASSIS_CONTROL,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				&arg,
										 				1,
														&response,
														&response_len);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd, "POWER_CYCLE"))
		{
		unsigned char	arg = 0x02;

		if (ctx->bios_cfg.restricted_mode)
			{
			validate_transaction(ctx, job_id, client_fd, -EPERM, NULL);
			goto out_ok;
			}
		do_emp_skipreinit(ctx);
		rc = do_emp_transaction(ctx,
										 				CMD_CHASSIS_CONTROL,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				&arg,
										 				1,
														&response,
														&response_len);
		if (!rc)
			do_emp_submitpass(ctx);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd, "HARD_RESET"))
		{
		unsigned char	arg = 0x03;

		if (ctx->bios_cfg.restricted_mode)
			{
			validate_transaction(ctx, job_id, client_fd, -EPERM, NULL);
			goto out_ok;
			}
		do_emp_skipreinit(ctx);
		rc = do_emp_transaction(ctx,
										 				CMD_CHASSIS_CONTROL,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				&arg,
										 				1,
														&response,
														&response_len);
		if (!rc)
			do_emp_submitpass(ctx);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd, "CHASSIS_STATUS"))
		{
		if (!(rc = do_emp_transaction(ctx,
										 				CMD_GET_CHASSIS_STATUS,
										 				CHASSIS_REQUEST,
										 				0x00,
										 				EMP_SLAVE_ADDRESS,
										 				EMP_APPLICATION_SLAVE_ADDRESS,
										 				NULL,
										 				0,
														&response,
														&response_len)))
			{
			if (response[6] == (unsigned char) 0x00)
				{
				CHASSIS_STATUS_RESPONSE	*status_resp;
				status_resp = (CHASSIS_STATUS_RESPONSE *) &response[6];
  			sprintf(resp_msg,
							"%d:%d:FOR:CSTATUS:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s:%s",
							job_id,
							client_fd,
          		(status_resp->ps_power_on             ? "YES" : "NO"),
          		(status_resp->ps_power_overload       ? "YES" : "NO"),
          		(status_resp->ps_interlock_active     ? "YES" : "NO"),
          		(status_resp->ps_power_fault          ? "YES" : "NO"),
          		(status_resp->ps_power_control_fault  ? "YES" : "NO"),
          		(status_resp->ps_power_restore_policy ? "YES" : "NO"),
          		(status_resp->lpe_ac_failed           ? "YES" : "NO"),
          		(status_resp->lpe_power_overload      ? "YES" : "NO"),
          		(status_resp->lpe_interlock_power_down? "YES" : "NO"),
          		(status_resp->lpe_power_fault         ? "YES" : "NO"),
          		(status_resp->lpe_commanded_onoff     ? "YES" : "NO"),
          		(status_resp->mcs_intrusion_active    ? "YES" : "NO"),
          		(status_resp->mcs_secure_mode_active  ? "YES" : "NO"),
          		(status_resp->mcs_drive_fault         ? "YES" : "NO"),
          		(status_resp->mcs_cooling_fan_fault   ? "YES" : "NO"));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				}
			}
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd, "NODE_STATUS"))
		{
		char	status[64];

		if (ctx->console_mode)
			strcpy(status,"CONSOLE");
		else if (!ctx->negotiation_complete)
			strcpy(status,"NEGOTIATING");
		else if (ctx->protocol_detected)
			strcpy(status,"DETECTED");
		else
			strcpy(status,"NOT_DETECTED");
			
		sprintf(resp_msg,
						"%d:%d:FOR:NODESTATUS:%s:%s",
						job_id,
						client_fd,
						ctx->device_address,
						status);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		}
	else if (!strcasecmp(ipc_cmd, "NODE_INV_INFO"))
		{
		sprintf(resp_msg,"%d:%d:FOR:NODEINVINFO:",
						job_id,
						client_fd);
		if (ctx->fru_data.board_manufacturer)
			{
			if (ctx->fru_data.board_manufacturer[0])
				strcat(resp_msg,ctx->fru_data.board_manufacturer);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.board_product_name)
			{
			if (ctx->fru_data.board_product_name[0])
				strcat(resp_msg,ctx->fru_data.board_product_name);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.board_part_number)
			{
			if (ctx->fru_data.board_part_number[0])
				strcat(resp_msg,ctx->fru_data.board_part_number);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.board_serial)
			{
			if (ctx->fru_data.board_serial[0])
				strcat(resp_msg,ctx->fru_data.board_serial);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		/* eth0 address no longer done this way */
		strcat(resp_msg,"N/A");

		strcat(resp_msg,":");
		if (ctx->fru_data.product_manuf)
			{
			if (ctx->fru_data.product_manuf[0])
				strcat(resp_msg,ctx->fru_data.product_manuf);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.product_name)
			{
			if (ctx->fru_data.product_name[0])
				strcat(resp_msg,ctx->fru_data.product_name);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.product_part)
			{
			if (ctx->fru_data.product_part[0])
				strcat(resp_msg,ctx->fru_data.product_part);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.product_ver)
			{
			if (ctx->fru_data.product_ver[0])
				strcat(resp_msg,ctx->fru_data.product_ver);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.product_serial)
			{
			if (ctx->fru_data.product_serial[0])
				strcat(resp_msg,ctx->fru_data.product_serial);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");
		strcat(resp_msg,":");
		if (ctx->fru_data.product_asset)
			{
			if (ctx->fru_data.product_asset[0])
				strcat(resp_msg,ctx->fru_data.product_asset);
			else
				strcat(resp_msg,"N/A");
			}
		else
			strcat(resp_msg,"N/A");

		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		}
	else if (!strcasecmp(ipc_cmd, "CHASSIS_CAPABILITIES"))
		{
		CAPABILITIES_RESPONSE	*cap;
		int	i;

		for (i=0;i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
										 							 CMD_GET_CHASSIS_CAPABILITIES,
										 							 CHASSIS_REQUEST,
										 							 0x00,
										 							 EMP_SLAVE_ADDRESS,
										 							 EMP_APPLICATION_SLAVE_ADDRESS,
										 							 NULL,
										 							 0,
																	 &response,
																	 &response_len)) < 0)
				{
				validate_transaction(ctx, job_id, client_fd, rc, response);
				goto out_ok;	
				}
			if (response[6] == (unsigned char) 0x00)
				break;
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			validate_transaction(ctx, job_id, client_fd, rc, response);
			free(response);
			goto out_ok;
			}
		cap = (CAPABILITIES_RESPONSE *) &response[6];

		sprintf(resp_msg,"%d:%d:FOR:CCAPABLE:",
						job_id,
						client_fd);
		if (cap->provides_intrusion)
			strcat(resp_msg,"YES:");
		else
			strcat(resp_msg,"NO:");
		if (cap->provides_secure)
			strcat(resp_msg,"YES:");
		else
			strcat(resp_msg,"NO:");
		if (cap->provides_fp_nmi)
			strcat(resp_msg,"YES:");
		else
			strcat(resp_msg,"NO:");
		if (cap->provides_power_interlock)
			strcat(resp_msg,"YES");
		else
			strcat(resp_msg,"NO");
		free(response);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		validate_transaction(ctx, job_id, client_fd, rc, response);
		}
	else if (!strcasecmp(ipc_cmd, "GET_SENSOR_LIST"))
		{
		SENSOR	*list;
		char		tmp[128], sensor_class[128];

		list = ctx->sensor_list;
  	while(list)
    	{
    	if (list->rec_header.record_type == 0x01)
      	{
 	     	decode_string(list->sensor_type.type_01_sensor.id_string_type_code,
 	     	              0,
 	     	              &list->sensor_type.type_01_sensor.id_string[0],
 	     	              &tmp[0],
 	     	              list->sensor_type.type_01_sensor.id_string_num_bytes);
 	    	if (list->sensor_type.type_01_sensor.sens_capable_reading_type == ANALOG)
					strcpy(sensor_class,"ANALOG");
 	     	else
 	       	strcpy(sensor_class,"UNSUPPORTED");
 	     	sprintf(resp_msg,
 	             "%d:%d:FOR:SENSOR:%d:%s:%s",
								job_id,
								client_fd,
 	              list->sensor_type.type_01_sensor.sensor_number,
 	              tmp,
 	              sensor_class);
 	     	}
 	   	else if (list->rec_header.record_type == 0x02)
 	     	{
 	     	int class;
 	     	decode_string(list->sensor_type.type_02_sensor.id_string_type_code,
 	     	              0,
 	     	              &list->sensor_type.type_02_sensor.id_string[0],
 		                  &tmp[0],
 	     	              list->sensor_type.type_02_sensor.id_string_num_bytes);
      	class = sensor_get_class(list->sensor_type.type_02_sensor.event_reading_type_code);
      	if (class == SENSOR_CLASS_DIGITAL)
        	strcpy(sensor_class,"DIGITAL");
      	else if (class == SENSOR_CLASS_DISCRETE)
        	strcpy(sensor_class,"UNSUPPORTED");
      	else
        	strcpy(sensor_class,"UNSUPPORTED");
      	sprintf(resp_msg,
              	"%d:%d:FOR:SENSOR:%d:%s:%s",
								job_id,
								client_fd,
              	list->sensor_type.type_01_sensor.sensor_number,
              	tmp,
              	sensor_class);
      	}
    	else
      	{
      	list = list->next;
      	continue;
     	 	}
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
    	list = list->next;
    	}
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		}
	else if (!strcasecmp(ipc_cmd, "GET_SENSOR_THRESHOLDS"))
		{
		SENSOR	*sens;
		int			sensor_number;
		char		*blum;
		int			src;
		if (!(blum = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;

		if (!(sensor_number = atoi(blum)))
			{
			validate_transaction(ctx, job_id, client_fd, -EBADSLT, NULL);
			goto out_ok;
			}
		sens = ctx->sensor_list;
		while(sens)
			{
			if ((src = lookup_sensor_number(sens)) == sensor_number)
				break;
    	else if (src < 0)
      	{
				validate_transaction(ctx, job_id, client_fd, -EOPNOTSUPP, NULL);
				goto out_ok;
      	}
			sens = sens->next;
			}
		if (!sens)
			{
			validate_transaction(ctx, job_id, client_fd, -EBADSLT, NULL);
			goto out_ok;
			}
		if (sens->rec_header.record_type == 0x01)
			{ 
			double lower_noncrit = 0; 
			double lower_non_recov = 0; 
			double lower_crit = 0; 
			double upper_noncrit = 0; 
			double upper_non_recov = 0; 
			double upper_crit = 0; 
			double minimum = 0; 
			double  maximum = 0;

			if (sens->sensor_type.type_01_sensor.sens_capable_reading_type != ANALOG)
				{  
				validate_transaction(ctx, job_id, client_fd, -EOPNOTSUPP, NULL);
				goto out_ok;
				}  
			/* It's an analog sensor... return the thresholds */
			if (sens->sensor_type.type_01_sensor.read_thres_mask_lower_non_crit)
				lower_noncrit = linearize_sensor_reading(sens,
                        sens->sensor_type.type_01_sensor.lower_non_critical);
			if (sens->sensor_type.type_01_sensor.read_thres_mask_lower_non_recov)
				lower_non_recov = linearize_sensor_reading(sens,
                      sens->sensor_type.type_01_sensor.lower_non_recoverable);
			if (sens->sensor_type.type_01_sensor.read_thres_mask_lower_crit)
				lower_crit = linearize_sensor_reading(sens,
                       sens->sensor_type.type_01_sensor.lower_critical);
			if (sens->sensor_type.type_01_sensor.read_thres_mask_upper_non_crit)
				upper_noncrit = linearize_sensor_reading(sens,
                          sens->sensor_type.type_01_sensor.upper_non_critical);
			if (sens->sensor_type.type_01_sensor.read_thres_mask_upper_non_recov)
				upper_non_recov = linearize_sensor_reading(sens,
                      sens->sensor_type.type_01_sensor.upper_non_recoverable);
			if (sens->sensor_type.type_01_sensor.read_thres_mask_upper_crit)
				upper_crit = linearize_sensor_reading(sens,  
                        sens->sensor_type.type_01_sensor.upper_critical);
			maximum  = linearize_sensor_reading(sens,
                      sens->sensor_type.type_01_sensor.sensor_maximum_reading);
			minimum  = linearize_sensor_reading(sens,
                    sens->sensor_type.type_01_sensor.sensor_minimum_reading);
      
			sprintf(resp_msg,                       
      				"%d:%d:FOR:THRESHOLD:%d:%.2f:%.2f:%.2f:%.2f:%.2f:%.2f:%.2f:%.2f",
							job_id,
							client_fd,
      				sensor_number,
      				minimum,                            
      				lower_noncrit,
      				lower_non_recov,
      				lower_crit,
      				upper_noncrit,
      				upper_non_recov,
      				upper_crit,
      				maximum);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			}         
		else if (sens->rec_header.record_type == 0x02)
			{         
			validate_transaction(ctx, job_id, client_fd, -EOPNOTSUPP, NULL);
			goto out_ok;
			}
		else
			{
			validate_transaction(ctx, job_id, client_fd, -EOPNOTSUPP, NULL);
			goto out_ok;
			} 
		validate_transaction(ctx, job_id, client_fd, 0, NULL);
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

out_ok:
	if (response)
		free(response);
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx,ctx->thread_ipc_write, resp_msg);
	if (response)
		free(response);
	return;
}

static int do_emp_rearm(CONTEXT *ctx)
{
	int						i, rc;
	__uint32_t		response_len;
	unsigned char	*response;
	char					log_msg[255];

	for (i=0;i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_REARM_EMP_CONNECTION_TIMER,
									 							 APP_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s EMP connectionr rearm failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	free(response);
	return(0);
}

static int do_emp_restart(CONTEXT			*ctx,
													__uint32_t	job_id,
													int					client_fd)
{
	char	resp_msg[255];
	int		rc, rrc = 0;

	ctx->enable_driver_checks = 0;
	ctx->negotiation_complete = 0;
	emp_disengage(ctx);
	sprintf(resp_msg,"%d:%d:FOR:STATUS:ENGAGING",job_id,client_fd);
	send_msg(ctx,ctx->thread_ipc_write, resp_msg);
	if ((rc = emp_engage(ctx)) < 0)
		{
		sprintf(resp_msg,"%d:%d:FOR:STATUS:ENGAGE_FAILED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);

		sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(-rc));
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		}
	else
		{
#ifdef DETECT_DEBUG
		printf("Thread %s calling detect_protocol as part of restart process\n",ctx->node_id);
#endif
		if ((rc = thread_detect_protocol(ctx)) < 0)
			{
			if (rc == -EUNATCH)
				{
				ctx->console_mode = 1;
				sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(EUNATCH));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				}
			else
				{
				sprintf(resp_msg,"%d:%d:FOR:STATUS:PROTOCOL_UNAVAILABLE",
								job_id,
								client_fd);
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				}
			rrc = rc;
			goto restart_out;
			}
		else
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:PROTOCOL_DETECTED",
							job_id,
							client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			}
		ctx->enable_driver_checks = 1;

		if (!(rc = do_emp_submitpass(ctx)))
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:CONNECTION_ACCEPTED",
							job_id,
							client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			}
		else if (rc == -EACCES)
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:CONNECTION_DENIED",
							job_id,
								client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EACCES;
			goto restart_out;
			}
		else if (rc == -EDQUOT)
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:INTERFACE_TEMP_LOCKED",
							job_id,
							client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EDQUOT;
			goto restart_out;
			}
		else if (rc == -EUNATCH)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(EUNATCH));
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EUNATCH;
			goto restart_out;
			}
		if (do_emp_getdeviceid(ctx) == -EUNATCH)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(EUNATCH));
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EUNATCH;
			goto restart_out;
			}
		if (do_emp_getchassiscapabilities(ctx) == -EUNATCH)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(EUNATCH));
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EUNATCH;
			goto restart_out;
			}
		if (do_emp_getempconfiguration(ctx) == -EUNATCH)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(EUNATCH));
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			rrc = -EUNATCH;
			goto restart_out;
			}

		if (ctx->capabilities.has_fru)
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:DOWNLOADING_FRU",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			if (do_emp_downloadfru(ctx) == -EUNATCH)
				{
				sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(EUNATCH));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				rrc = -EUNATCH;
				goto restart_out;
				}
			}
		if (ctx->capabilities.has_sdr)
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:DOWNLOADING_SDR",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			if (do_emp_downloadsdr(ctx) == -EUNATCH)
				{
				sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(EUNATCH));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				rrc = -EUNATCH;
				goto restart_out;
				}
			}
		if ((ctx->capabilities.has_sel) && (!ctx->bios_cfg.restricted_mode))
			{
			sprintf(resp_msg,"%d:%d:FOR:STATUS:DOWNLOADING_SEL",job_id,client_fd);
			send_msg(ctx,ctx->thread_ipc_write, resp_msg);
			if (do_emp_downloadsel(ctx) == -EUNATCH)
				{
				sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
								strerror(EUNATCH));
				send_msg(ctx,ctx->thread_ipc_write, resp_msg);
				rrc = -EUNATCH;
				goto restart_out;
				}
			}
		sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		}

restart_out:
	ctx->negotiation_complete = 1;
	ctx->enable_driver_checks = 1;
	return(rrc);
}

static int do_emp_skipreinit(CONTEXT *ctx)
{
	unsigned char	*response;
	int						response_len;
	unsigned char	arg = 0;
	int						rc;
	int						i;
	char					log_msg[255];

	arg |= 0x01;
	if (ctx->console_switch_allow)
		arg |= 0x02;

	for (i=0;i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_SET_EMP_REINIT,
									 							 APP_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 &arg,
									 							 sizeof(arg),
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s skip_reinit failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	free(response);
	return(0);
}

/*
 * Issue a CHASSIS_CAPABILITIES command
 * and stuff data into context
 */
static int do_emp_getchassiscapabilities(CONTEXT *ctx)
{
	CAPABILITIES_RESPONSE	*cap;
	int										i,rc;
	unsigned char					*response;
	int										response_len;
	char									log_msg[255];

	for (i=0;i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_GET_CHASSIS_CAPABILITIES,
									 							 CHASSIS_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			{
			return(rc);
			}
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		free(response);
		return(-EIO);
		}
	cap = (CAPABILITIES_RESPONSE *) &response[6];
	if (cap->provides_intrusion)
		ctx->capabilities.has_intrusion = 1;
	if (cap->provides_fp_nmi)
		ctx->capabilities.has_fpnmi = 1;
	if (cap->provides_power_interlock)
		ctx->capabilities.has_power_interlock = 1;

	ctx->addrs.fru_addr = cap->fru_addr;
	ctx->addrs.sdr_addr = cap->sdr_addr;
	ctx->addrs.sel_addr = cap->sel_addr;
	ctx->addrs.sm_addr = cap->sm_addr;
	ctx->addrs.icmb_addr = cap->icmb_addr;
	free(response);

#ifdef EMPHW_DEBUG
	sprintf(
		log_msg,
		"0:0:LOG:Thread %s fru@0x%.2x, sdr@0x%.2x sel@0x%.2x sm@0x%.2x icmb@0x%.2x",
		ctx->node_id,
		ctx->addrs.fru_addr,
		ctx->addrs.sdr_addr,
		ctx->addrs.sel_addr,
		ctx->addrs.sm_addr,
		ctx->addrs.icmb_addr);
					
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
	return(0);
}

/*
 * Issue the GET_EMP_CONFIGURATION command.
 * Stuff data into context
 */
static int do_emp_getempconfiguration(CONTEXT *ctx)
{
	EMP_CONFIG_RESP	*cfg;
	unsigned char		*response;
	int							response_len;
	int							rc, i;
	char						log_msg[255];

	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_GET_EMP_CONFIGURATION,
									 							 APP_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s get_emp_configuration failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	cfg = (EMP_CONFIG_RESP *) &response[6];

	ctx->bios_cfg.activation_mode = cfg->activation_mode;
	ctx->bios_cfg.restricted_mode = cfg->restricted_mode_enabled;
	ctx->bios_cfg.direct_connect = cfg->direct_connect_mode;
	ctx->bios_cfg.dcd_activation = cfg->dcd_activation_enabled;
	ctx->bios_cfg.password_null = cfg->password_null;
	free(response);
	return(0);
}

/*
 * Issue the GET_DEVICE_ID command.
 * Stuff data into context
 */
static int do_emp_getdeviceid(CONTEXT *ctx)
{
	DEVICE_ID_RESPONSE	*id;
	unsigned char				*response;
	int									response_len;
	int									rc, i;
	char								log_msg[255];
	char								*iana_name;

	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_GET_DEVICE_ID,
									 							 APP_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s get_device_id failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	id = (DEVICE_ID_RESPONSE *) &response[6];

	ctx->addrs.fru_devid = id->device_id;
	ctx->addrs.fru_devid = 0x00; /* Defined by IPMI 1.0 r1.1 spec pg 156 */

	ctx->ver.manufacturer_id = 0;
	ctx->ver.manufacturer_id  = (((unsigned int) id->manufacturer_id3) << 16);
	ctx->ver.manufacturer_id |= (((unsigned int) id->manufacturer_id2) << 8);
	ctx->ver.manufacturer_id |= ((unsigned int) id->manufacturer_id1);

	ctx->ver.ipmi_ver_major = id->ipmi_version_major;
	ctx->ver.ipmi_ver_minor = id->ipmi_version_minor;
	ctx->ver.bmc_ver_major = id->major_firmware_revision;
	ctx->ver.bmc_ver_minor = id->minor_firmware_revision;
	ctx->ver.device_revision = id->device_revision;

	ctx->capabilities.has_sensor = id->supports_sensor_device;
	ctx->capabilities.has_sdr = id->supports_sdr_device;
	ctx->capabilities.has_sel = id->supports_sel_device;
	ctx->capabilities.has_fru = id->supports_fru_device;
	ctx->capabilities.has_ipmb_recv = id->supports_ipmb_receiver;
	ctx->capabilities.has_ipmb_gen = id->supports_ipmb_generator;
	ctx->capabilities.has_bridge = id->supports_bridge;
	ctx->capabilities.has_chassis = id->supports_chassis_device;

	iana_name = iana_lookup(ctx->ver.manufacturer_id);

#ifdef EMPHW_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s ID (IPMI %d.%d) (BMC %x.%x) (MFID 0x%x [%s]) (HW rev %d)",
				 	ctx->node_id,
				 	ctx->ver.ipmi_ver_major,
				 	ctx->ver.ipmi_ver_minor,
				 	ctx->ver.bmc_ver_major,
				 	ctx->ver.bmc_ver_minor,
				 	ctx->ver.manufacturer_id,
					(iana_name != NULL ? iana_name : "Invalid IANA Number"),
				 	ctx->ver.device_revision);
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
#ifdef EMPHW_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s provides (SEN=%d)(SDR=%d)(SEL=%d)(FRU=%d)(IPR=%d)(IPG=%d)(BRDG=%d)(CHAS=%d)",
				 	ctx->node_id,
				 	ctx->capabilities.has_sensor,
				 	ctx->capabilities.has_sdr,
				 	ctx->capabilities.has_sel,
				 	ctx->capabilities.has_fru,
				 	ctx->capabilities.has_ipmb_recv,
				 	ctx->capabilities.has_ipmb_gen,
				 	ctx->capabilities.has_bridge,
				 	ctx->capabilities.has_chassis);
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
	free(response);
	return(0);
}

/* FRU Processing stuff */
static int do_emp_downloadfru(CONTEXT *ctx)
{
	FRU_INV_AREA_INFO				*fru;
	FRU_INV_DATA_REQ_IPMI09	req_09;
	FRU_INV_DATA_REQ_IPMI10	req_10;
	unsigned char						*arg;
	unsigned char						*response;
	int											response_len;
	int											rc, i, count;
	unsigned int						fru_pos, fru_size;
	unsigned int						fru_bytes_remaining_to_load;
	unsigned char						num;
	char										log_msg[255];
	int											ipmi10_flag;
	unsigned char						ipmi10_arg;
	int											size;
#ifdef FRU_DEBUG
	int											dbg;
#endif

	if (ctx->ver.ipmi_ver_major >= 1)
		ipmi10_flag = 1;
	else
		ipmi10_flag = 0;

	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if (ipmi10_flag)
			{
			ipmi10_arg = ctx->addrs.fru_devid;
			size = sizeof(ipmi10_arg);
			}
		else
			{
			ipmi10_arg = 0;
			size = 0;
			}
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_GET_FRU_INV_AREA_INFO,
									 							 STORAGE_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 &ipmi10_arg,
									 							 size,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		sprintf(log_msg,
						"0:0:LOG:Thread %s get_fru_inv_area_info failed. (0x%.2x)",
						ctx->node_id,
						response[6]);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s get_fru_inv_area_info failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	fru = (FRU_INV_AREA_INFO *) &response[6];
	fru_pos = 0;
	fru_bytes_remaining_to_load = fru->size;
	fru_size = fru->size;

	fru_free(ctx);

	ctx->fru_buffer = malloc(fru->size);
	memset(ctx->fru_buffer, 0, fru->size);
	free(response);
	if (ipmi10_flag)
		{
		req_10.dev_id= ctx->addrs.fru_devid;
		req_10.offset = 0;
		req_10.count = 16;
		arg = (unsigned char *) &req_10;
		size = sizeof(req_10);
		}
	else
		{
		req_09.offset = 0;
		req_09.count = 16;
		arg = (unsigned char *) &req_09;
		size = sizeof(req_09);
		}

	while(fru_bytes_remaining_to_load)
		{
#ifdef FRU_DEBUG
		if (ipmi10_flag)
			printf("FRU blk req (o %d, c %d, brtl %d dev 0x%x (ipmi10))\n",
							req_10.offset,
							req_10.count,
							fru_bytes_remaining_to_load,
							req_10.dev_id);
		else
			printf("FRU blk req (o %d, c %d, brtl %d (ipmi09))\n",
							req_09.offset,
							req_09.count,
							fru_bytes_remaining_to_load);
#endif
		for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
									 							 	 CMD_READ_FRU_INV_DATA,
									 							 	 STORAGE_REQUEST,
									 							 	 0x00,
									 							 	 EMP_SLAVE_ADDRESS,
									 							 	 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 	 (unsigned char*) arg,
									 							 	 size,
																 	 &response,
																 	 &response_len)) < 0)
				return(rc);
			if (response[6] == (unsigned char) 0x00)
				break;
			sprintf(log_msg,
							"0:0:LOG:Thread %s get_fru_inv_data failed. (0x%.2x)",
							ctx->node_id,response[6]);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s get_fru_inv_data failed. Retries exceeded",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-EREMOTEIO);
			}
		num = response[7];
#ifdef FRU_DEBUG
		printf("Thread %s got %d bytes of FRU data\n",ctx->node_id,num);
		for (dbg = 0; dbg < num; dbg++)
			printf("%.2x ",response[8+dbg]);
		printf("\n");
#endif
		memcpy(&ctx->fru_buffer[fru_pos], &response[8], num);
		free(response);

		fru_pos += num;
		fru_bytes_remaining_to_load -= num;
		if (fru_bytes_remaining_to_load == 0)
			break;
		if (fru_bytes_remaining_to_load <= 16)
			count = fru_bytes_remaining_to_load;
		else
			count = 16;
		if (ipmi10_flag)
			{
			req_10.count = count;
			req_10.offset = fru_pos;
			req_10.dev_id = ctx->addrs.fru_devid;
			size = sizeof(req_10);
			arg = (unsigned char *) &req_10;
			}
		else
			{
			req_09.count = count;
			req_09.offset = fru_pos;
			size = sizeof(req_09);
			arg = (unsigned char *) &req_09;
			}
		}
#ifdef FRU_DEBUG
	printf("Thread %s FRU download complete\n",ctx->node_id);
#endif
#ifdef FRU_DEBUG
	{
	char	fname[255];
	int		fd;

	sprintf(fname,"/tmp/fru.%s",ctx->node_id);
	if ((fd = open(fname,O_WRONLY | O_CREAT)) < 0)
		perror("open");
	else
		{
		write(fd, ctx->fru_buffer, fru_size);
		close(fd);
		}
	}
#endif

	parse_fru(ctx);

#ifdef FRU_DEBUG
	printf("Thread %s FRU parsing complete\n",ctx->node_id);
#endif
	return(0);
}

/*
 * Set the asset tag on the machine
 */
static int do_emp_set_at(CONTEXT *ctx, char *tag)
{
	int							rc, response_len, i;
	unsigned char		*response;
	char						log_msg[255];
	unsigned short	*fru_write_offset;
	unsigned char		*fru_device_id; // Required only for IPMI 1.0
	unsigned char		*write_buffer;
	int							data_len = 0;
  struct type_length
    {
    unsigned char num_bytes :6;
    unsigned char type_code :2;
    } *tl;

	if (ctx->fru_data.fru_at_offset < 0)
		return(-ELIBBAD);
	if ((int)strlen(tag) > ctx->fru_data.fru_at_reserved_bytes)
		return(-EOVERFLOW);
	
	if (ctx->ver.ipmi_ver_major >= 0x01)
		{
		return(-ENOSYS);
		data_len = ctx->fru_data.fru_at_reserved_bytes +
							 sizeof(unsigned char) +
							 sizeof(unsigned short) +
							 sizeof(struct type_length);
		if (!(write_buffer = malloc(data_len)))
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s unable to malloc fru buffer",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-ENOMEM);
			}
		memset(write_buffer, 0, data_len);
		fru_device_id = (unsigned char *) &write_buffer[0];
		fru_write_offset = (unsigned short *) &write_buffer[1];
		tl = (struct type_length *) &write_buffer[3];
		*fru_device_id = 0x00; // unknown
		*fru_write_offset = ctx->fru_data.fru_at_offset;
		tl->type_code = 0x03;
		tl->num_bytes = ctx->fru_data.fru_at_reserved_bytes;
  	memcpy(&write_buffer[4], tag, strlen(tag));
		}
	else
		{
		data_len = ctx->fru_data.fru_at_reserved_bytes +
							 sizeof(unsigned short) +
							 sizeof(struct type_length);
		if (!(write_buffer = malloc(data_len)))
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s unable to malloc fru buffer",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-ENOMEM);
			}
		memset(write_buffer, 0, data_len);
		fru_write_offset = (unsigned short *) &write_buffer[0];
		tl = (struct type_length *) &write_buffer[2];
		*fru_write_offset = ctx->fru_data.fru_at_offset;
		tl->type_code = 0x03;
		tl->num_bytes = ctx->fru_data.fru_at_reserved_bytes;
  	memcpy(&write_buffer[3], tag, strlen(tag));
		}

	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
							 							 	 	 CMD_WRITE_FRU_INV_DATA,
							 							 	 	 STORAGE_REQUEST,
							 							 	 	 0x00,
							 							 	 	 EMP_SLAVE_ADDRESS,
							 							 	 	 EMP_APPLICATION_SLAVE_ADDRESS,
							 							 	 	 (unsigned char *) write_buffer,
							 							 	 	 data_len,
														 	 	 &response,
														 	 	 &response_len)) < 0)
			return(rc);
		if (response[6] == 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s write_fru_inv_data failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	if (ctx->fru_data.product_asset)
		free(ctx->fru_data.product_asset);
	ctx->fru_data.product_asset = malloc((strlen(tag)+1));
  memcpy(&ctx->fru_data.product_asset[0], tag, strlen(tag));
/* force fru reload
	memcpy(&ctx->fru_buffer[ctx->fru_data.fru_at_offset],
				 tag,
				 sizeof(asset_tag_pkt_ipmi09.data));
 */
	free(response);
	do_emp_downloadfru(ctx);
	return(0);
}

static void fru_free(CONTEXT *ctx)
{
	if (ctx->fru_buffer)
		{
		free(ctx->fru_buffer);
		ctx->fru_buffer = NULL;
		}
	if (ctx->fru_data.board_manufacturer)
		{
		free(ctx->fru_data.board_manufacturer);
		ctx->fru_data.board_manufacturer = NULL;
		}
	if (ctx->fru_data.board_product_name)
		{
		free(ctx->fru_data.board_product_name);
		ctx->fru_data.board_product_name = NULL;
		}
	if (ctx->fru_data.board_serial)
		{
		free(ctx->fru_data.board_serial);
		ctx->fru_data.board_serial = NULL;
		}
	if (ctx->fru_data.board_part_number)
		{
		free(ctx->fru_data.board_part_number);
		ctx->fru_data.board_part_number = NULL;
		}
	if (ctx->fru_data.product_manuf)
		{
		free(ctx->fru_data.product_manuf);
		ctx->fru_data.product_manuf = NULL;
		}
	if (ctx->fru_data.product_name)
		{
		free(ctx->fru_data.product_name);
		ctx->fru_data.product_name = NULL;
		}
	if (ctx->fru_data.product_part)
		{
		free(ctx->fru_data.product_part);
		ctx->fru_data.product_part = NULL;
		}
	if (ctx->fru_data.product_ver)
		{
		free(ctx->fru_data.product_ver);
		ctx->fru_data.product_ver = NULL;
		}
	if (ctx->fru_data.product_serial)
		{
		free(ctx->fru_data.product_serial);
		ctx->fru_data.product_serial = NULL;
		}
	if (ctx->fru_data.product_asset)
		{
		free(ctx->fru_data.product_asset);
		ctx->fru_data.product_asset = NULL;
		}
}

static int do_emp_downloadsdr(CONTEXT *ctx)
{
	RESERVE_SDR_REPOSITORY_RESPONSE	*sdr_res_resp;
	SDR_REPOSITORY_INFO_RESPONSE		*sdr_info_resp;
	GET_SDR_REQUEST									request;
	GET_SDR_RESPONSE								*sdr_get_resp;
	SDR_HEADER											*sdr_hdr;
	unsigned short									sdr_current_record;
	unsigned short									sdr_next_record;
	unsigned short									sdr_reservation_id;
	unsigned int										sdr_buffer_pos;
	unsigned char										*response;
	int															sdr_record_bytes_remaining_to_load;
	int															response_len, rc, i;
	int															sdr_hunk_size;
	unsigned char										sdr_buffer[64];
	char														log_msg[255];

	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_RESERVE_SDR_REPOSITORY,
									 							 STORAGE_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s reserve_sdr failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	sdr_res_resp = (RESERVE_SDR_REPOSITORY_RESPONSE *) &response[6];
	sdr_reservation_id = sdr_res_resp->res_id;
	free(response);
#ifdef SDR_DEBUG
	printf("Thread %s downloading SDR repository info\n",ctx->node_id);
#endif
	for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
		{
		if ((rc = do_emp_transaction(ctx,
									 							 CMD_GET_SDR_REPOSITORY_INFO,
									 							 STORAGE_REQUEST,
									 							 0x00,
									 							 EMP_SLAVE_ADDRESS,
									 							 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 NULL,
									 							 0,
																 &response,
																 &response_len)) < 0)
			return(rc);
		if (response[6] == (unsigned char) 0x00)
			break;
		free(response);
		usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
		}
	if (i == EMP_HIGH_LEVEL_RETRY_MAX)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s get_sdr_rep_info failed. Retries exceeded",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(-EREMOTEIO);
		}
	sdr_info_resp = (SDR_REPOSITORY_INFO_RESPONSE *) &response[6];
	free(response);

#ifdef SDR_DEBUG
	printf("Thread %s downloading SDR records\n",ctx->node_id);
#endif
	/* Initial record to read */
	sdr_next_record = 0x0000;
	sdr_current_record = 0x0000;
	while(1)
		{
		request.reservation_id = sdr_reservation_id;
		request.record_id = sdr_current_record;
		request.offset_into_record = 0;
		request.bytes_to_read = sizeof(SDR_HEADER);
		for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
									 								 CMD_GET_SDR,
									 							 	 STORAGE_REQUEST,
									 							 	 0x00,
									 							 	 EMP_SLAVE_ADDRESS,
									 							 	 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 	 (unsigned char *) &request,
									 							 	 sizeof(request),
																 	 &response,
																 	 &response_len)) < 0)
				return(rc);
			if (response[6] == (unsigned char) 0x00)
				break;
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s get_sdr failed. Retries exceeded (record 0x%x)",
							ctx->node_id,request.record_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-EREMOTEIO);
			}
	
		sdr_get_resp = (GET_SDR_RESPONSE *) &response[6];
		sdr_next_record = sdr_get_resp->next_record;
		sdr_hdr = (SDR_HEADER *) &response[9];
		sdr_record_bytes_remaining_to_load = sdr_hdr->record_length +
																				sizeof(SDR_HEADER);
		free(response);

		/* Download the full record */
		sdr_buffer_pos = 0;
		if (sdr_record_bytes_remaining_to_load <= 16)
			sdr_hunk_size = sdr_record_bytes_remaining_to_load;
		else
			sdr_hunk_size = 16;
		while(sdr_record_bytes_remaining_to_load)
			{
#ifdef SDR_DEBUG
			printf("SDR block request (record 0x%.4x, size %d, remain %d)\n",
						 sdr_current_record,
						 sdr_hunk_size,
						 sdr_record_bytes_remaining_to_load);
#endif
			memset(&request, 0, sizeof(request));
			request.reservation_id = sdr_reservation_id;
			request.record_id = sdr_current_record;
			request.offset_into_record = sdr_buffer_pos;
			request.bytes_to_read = sdr_hunk_size;
	
			for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
				{
				if ((rc = do_emp_transaction(ctx,
									 							 	 	 CMD_GET_SDR,
									 							 	 	 STORAGE_REQUEST,
									 							 	 	 0x00,
									 							 	 	 EMP_SLAVE_ADDRESS,
									 							 	 	 EMP_APPLICATION_SLAVE_ADDRESS,
									 							 	 	 (unsigned char *) &request,
									 							 	 	 sizeof(request),
																 	 	 &response,
																 	 	 &response_len)) < 0)
					return(rc);
				if (response[6] == (unsigned char) 0x00)
					break;
				free(response);
				usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
				}
			if (i == EMP_HIGH_LEVEL_RETRY_MAX)
				{
				sprintf(log_msg,
								"0:0:LOG:Thread %s get_sdr failed. Retries exceeded",
								ctx->node_id);
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
				return(-EREMOTEIO);
				}
			sdr_get_resp = (GET_SDR_RESPONSE *) &response[6];
			memcpy(&sdr_buffer[sdr_buffer_pos],
					 	 &response[9],
					 	 sdr_hunk_size);
#ifdef SDR_DEBUG
			printf("Hunk of size %d downloaded (remain %d)\n",
							sdr_hunk_size,
							sdr_record_bytes_remaining_to_load);
#endif
			sdr_buffer_pos += sdr_hunk_size;
			sdr_record_bytes_remaining_to_load -= sdr_hunk_size;
			if (sdr_record_bytes_remaining_to_load <= 16)
				sdr_hunk_size = sdr_record_bytes_remaining_to_load;
			else
				sdr_hunk_size = 16;
			free(response);
			}
#ifdef SDR_DEBUG
printf("Thread %s SDR record 0x%.4x downloaded\n",ctx->node_id, sdr_current_record);
#endif
		/* Record download complete */
		process_sdr_record(ctx, &sdr_buffer[0]);

		if (sdr_next_record == (unsigned short) 0xffff)
			{
#ifdef SDR_DEBUG
			sprintf(log_msg,
							"0:0:LOG:Thread %s SDR record downloaded complete (%d sensors)",
						 	ctx->node_id,
						 	return_num_processed_sensors(ctx));
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif
			break;
			}
		else
			{
			sdr_current_record = sdr_next_record;
			}
		}
#ifdef SDR_DEBUG
	printf("Thread %s SDR download complete\n",ctx->node_id);
#endif
	return(0);
}

#define BOARD_AREA_NUM_FIELDS 4
#define BOARD_MANUFACTURER    0
#define BOARD_PRODUCT_NAME    1
#define BOARD_SERIAL          2
#define BOARD_PART            3

#define PRODUCT_AREA_NUM_FIELDS 6
#define PRODUCT_MANUFACTURER    0
#define PRODUCT_NAME            1
#define PRODUCT_PART            2
#define PRODUCT_VERSION         3
#define PRODUCT_SERIAL          4
#define PRODUCT_ASSET           5

static void parse_fru(CONTEXT *ctx)
{
	char							board_area[BOARD_AREA_NUM_FIELDS][64];
	char							product_area[PRODUCT_AREA_NUM_FIELDS][64];
  FRU_COMMON_HEADER *common_hdr;
  unsigned char     *p;
	int								i;
  unsigned char     lang;
	char							log_msg[255];
  struct type_length
    {
    unsigned char num_bytes :6;
    unsigned char type_code :2;
    } *tl;

	memset(board_area, 0, sizeof(board_area));
	memset(product_area, 0, sizeof(product_area));
  common_hdr = (FRU_COMMON_HEADER *) &ctx->fru_buffer[0];

  if ((common_hdr->board_area_offset) && (common_hdr->board_area_offset != 0xff))
    {
    p = &ctx->fru_buffer[(common_hdr->board_area_offset * 8)];
    lang = p[2];
    p +=6;
		for (i=0; i < BOARD_AREA_NUM_FIELDS; i++)
			{
			tl = (struct type_length *) p;
			p++;
			if (tl->num_bytes != 0)
				decode_string(tl->type_code, lang, p, board_area[i], tl->num_bytes);
			board_area[i][tl->num_bytes] = 0;
			p += tl->num_bytes;
			}
    }
  else
		{
    sprintf(log_msg,
						"0:0:LOG:FRU on %s does not contain a Board Info Area!",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}

  /*
   * Now parse product info area
   */
  if ((common_hdr->product_info_offset) && (common_hdr->product_info_offset != 0xff))
    {
    p = &ctx->fru_buffer[(common_hdr->product_info_offset * 8)];
    lang = p[2];
    p+=3;
		for (i=0; i < PRODUCT_AREA_NUM_FIELDS; i++)
			{
			tl = (struct type_length *) p;
			if (i == PRODUCT_ASSET)
				{
				ctx->fru_data.fru_at_reserved_bytes = tl->num_bytes;
    		ctx->fru_data.fru_at_offset =
					(((unsigned int)p)-((unsigned int)ctx->fru_buffer));
				}
			p++;
			if (tl->num_bytes != 0)
				decode_string(tl->type_code, lang, p, product_area[i], tl->num_bytes);
			product_area[i][tl->num_bytes] = 0;
			p += tl->num_bytes;
			}
    }
  else
		{
		ctx->fru_data.fru_at_offset = -1;
    sprintf(log_msg,
						"0:0:LOG:FRU on %s does not contain a Product Info Area!",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	fru_free(ctx);

	ctx->fru_data.board_manufacturer = 
			malloc((strlen(board_area[BOARD_MANUFACTURER])+1));
	strcpy(ctx->fru_data.board_manufacturer, board_area[BOARD_MANUFACTURER]);

	ctx->fru_data.board_product_name = 
			malloc((strlen(board_area[BOARD_PRODUCT_NAME])+1));
	strcpy(ctx->fru_data.board_product_name, board_area[BOARD_PRODUCT_NAME]);

	ctx->fru_data.board_serial = 
			malloc((strlen(board_area[BOARD_SERIAL])+1));
	strcpy(ctx->fru_data.board_serial, board_area[BOARD_SERIAL]);

	ctx->fru_data.board_part_number = 
			malloc((strlen(board_area[BOARD_PART])+1));
	strcpy(ctx->fru_data.board_part_number, board_area[BOARD_PART]);


	ctx->fru_data.product_manuf = 
			malloc((strlen(product_area[PRODUCT_MANUFACTURER])+1));
	strcpy(ctx->fru_data.product_manuf, product_area[PRODUCT_MANUFACTURER]);

	ctx->fru_data.product_name = 
			malloc((strlen(product_area[PRODUCT_NAME])+1));
	strcpy(ctx->fru_data.product_name, product_area[PRODUCT_NAME]);

	ctx->fru_data.product_part = 
			malloc((strlen(product_area[PRODUCT_PART])+1));
	strcpy(ctx->fru_data.product_part, product_area[PRODUCT_PART]);

	ctx->fru_data.product_ver = 
			malloc((strlen(product_area[PRODUCT_VERSION])+1));
	strcpy(ctx->fru_data.product_ver, product_area[PRODUCT_VERSION]);

	ctx->fru_data.product_serial = 
			malloc((strlen(product_area[PRODUCT_SERIAL])+1));
	strcpy(ctx->fru_data.product_serial, product_area[PRODUCT_SERIAL]);

	ctx->fru_data.product_asset = 
			malloc((strlen(product_area[PRODUCT_ASSET])+1));
	strcpy(ctx->fru_data.product_asset, product_area[PRODUCT_ASSET]);


#if 0
  for (i=0; i < BOARD_AREA_NUM_FIELDS; i++)
    printf("Field %d = %s\n",i,board_area[i]);
  for (i=0; i < PRODUCT_AREA_NUM_FIELDS; i++)
    printf("Field %d = %s\n",i,product_area[i]);
#endif
}

static void process_sdr_record(CONTEXT *ctx, unsigned char *buffer)
{
	SDR_HEADER		*rec_header;
	SENSOR				*sensor_new;
	char					tmp[255];
	char					log_msg[255];

	rec_header = (SDR_HEADER *) &buffer[0];

  if ((rec_header->record_type != 0x01) &&
      (rec_header->record_type != 0x02) &&
      (rec_header->record_type != 0x10) &&
      (rec_header->record_type != 0xC0))
    {
    sprintf(log_msg,
						"0:0:LOG:Bad SDR header (record 0x%.4x, type 0x%x)",
					 	rec_header->record_id,
					 	rec_header->record_type);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
    return;
    }

#if 0
  printf("process_sdr_record(): Processing record 0x%.4x, type 0x%x\n",
          rec_header->record_id,
          rec_header->record_type);
#endif
	sensor_new = sensor_alloc(ctx);
	sensor_new->rc = -ENODATA;

  memcpy(&sensor_new->rec_header,
				 rec_header,
         sizeof(SDR_HEADER));

  if (sensor_new->rec_header.record_type == 0x01)
    {
    memcpy(&sensor_new->sensor_type.type_01_sensor,
           &buffer[0],
           sensor_new->rec_header.record_length+sizeof(SDR_HEADER));
#if 0
    decode_string(sensor_new->sensor_type.type_01_sensor.id_string_type_code,
                  0,
                  &sensor_new->sensor_type.type_01_sensor.id_string[0],
                  &tmp[0],
                  sensor_new->sensor_type.type_01_sensor.id_string_num_bytes);
    printf("Internal Sensor Found -> '%s'\n",tmp);
#endif
    }
  else if (sensor_new->rec_header.record_type == 0x02)
    {
    memcpy(&sensor_new->sensor_type.type_02_sensor,
           &buffer[0],
           sensor_new->rec_header.record_length + sizeof(SDR_HEADER));
#if 0
    decode_string(sensor_new->sensor_type.type_02_sensor.id_string_type_code,
                  0,
                  &sensor_new->sensor_type.type_02_sensor.id_string[0],
                  &tmp[0],
                  sensor_new->sensor_type.type_02_sensor.id_string_num_bytes);      
		printf("Discrete/Digital Sensor Found -> '%s'\n",tmp);
#endif
    }
  else if (sensor_new->rec_header.record_type == 0x10)
    {
    memcpy(&sensor_new->sensor_type.type_10_sensor,
           &buffer[0],
           sensor_new->rec_header.record_length+sizeof(SDR_HEADER));
    if (!return_device_type_entry(
						sensor_new->sensor_type.type_10_sensor.device_type,
            sensor_new->sensor_type.type_10_sensor.device_type_modifier,
            &tmp[0]))
			{
      sprintf(log_msg,
							"0:0:LOG:Unable to lookup device with code 0x%x, modifier 0x%x",
            	sensor_new->sensor_type.type_10_sensor.device_type,
            	sensor_new->sensor_type.type_10_sensor.device_type_modifier);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
      }
    else
			{
#if 0
      printf("process_sdr_record(): IPMB '%s' found \n",tmp);
#endif
      }
    }
  else if (sensor_new->rec_header.record_type == 0xC0)
    {
#if 0
    memcpy(&sensor_new->sensor_type.type_c0_sensor,
           &buffer[0],
           sensor_new->rec_header.record_length +5);
    printf("process_sdr_record(): OEM Record Found\n");
#endif
    }
}

static int do_emp_submitpass(CONTEXT *ctx)
{
	char					pass[8];
	int						rc;
	unsigned char	*response;
	int						response_len;
	char					log_msg[255];

	memset(pass, 0, sizeof(pass));
	if (ctx->password)
		{
		if (strcasecmp(ctx->password,"NONE"))
			strncpy(pass, ctx->password, sizeof(pass));
		}
	rc = do_emp_transaction(ctx,
									 				CMD_SUBMIT_PASSWORD,
									 				APP_REQUEST,
									 				0x00,
									 				EMP_SLAVE_ADDRESS,
									 				EMP_APPLICATION_SLAVE_ADDRESS,
									 				pass,
									 				8,
													&response,
													&response_len);
	if (rc != 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Node %s password submission failed (%s)",
						ctx->node_id,
						strerror(-rc));
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	else if (rc == 0)
		{
		if (response[6] == 0x00)
			ctx->session_active = 1;
		else
			{
			ctx->session_active = 0;
			if (response[6] == 0x01)
				{
				sprintf(log_msg,
								"0:0:LOG:Node %s password submission failed (Bad Password)",
								ctx->node_id);
				rc = -EACCES;
				}
			else if (response[6] == 0x02)
				{
				sprintf(log_msg,
								"0:0:LOG:Node %s password submission failed (Interface Locked for 30 seconds)",
								ctx->node_id);
				rc = -EDQUOT;
				}
			else
				{
				sprintf(log_msg,
								"0:0:LOG:Node %s password submission failed (0x%.2x)",
								ctx->node_id,
								response[6]);
				rc = -EACCES;
				}
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			}
		free(response);
		}
	return(rc);
}

/*
 * Validate the return of a do_emp_transaction call
 * Send the proper string to the client.
 * rc should be the return code from do_emp_transaction
 * response should be the response packet or NULL if N/A
 */
static void validate_transaction(CONTEXT				*ctx,
																 __uint32_t			job_id,
																 int						client_fd,
																 int						rc,
																 unsigned char	*response)
{
	char	resp_msg[255];

	if (rc != 0)
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(-rc));
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		return;
		}
	if (response)
		{
		if (response[6] != 0x00)
			{
			sprintf(resp_msg,
							"%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							return_ccstr(response[6]));
			}
		else
			sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		}
	else if (!rc)
		{
		sprintf(resp_msg,"%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx,ctx->thread_ipc_write, resp_msg);
		}
}

static int thread_detect_protocol(CONTEXT *ctx)
{
	DRIVER_IPC			msg;
	struct timeval	to;
	fd_set					read_fds;
	pthread_t				self;
	int							rc;

	self = pthread_self();

	if (pthread_equal(self, ctx->driver_thread))
		{
		printf("ACK! Driver thread called thread_detect_protocol!\n");
		}
	/* Instruct the driver to detect protocol */
	memset(&msg, 0, sizeof(msg));
	msg.request = DIPCREQ_DET;
	if (write(ctx->thread_pwrite, (void *) &msg, sizeof(DRIVER_IPC))<0)
		{
		printf("unable to write to driver (%m)\n");
		return(-EIO);
		}

	/* Wait for response pkt from driver */
	FD_ZERO(&read_fds);
	FD_SET(ctx->thread_pread, &read_fds);
	to.tv_sec = 30;
	to.tv_usec = 0;
	if ((rc = select((ctx->thread_pread +1), &read_fds, NULL, NULL, &to)) < 0)
		{
		printf("select failure (%m)\n");
		return(-EIO);
		}
	if (!rc)
		{
		printf("Timeout waiting for response from driver\n");
		return(-ETIMEDOUT);
		}
	/* Pull the response data out */
	if (read(ctx->thread_pread, (void *) &msg, sizeof(DRIVER_IPC)) < 0)
		{
		printf("unable to read response\n");
		return(-EIO);
		}
	if (msg.request != DIPCREQ_DET)
		{
		printf("Thread %s expected DET got %d\n",ctx->node_id,msg.request);
		return(-EIO);
		}
	return(msg.rc);
}

static int do_emp_transaction(CONTEXT				*ctx,
											 				unsigned char	cmd,
											 				unsigned char	netfn,
											 				unsigned char	lun,
											 				unsigned char	rs_slave_addr,
											 				unsigned char	rq_slave_addr,
											 				unsigned char	*data,
											 				int						I_dlen,
															unsigned char	**response_buffer,
															int						*response_len)
{
	unsigned char		raw_send[EMP_MAX_MSG_LEN*2];
	unsigned char		norm_send[EMP_MAX_MSG_LEN*2];
	REQHDR					*hdr = (REQHDR *) &norm_send[1];
	int							len, I_c1;
	unsigned char		cksum;
	int							rc;
	DRIVER_IPC			msg;
	struct timeval	to;
	fd_set					read_fds;
	char						log_msg[255];

#ifdef TRANS_DEBUG
	printf("[%s] cmd 0x%.2x\n",ctx->node_id, cmd);
#endif
	*response_buffer = NULL;

	if (ctx->console_mode)
		return(-EUNATCH);
	if (!ctx->protocol_detected)
		return(-ENOPROTOOPT);

	memset(raw_send, 0, sizeof(raw_send));
	memset(norm_send, 0, sizeof(norm_send));

  hdr->rs_slave_addr= rs_slave_addr;
  hdr->net_fn       = netfn;
  hdr->lun          = lun;
  hdr->cksum1       = ((0xff - hdr->rs_slave_addr - 
											(*(unsigned char *)&norm_send[2]))+1);
  hdr->rq_slave_addr= rq_slave_addr;
  hdr->rqseq        = ctx->sequencenumber+=4;
  hdr->cmd          = cmd;
	if (I_dlen)
  	memcpy(&norm_send[7],data,I_dlen);
  cksum=0xff - hdr->rq_slave_addr - hdr->rqseq - hdr->cmd;
  for (I_c1=7;I_c1<(7+I_dlen);I_c1++)
    cksum=cksum - norm_send[I_c1];
  norm_send[7+I_dlen]= (cksum+1);

	len = unnormalize_data(&norm_send[1],
												 &raw_send[1],
												 (8+I_dlen));

	raw_send[0] = (unsigned char) EMP_START_CHAR;
	raw_send[len] = (unsigned char) EMP_STOP_CHAR;
	len++;

	/* Instruct the driver to send the packet out */
	memset(&msg, 0, sizeof(msg));
	msg.request = DIPCREQ_TRANS;
	msg.cmd = cmd;
	memcpy(&msg.pkt_buffer, raw_send, len);
	msg.pkt_len = len;
	if (write(ctx->thread_pwrite, (void *) &msg, sizeof(DRIVER_IPC))<0)
		{
		printf("unable to write to driver (%m)\n");
		return(-EIO);
		}

	/* Wait for response pkt from driver */
	FD_ZERO(&read_fds);
	FD_SET(ctx->thread_pread, &read_fds);
	to.tv_sec = 10;
	to.tv_usec = 0;
	if ((rc = select((ctx->thread_pread +1), &read_fds, NULL, NULL, &to)) < 0)
		{
		printf("select failure (%m)\n");
		return(-EIO);
		}
	if (!rc)
		{
		printf("select timeout for response\n");
		return(-ETIMEDOUT);
		}
	/* Pull the response data out */
	if (read(ctx->thread_pread, (void *) &msg, sizeof(DRIVER_IPC)) < 0)
		{
		printf("unable to read response\n");
		return(-EIO);
		}
	if (msg.request == DIPCREQ_CON)
		{
		/* We got disconnected.. must resubmit */
		sprintf(log_msg,
						"0:0:LOG:Thread %s connection dropped in transaction. Re-connecting",
						ctx->node_id);
		
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		do_emp_submitpass(ctx);
		return(-EAGAIN);
		}
	else if (msg.request != DIPCREQ_RESP)
		{
		printf("Thread %s expected RESP got %d\n",ctx->node_id, msg.request);
		return(-EIO);
		}
	if (msg.rc != 0)
		return(-rc);

	*response_buffer = malloc(msg.pkt_len);
	memcpy(*response_buffer, &msg.pkt_buffer[0], msg.pkt_len);
	*response_len = msg.pkt_len;
	return(0);
}

/*
 * SDR decode tables and tools
 */

/*
 * Tables and whatnot for looking up english..
 */

struct  ipmb_i2c_device_type_codes
  {
  unsigned  char  code;
  unsigned  char  modifier_type;  /* 00 = no modifiers, 0x1 = bit, 0x2 = byte */
  unsigned  char  modifier_code;
  unsigned  char  *name;
  };

struct  ipmb_i2c_device_type_codes  ipmb_i2c_device_type_table[] = {
  { 0x00, 0x01, 0x00, "IPMB Device (Firmware Transfer Support)" },
  { 0x00, 0x01, 0x01, "IPMB Device (ICMB Bridge)"               },
  { 0x00, 0x01, 0x02, "IPMB Device (Event Generator)"           },
  { 0x00, 0x01, 0x03, "IPMB Device (Primary Event Receiver)"    },
  { 0x00, 0x01, 0x04, "IPMB Device (SDR Repository Device)"     },
  { 0x00, 0x01, 0x05, "IPMB Device (SEL Device)"                },
  { 0x00, 0x01, 0x06, "IPMB Device (Chassis Device)"            },
  { 0x00, 0x01, 0x07, "IPMB Device (FRU Inventory Device)"      },

  { 0x01, 0x01, 0x00, "SMM Card IPMB Device (Firmware Transfer Support)"  },
  { 0x01, 0x01, 0x01, "SMM Card IPMB Device (ICMB Bridge)"                },
  { 0x01, 0x01, 0x02, "SMM Card IPMB Device (Event Receiver)"             },
  { 0x01, 0x01, 0x03, "SMM Card IPMB Device (Reserved)"                   },
  { 0x01, 0x01, 0x04, "SMM Card IPMB Device (Reserved)"                   },
  { 0x01, 0x01, 0x05, "SMM Card IPMB Device (Reserved)"                   },
  { 0x01, 0x01, 0x06, "SMM Card IPMB Device (Reserved)"                   },
  { 0x01, 0x01, 0x07, "SMM Card IPMB Device (Reserved)"                   },

  { 0x02, 0x00, 0x00, "DS1624 Temperature Sensor / EEPROM or Equivalent"  },
  { 0x03, 0x00, 0x00, "DS1621 Temperature Sensor or Equivalent"           },
  { 0x04, 0x00, 0x00, "LM75 Temperature Sensor or Equivalent"             },

  { 0x05, 0x02, 0x00, "Heceta ASIC or equivalent (HECETA 1 [LM 78])"  },
  { 0x05, 0x02, 0x01, "Heceta ASIC or equivalent (HECETA 2 [LM 79])"  },
  { 0x05, 0x02, 0x02, "Heceta ASIC or equivalent (HECETA 3)"          },

  { 0x08, 0x02, 0x00, "EEPROM, 24C01 or equivalent (Use: unspecified)"                },
  { 0x08, 0x02, 0x01, "EEPROM, 24C01 or equivalent (Use: DIMM Memory ID)"             },
  { 0x08, 0x02, 0x02, "EEPROM, 24C01 or equivalent (Use: FRU Inventory)"              },
  { 0x08, 0x02, 0x03, "EEPROM, 24C01 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x09, 0x02, 0x00, "EEPROM, 24C02 or equivalent (Use: unspecified)"                },
  { 0x09, 0x02, 0x01, "EEPROM, 24C02 or equivalent (Use: DIMM Memory ID)"             },
  { 0x09, 0x02, 0x02, "EEPROM, 24C02 or equivalent (Use: FRU Inventory)"              },
  { 0x09, 0x02, 0x03, "EEPROM, 24C02 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0A, 0x02, 0x00, "EEPROM, 24C04 or equivalent (Use: unspecified)"                },
  { 0x0A, 0x02, 0x01, "EEPROM, 24C04 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0A, 0x02, 0x02, "EEPROM, 24C04 or equivalent (Use: FRU Inventory)"              },
  { 0x0A, 0x02, 0x03, "EEPROM, 24C04 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0B, 0x02, 0x00, "EEPROM, 24C08 or equivalent (Use: unspecified)"                },
  { 0x0B, 0x02, 0x01, "EEPROM, 24C08 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0B, 0x02, 0x02, "EEPROM, 24C08 or equivalent (Use: FRU Inventory)"              },
  { 0x0B, 0x02, 0x03, "EEPROM, 24C08 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0C, 0x02, 0x00, "EEPROM, 24C16 or equivalent (Use: unspecified)"                },
  { 0x0C, 0x02, 0x01, "EEPROM, 24C16 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0C, 0x02, 0x02, "EEPROM, 24C16 or equivalent (Use: FRU Inventory)"              },
  { 0x0C, 0x02, 0x03, "EEPROM, 24C16 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0D, 0x02, 0x00, "EEPROM, 24C17 or equivalent (Use: unspecified)"                },
  { 0x0D, 0x02, 0x01, "EEPROM, 24C17 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0D, 0x02, 0x02, "EEPROM, 24C17 or equivalent (Use: FRU Inventory)"              },
  { 0x0D, 0x02, 0x03, "EEPROM, 24C17 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0E, 0x02, 0x00, "EEPROM, 24C32 or equivalent (Use: unspecified)"                },
  { 0x0E, 0x02, 0x01, "EEPROM, 24C32 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0E, 0x02, 0x02, "EEPROM, 24C32 or equivalent (Use: FRU Inventory)"              },
  { 0x0E, 0x02, 0x03, "EEPROM, 24C32 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x0F, 0x02, 0x00, "EEPROM, 24C64 or equivalent (Use: unspecified)"                },
  { 0x0F, 0x02, 0x01, "EEPROM, 24C64 or equivalent (Use: DIMM Memory ID)"             },
  { 0x0F, 0x02, 0x02, "EEPROM, 24C64 or equivalent (Use: FRU Inventory)"              },
  { 0x0F, 0x02, 0x03, "EEPROM, 24C64 or equivalent (Use: System Processor SECC FRU)"  },

  { 0x10, 0x00, 0x00, "Other FRU Inventory Device Behind Management Controller" },

  { 0x14, 0x00, 0x00, "PCF 8570 256 byte RAM or equivalent"                           },
  { 0x15, 0x00, 0x00, "PCF 8573 clock/calendar or equivalent"                         },
  { 0x16, 0x00, 0x00, "PCF 8574A I/O port or equivalent"                              },
  { 0x17, 0x00, 0x00, "PCF 8583 clock/calendar or equivalent"                         },
  { 0x18, 0x00, 0x00, "PCF 8593 clock/calendar or equivalent"                         },
  { 0x19, 0x00, 0x00, "Clock/calendar, type not specified"                            },
  { 0x1A, 0x00, 0x00, "PCF 8591 A/D, D/A Converter or equivalent"                     },
  { 0x1B, 0x00, 0x00, "I/O port, specific device not specified"                       },
  { 0x1C, 0x00, 0x00, "A/D Converter, specific device not specified"                  },
  { 0x1D, 0x00, 0x00, "D/A Converter, specific device not specified"                  },
  { 0x1E, 0x00, 0x00, "A/D,D/A Converter, specific device not specified"              },
  { 0x1F, 0x00, 0x00, "LCD Controller/Driver, specific device not specified"          },
  { 0x20, 0x00, 0x00, "Core Logic Device, specific device not specified"              },
  { 0x21, 0x00, 0x00, "LMC6874 Intelligent Battery Controller or equivalent"          },
  { 0x22, 0x00, 0x00, "Intelligent Battery Controller, specific device not specified" },
  { 0x23, 0x00, 0x00, "Combo Management ASIC, specific device not specified"          },
  { 0xFF, 0xFF, 0xFF, NULL}};


struct  device_location_codes
  {
  unsigned  char  code;
  char            *name;
  };

static struct  device_location_codes device_location_table[] = {
  { 0x00, "Reserved"                  },
  { 0x01, "Other"                     },
  { 0x02, "Unknown (Unspecified)"     },
  { 0x03, "Processor"                 },
  { 0x04, "Disk or Disk Bay"          },
  { 0x05, "Peripheral Bay"            },
  { 0x06, "System Management Module"  },
  { 0x07, "System Board"              },
  { 0x08, "Memory Module"             },
  { 0x09, "Processor Module"          },
  { 0x0A, "Power Unit"                },
  { 0x0B, "Add-in Card"               },
  { 0x0C, "Front Panel Board"         },
  { 0x0D, "Back Panel Board"          },
  { 0x0E, "Drive Backplane"           },
  { 0xFF, NULL}};

struct  entity_id_codes
  {
  unsigned  char  entity_id;
  char            *name;
  };

static struct entity_id_codes  entity_id_table[] = {
  { 0x00, "ID Unspecified"              },
  { 0x01, "Main System Board"           },
  { 0x02, "System Board 2"              },
  { 0x03, "System Board 3"              },
  { 0x04, "System Board 4"              },
  { 0x08, "System Expansion Board 1"    },
  { 0x09, "System Expansion Board 2"    },
  { 0x0A, "System Expansion Board 3"    },
  { 0x0B, "System Expansion Board 4"    },
  { 0x10, "Processor Board 1"           },
  { 0x11, "Processor Board 2"           },
  { 0x12, "Processor Board 3"           },
  { 0x13, "Processor Board 4"           },
  { 0x18, "Processor 1"                 },
  { 0x19, "Processor 2"                 },
  { 0x1A, "Processor 3"                 },
  { 0x1B, "Processor 4"                 },
  { 0x1C, "Processor 5"                 },
  { 0x1D, "Processor 6"                 },
  { 0x1E, "Processor 7"                 },
  { 0x1F, "Processor 8"                 },
  { 0x20, "Memory Module 1"             },
  { 0x21, "Memory Module 2"             },
  { 0x22, "Memory Module 3"             },
  { 0x23, "Memory Module 4"             },
  { 0x28, "Disk Backplane 1"            },
  { 0x29, "Disk Backplane 2"            },
  { 0x2A, "Disk Backplane 3"            },
  { 0x2B, "Disk Backplane 4"            },
  { 0x2C, "Power Supply 1"              },
  { 0x2D, "Power Supply 2"              },
  { 0x2E, "Power Supply 3"              },
  { 0x2F, "Power Supply 4"              },
  { 0x30, "Power Supply 5"              },
  { 0x31, "Power Supply 6"              },
  { 0x32, "Power Supply 7"              },
  { 0x33, "Power Supply 8"              },
  { 0x34, "Power Management Board 1"    },
  { 0x35, "Power Management Board 2"    },
  { 0x36, "Power Management Board 3"    },
  { 0x37, "Power Management Board 4"    },
  { 0x3C, "Power Module 1"              },
  { 0x3D, "Power Module 2"              },
  { 0x3E, "Power Module 3"              },
  { 0x3F, "Power Module 4"              },
  { 0x44, "System Management Card 1"    },
  { 0x45, "System Management Card 2"    },
  { 0x46, "System Management Card 3"    },
  { 0x47, "System Management Card 4"    },
  { 0x4C, "Chassis Board 1"             },
  { 0x4D, "Chassis Board 2"             },
  { 0x4E, "Chassis Board 3"             },
  { 0x4F, "Chassis Board 4"             },
  { 0x54, "Chassis Front Panel Board 1" },
  { 0x55, "Chassis Front Panel Board 2" },
  { 0x56, "Chassis Front Panel Board 3" },
  { 0x57, "Chassis Front Panel Board 4" },
  { 0x5C, "Chassis Back Panel Board 1"  },
  { 0x5D, "Chassis Back Panel Board 2"  },
  { 0x5E, "Chassis Back Panel Board 3"  },
  { 0x5F, "Chassis Back Panel Board 4"  },
  { 0x64, "Chassis Management Board 1"  },
  { 0x65, "Chassis Management Board 2"  },
  { 0x66, "Chassis Management Board 3"  },
  { 0x67, "Chassis Management Board 4"  },
  { 0x6C, "Disk Drive Bay 1"            },
  { 0x6D, "Disk Drive Bay 2"            },
  { 0x6E, "Disk Drive Bay 3"            },
  { 0x6F, "Disk Drive Bay 4"            },
  { 0x70, "Disk Drive Bay 5"            },
  { 0x71, "Disk Drive Bay 6"            },
  { 0x72, "Disk Drive Bay 7"            },
  { 0x73, "Disk Drive Bay 8"            },
  { 0x74, "Disk Drive Bay 9"            },
  { 0x75, "Disk Drive Bay 10"           },
  { 0x76, "Disk Drive Bay 11"           },
  { 0x77, "Disk Drive Bay 12"           },
  { 0x78, "Disk Drive Bay 13"           },
  { 0x79, "Disk Drive Bay 14"           },
  { 0x7A, "Disk Drive Bay 15"           },
  { 0x7B, "Disk Drive Bay 16"           },
  { 0x7C, "Peripheral Bay 1"            },
  { 0x7D, "Peripheral Bay 2"            },
  { 0x7E, "Peripheral Bay 3"            },
  { 0x7F, "Peripheral Bay 4"            },
  { 0x80, "Peripheral Bay 5"            },
  { 0x81, "Peripheral Bay 6"            },
  { 0x82, "Peripheral Bay 7"            },
  { 0x83, "Peripheral Bay 8"            },
  { 0xFF, NULL                          }};

#define STRING_DATA_TYPE_BINARY         0x00
#define STRING_DATA_TYPE_BCD_PLUS       0x01
#define STRING_DATA_TYPE_SIX_BIT_ASCII  0x02
#define STRING_DATA_TYPE_LANG_DEPENDANT 0x03

static void decode_string(unsigned char type,
                   				unsigned char language_code,
                   				unsigned char *source,
                   				char					*target,
                   				int						size)
{
  unsigned char *s = &source[0];
  unsigned char *d = &target[0];

  memset(target, 0, size);
  if (type == STRING_DATA_TYPE_BCD_PLUS)
    {
    while(size)
      {
//   if ((*s >= (unsigned char) 0x00) && (*s <= (unsigned char) 0x09))
      if (*s <= (unsigned char) 0x09)
        *d++ = (unsigned char) 0x30 + (unsigned char) *s++;
      else
        {
        if (*s == (unsigned char) 0x0A)
          *d++ = (unsigned char) ' ';
        else if (*s == (unsigned char) 0x0B)
          *d++ = (unsigned char) '-';
        else if (*s == (unsigned char) 0x0C)
          *d++ = (unsigned char) '.';
        else if ((*s <= (unsigned char) 0x0D) && (*s <= (unsigned char) 0x0F))
          {
          *d++ = (unsigned char) '*';
          }
        s++;
        }
      size --;
      }
    }
  else if (type == STRING_DATA_TYPE_SIX_BIT_ASCII)
    {
		printf("Six bit ASCII decode not supported\n");
    }
  else if (type == STRING_DATA_TYPE_LANG_DEPENDANT)
    {
    if ((language_code == 0x00) || (language_code == 0x25))
      {
      strncpy(target, source, size);
      target[size] = 0x0;
      }
    else
      {
			printf("Language 0x%x dependant decode not supported\n",language_code);
      return;
      }
    }
  else
    {
		sprintf(target,"Unknown Language Type 0x%.2x",type);
    return;
    }
}

static SENSOR	*sensor_alloc(CONTEXT *ctx)
{
	SENSOR	*new;
	SENSOR	*scan;

  if (!(new = (SENSOR *) malloc(sizeof(SENSOR))))
    return(NULL);
  memset(new, 0, sizeof(SENSOR));

  if (!ctx->sensor_list)
    ctx->sensor_list = new;
  else
    {
		scan = ctx->sensor_list;
	
		while(scan->next)
			scan = scan -> next;
		scan->next = new;
    }
  return(new);
}

static void sensor_free(CONTEXT *ctx, SENSOR *sensor)
{
	SENSOR	*scan;

  if (sensor == ctx->sensor_list)
    {
		ctx->sensor_list = ctx->sensor_list->next;
    }
  else
    {
		scan = ctx->sensor_list;
		while(scan->next !=sensor)
			scan = scan ->next;
		scan->next = sensor->next;
    }
	free(sensor);
}

/*
 * Looks up the proper type_code and type_modifier in the ipmb i2c device
 * type table and returns the type. If multiple types are available, they
 * are returned '*' separated
 */
int return_device_type_entry(unsigned char type_code,
                             unsigned char type_modifier,
                             char          *buffer)
{
  int i;

  for (i=0; ipmb_i2c_device_type_table[i].code != 0xFF; i++)
    {
    if (ipmb_i2c_device_type_table[i].code == type_code)
      {
      if (ipmb_i2c_device_type_table[i].modifier_type == 0)
        {
        strcpy(buffer, ipmb_i2c_device_type_table[i].name);
        return(1);
        }
      else if (ipmb_i2c_device_type_table[i].modifier_type == 1)
        {
        unsigned char bitmask;

        strcpy(buffer, "");
        for (; ipmb_i2c_device_type_table[i].code == type_code; i++)
          {
          bitmask = 0x01 << ipmb_i2c_device_type_table[i].modifier_code;
          if (bitmask & type_modifier)
            {
            strcat(buffer, ipmb_i2c_device_type_table[i].name);
            strcat(buffer, "*");
            }
          }
        if (buffer[0] == (char) 0x00)
          {
          printf("return_device_type_entry(): T 0x%x, B mod mask 0x%x not found\n",
                 type_code,
                 type_modifier);
          return(0);
          }
        else
          {
          buffer[(strlen(buffer)-1)] = 0x00;
          }
        return(1);
        }
      else if (ipmb_i2c_device_type_table[i].modifier_type == 2)
        {
        for (; ipmb_i2c_device_type_table[i].code == type_code; i++)
          {
          if (ipmb_i2c_device_type_table[i].modifier_code == type_modifier)
            {
            strcpy(buffer, ipmb_i2c_device_type_table[i].name);
            return(1);
            }
          printf("return_device_type_entry(): T 0x%x, B mod 0x%x not found\n",
                 type_code,
                 type_modifier);
          return(0);
          }
        }
      else
        {
        printf("return_device_type_entry(): T 0x%x has bad M type 0x%x\n",
               ipmb_i2c_device_type_table[i].code,
               ipmb_i2c_device_type_table[i].modifier_type);
        return(0);
        }
      }
    }
  return(0);
}

static char *reading_type_to_str(unsigned char value)
{
  switch (value)
    {
    case NO_PRESENT_READING_SUPPORT:
      return("NO_PRESENT_READING_SUPPORT");
    case PRESENT_READING_TYPE_NOT_SPECIFIED:
      return("PRESENT_READING_TYPE_NOT_SPECIFIED");
    case DIGITAL_DISCRETE:
      return("DIGITAL_DISCRETE");
    case ANALOG:
      return("ANALOG");
    default:
      return("UNKNOWN");
    }
}

static char *linearization_to_str(unsigned char value)
{
  switch (value)
    {
    case LINEARIZATION_LINEAR:
      return("LINEAR");
    case LINEARIZATION_LN:
      return("LN");
    case LINEARIZATION_LOG10:
      return("LOG10");
    case LINEARIZATION_LOG2:
      return("LOG2");
    case LINEARIZATION_E:
      return("E");
    case LINEARIZATION_EXP10:
      return("EXP10");
    case LINEARIZATION_EXP2:
      return("EXP2");
    case LINEARIZATION_1_DIV_X:
      return("1 DIV X");
    case LINEARIZATION_SQR_X:
      return("SQR X");
    case LINEARIZATION_CUBE_X:
      return("CUBE X");
    case LINEARIZATION_SQRT_X:
      return("SQRT X");
    case LINEARIZATION_CUBE_MINUS_1_X:
      return("CUBE MINUS 1 X");
    default:
      return("UNKNOWN");
    }
}

double linearize_sensor_reading(SENSOR *slist, unsigned char reading)
{
  struct bit10 {
	int bits : 10;
  };
  struct bit10 M, B;
  char K1 = 0, K2 = 0;
#if 0
  char            M = 0, B = 0, K1 = 0, K2 = 0;
#endif
  double          pre_lin;
  unsigned char   x;
  unsigned char   function;

  x = reading;
  if (slist->rec_header.record_type == 0x01)
    {
    M.bits =  slist->sensor_type.type_01_sensor.M_ms2bits;
    M.bits =  M.bits << 8;
    M.bits |= (slist->sensor_type.type_01_sensor.M_ls8bits & 0xff);

    B.bits =  slist->sensor_type.type_01_sensor.B_ms2bits;
    B.bits =  B.bits << 8;
    B.bits |= (slist->sensor_type.type_01_sensor.B_ls8bits & 0xff);

    K1 = slist->sensor_type.type_01_sensor.B_exp;
    K2 = slist->sensor_type.type_01_sensor.R_exp;

    function = slist->sensor_type.type_01_sensor.linearization;
    }
  else
    {
    printf("linearize_sensor_reading(): record type 0x%x not supported\n",
            slist->rec_header.record_type);
    return(0);
    }
//  printf("x=%d, m=%d, b=%d, k1=%d, k2=%d, linear=%d\n",x,M.bits,B.bits,K1,K2,function);

  pre_lin = ((M.bits * x) + (B.bits * pow(10, K1))) * pow(10, K2);
  switch(function)
    {
    case LINEARIZATION_LINEAR:
      break;
    default:
      printf("linearize_sensor_reading(): linearization type %x not supported\n",function);
      return(0);
    }
  return(pre_lin);
}

static int return_num_processed_sensors(CONTEXT *ctx)
{
  SENSOR *sensor_scan =ctx->sensor_list;
  int         cnt=0;

  while(sensor_scan)
    {
    cnt++;
    sensor_scan = sensor_scan->next;
    }
  return(cnt);
}


/*
 * SEL support stuff
 */

static int do_emp_downloadsel(CONTEXT *ctx)
{
	SEL_RESERVATION_RESP	*sel_res_resp;
	SEL_CLEAR_REQ					sel_clear_req;
	SEL_INFO_RESP					*sel_info_resp;
	SEL_REQ								sel_req;
	SEL_RESP							*sel_resp;
	unsigned char					*response;
	int										response_len;
	int										num_entries;
	char									*sel_string;
	unsigned short				reservation_id;
	int										rc, i;
	char									log_msg[255];
	unsigned short				sel_records[255];

#ifdef SEL_DEBUG
	printf("Thread %s downloading SEL.\n",ctx->node_id);
#endif
	for (i=0; i < 255; i++)
		sel_records[i] = 0x0000;

	while(1)
		{
#ifdef SEL_DEBUG
		printf("Thread %s reserving SEL.\n",ctx->node_id);
#endif
		/* First reserve the SEL */
		for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
							 							 	 	 	 CMD_RESERVE_SEL,
							 							 	 	 	 STORAGE_REQUEST,
							 							 	 	 	 0x00,
							 							 	 	 	 EMP_SLAVE_ADDRESS,
							 							 	 	 	 EMP_APPLICATION_SLAVE_ADDRESS,
							 							 	 	 	 NULL,
							 							 	 	 	 0,
														 	 	 	 &response,
														 	 	 	 &response_len)) < 0)
				return(rc);
			if (response[6] == (unsigned char) 0x00)
				break;
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s reserve_sel failed. Retries exceeded",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-EREMOTEIO);
			}
		sel_res_resp = (SEL_RESERVATION_RESP *) &response[6];
		reservation_id = sel_res_resp->res_id;
		free(response);

#ifdef SEL_DEBUG
		printf("Thread %s SEL reservation ID 0x%.4x.\n",ctx->node_id,reservation_id);
#endif

		/* Find out how many entries */
		for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
							 							 	 	 	CMD_GET_SEL_INFO,
							 							 	 	 	STORAGE_REQUEST,
							 							 	 	 	0x00,
							 							 	 	 	EMP_SLAVE_ADDRESS,
							 							 	 	 	EMP_APPLICATION_SLAVE_ADDRESS,
							 							 	 	 	NULL,
							 							 	 	 	0,
														 	 	 	&response,
														 	 	 	&response_len)) < 0)
				return(rc);
			if (response[6] == (unsigned char) 0x00)
				break;
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s get_sel_info failed. Retries exceeded",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-EREMOTEIO);
			}
		sel_info_resp = (SEL_INFO_RESP *) &response[6];
		num_entries = sel_info_resp->num_entries;
		free(response);

#ifdef SEL_DEBUG
		printf("Thread %s SEL contains %d entries.\n",ctx->node_id,num_entries);
#endif

		if (num_entries == 0)
			return(0);

		/* Now download the SEL entries */
		/* First SEL entry */
		sel_req.resid = 0x0000;
		sel_req.recid = 0x0000;
		sel_req.offset = 0x00;
		sel_req.length = SEL_LENGTH_ALL;
		while(1)
			{
			for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
				{
#ifdef SEL_DEBUG
				printf("Thread %s SEL entry DL (res 0x%.4x, rec 0x%.4x).\n",
								ctx->node_id,
								reservation_id,
								sel_req.recid);
#endif
				if ((rc = do_emp_transaction(ctx,
																		CMD_GET_SEL_ENTRY,
							 							 	 	 		STORAGE_REQUEST,
							 							 	 	 	 	0x00,
							 							 	 	 	 	EMP_SLAVE_ADDRESS,
							 							 	 	 	 	EMP_APPLICATION_SLAVE_ADDRESS,
							 							 	 	 	 	(unsigned char *) &sel_req,
							 							 	 	 	 	sizeof(sel_req),
														 	 	 	 	&response,
														 	 	 	 	&response_len)) < 0)
					return(rc);
				if (response[6] == (unsigned char) 0x00)
					break;
				free(response);
				usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
				}
			if (i == EMP_HIGH_LEVEL_RETRY_MAX)
				{
				sprintf(log_msg,
								"0:0:LOG:Thread %s get_sel_entry failed. Retries exceeded\n",
								ctx->node_id);
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
				return(-EREMOTEIO);
				}
			sel_resp = (SEL_RESP *) &response[6];
			/* 
		 	 * Englishify the SEL event, and send it to our log, 
		 	 * and any watchers that may exist 
		 	 */
			if (response[6] == 0x00)
				{
				int		hunt;
				int		exists = 0;
				FILE	*Tfilep_ptr;
				int		slot;
				char	nexxus_msg[255];

				/* Check to see if we've already downloaded this entry */
				for (hunt = 0; hunt < 255; hunt++)
					if (sel_records[hunt] == sel_resp->recid)
						{
						exists = 1;
						break;
						}
				if (!exists)
					{
					/* Store the record id in case we download it again */
					for (hunt = 0; hunt < 255; hunt++)
						if (sel_records[hunt] == 0x0000)
							break;
					sel_records[hunt] = sel_resp->recid;

					/* Parse the entry */
					sel_string = parse_sel_entry(ctx, sel_resp);
					if (!(Tfilep_ptr = open_sel(ctx)))
						{
						sprintf(log_msg,
										"0:0:LOG:Thread %s unable to open log (%m)\n",
										ctx->node_id);
						send_msg(ctx,ctx->thread_ipc_write, log_msg);
						}
					else
						{
						fprintf(Tfilep_ptr, "%s\n",sel_string);
						close_sel(Tfilep_ptr);
						}
					/* Send msg to any watchers */
					for (slot = 0; slot < MAX_SEL_WATCHERS; slot ++)
						if (ctx->sel_watchers[slot] != -1)
							{
							sprintf(nexxus_msg,
											"0:%d:FOR:%s",
											ctx->sel_watchers[slot],
											sel_string);
							send_msg(ctx,ctx->thread_ipc_write, nexxus_msg);
							}
					free(sel_string);
					}
				else
					{
#ifdef SEL_DEBUG
					printf("Thread %s allready downloaded SEL record 0x%.4x.\n",
								 ctx->node_id,
								 sel_resp->recid);
#endif
					}
				}
			num_entries--;
			if (sel_resp->next_id == (unsigned short) 0xffff)
				break;
			else
				{
				sel_req.resid = 0x0000;
				sel_req.recid = sel_resp->next_id;
				sel_req.offset = 0x00;
				sel_req.length = SEL_LENGTH_ALL;
				}
			free(response);
			} /* End of download complete SEL loop */

#ifdef SEL_DEBUG
		printf("Thread %s erasing SEL.\n",ctx->node_id);
#endif

		/* Now erase the sel */
		memset(&sel_clear_req, 0, sizeof(sel_clear_req));
		sel_clear_req.res_id = reservation_id;
 		sel_clear_req.clr0=0x43;
 		sel_clear_req.clr1=0x4c;
 		sel_clear_req.clr2=0x52;
 		sel_clear_req.op=0xaa; /* Initiate erase */
		for (i = 0; i < EMP_HIGH_LEVEL_RETRY_MAX; i++)
			{
			if ((rc = do_emp_transaction(ctx,
					 							 	 	 	 	 CMD_CLEAR_SEL,
					 							 	 	 	 	 STORAGE_REQUEST,
					 							 	 	 	 	 0x00,
					 							 	 	 	 	 EMP_SLAVE_ADDRESS,
					 							 	 	 	 	 EMP_APPLICATION_SLAVE_ADDRESS,
					 							 	 	 	 	 (unsigned char *)&sel_clear_req,
					 							 	 	 	 	 sizeof(sel_clear_req),
												 	 	 	 	 &response,
												 	 	 	 	 &response_len)) < 0)
				return(rc);
			if (response[6] == (unsigned char) 0x00)
				break;
			else if (response[6] == (unsigned char) 0xc5)
				break; /* Reservation expired... redownload time */
			sprintf(log_msg,
							"0:0:LOG:Thread %s clear_sel failed. (0x%.2x)",
							ctx->node_id,
							response[6]);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			free(response);
			usleep(EMP_HIGH_LEVEL_RETRY_DELAY);
			}
		if (i == EMP_HIGH_LEVEL_RETRY_MAX)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s clear_sel failed. Retries exceeded",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			return(-EREMOTEIO);
			}
		else
			{
			if (response[6] == 0xc5)
				{
#ifdef SEL_DEBUG
				printf("Thread %s SEL reservation expired.\n",ctx->node_id);
#endif
				free(response);
				continue; /* Try again */
				}
#ifdef SEL_DEBUG
			printf("Thread %s SEL download and erasure complete.\n",ctx->node_id);
#endif
			return(0); /* AOK */
			}
		}
}

/*
 * Structures and constants for converting SEL entry data into English
 */
struct posterror
  {
  unsigned char code0;
  unsigned char code1;
  char          *msg;
  };
static struct posterror posterrlist[] = {
  { 0x62, 0x01, "BIOS unable to apply BIOS update to processor 1"        },
  { 0x63, 0x01, "BIOS unable to apply BIOS update to processor 2"        },
  { 0x64, 0x01, "BIOS does not support current stepping for processor 1" },
  { 0x65, 0x01, "BIOS does not support current stepping for processor 2" },

  { 0x00, 0x02, "Failure Fixed Disk"                  },
  { 0x10, 0x02, "Stuck Key"                           },
  { 0x11, 0x02, "Keyboard error"                      },
  { 0x12, 0x02, "Keyboard Controller Failed"          },
  { 0x13, 0x02, "Keyboard locked"                     },
  { 0x20, 0x02, "Monitor type does not match CMOS"    },
  { 0x30, 0x02, "System RAM Failed at offset"         },
  { 0x31, 0x02, "Shadow RAM Failed at offset"         },
  { 0x32, 0x02, "Extended RAM Failed at offset"       },
  { 0x50, 0x02, "System battery is dead"              },
  { 0x51, 0x02, "System CMOS checksum bad"            },
  { 0x60, 0x02, "System timer error"                  },
  { 0x70, 0x02, "Real time clock error"               },
  { 0x97, 0x02, "ECC Memory error in bank"            },
  { 0xb2, 0x02, "Incorrect Drive A: type"             },
  { 0xb3, 0x02, "Incorrect Drive B: type"             },
  { 0xd0, 0x02, "SYstem cache error - Cache disabled" },
  { 0xf5, 0x02, "DMA Test Failed"                     },
  { 0xf6, 0x02, "Software NMI Failed"                 },

  { 0x01, 0x04, "Invalid System Configuration Data" },
  { 0x03, 0x04, "Resource Conflict"                 },
  { 0x04, 0x04, "Resource Conflict"                 },
  { 0x05, 0x04, "Expansion ROM not initialized"     },
  { 0x06, 0x04, "Warning: IRQ not configured"       },

  { 0x04, 0x05, "Resource Conflict"             },
  { 0x05, 0x05, "Expansion ROM not initialized" },
  { 0x06, 0x05, "Warning: IRQ not configured"   },

  { 0x01, 0x06, "Device configuration changed"          },
  { 0x02, 0x06, "Configuration error - device disabled" },

  { 0x00, 0x81, "processor 0 failed BIST"                            },
  { 0x01, 0x81, "processor 1 failed BIST"                            },
  { 0x04, 0x81, "processor 0 Internal Error (IERR) failure"          },
  { 0x05, 0x81, "processor 1 Internal Error (IERR) failure"          },
  { 0x06, 0x81, "processor 0 Thermal Trip failure"                   },
  { 0x07, 0x81, "processor 1 Thermal Trip failure"                   },
  { 0x08, 0x81, "Watchdog Timer failed on last boot, BSP switched"   },
  { 0x0a, 0x81, "processor 1 failed initialization on last boot"     },
  { 0x0b, 0x81, "processor 0 failed initialization on last boot"     },
  { 0x0c, 0x81, "processor 0 disabled, system in Uni-processor mode" },
  { 0x0d, 0x81, "processor 1 disabled, system in Uni-processor mode" },
  { 0x0e, 0x81, "processor 0 failed FRB Level 3 timer"               },
  { 0x0f, 0x81, "processor 1 failed FRB Level 3 timer"               },
  { 0x10, 0x81, "Server Management Interface failed to function"     },
  { 0x20, 0x81, "IOP sub-system is not functional"                   },
  { 0x50, 0x81, "NVRAM (CMOS) Cleared by Jumper"                     },
  { 0x51, 0x81, "NVRAM (CMOS) Checksum Error, NVRAM cleared"         },
  { 0x52, 0x81, "NVRAM (CMOS) Data Invalid, NVRAM cleared"           },
  { 0x00, 0x00, NULL}};

/*
 * Converts a binary SEL entry into the English equivalents
 * Note: Caller must free buffer 
 */
static char *parse_sel_entry(CONTEXT *ctx, SEL_RESP *entry)
{
  SENSOR	*slist;
  char   	*buffer;
  int    	i;
  char   	record_id[8];
  char   	type[32];
  char   	time[32];
  char   	generator_id[32];
  char   	emvrev[8];
  char   	*sensortype;
  char   	sensorname[64];
  char   	*descp = NULL;
  char		description[64];
	char		log_msg[255];

	memset(description, 0, sizeof(description));
  sprintf(record_id, "0x%.4x", entry->recid);
  if (entry->type == 0x02)
    strcpy(type, "System Event Record");
  else
    sprintf(type, "Unknown (0x%.2x)", entry->type);
  sprintf(time, "0x%.8x",entry->timestamp);
  
  /*
   * Englishify the generator ID fields
   */
  if (entry->id_type == 0)
    sprintf(generator_id, "IPMB addr 0x%x (LUN %d)",
            entry->id_data,
            entry->ipmb_lun);
  else      
    sprintf(generator_id, "SSID 0x%x",entry->id_data);
    
  sprintf(emvrev, "0x%.2x", entry->evmrev);
  
  /*
   * Identify the type of sensor (temp, fan, volt, etc)
   */
  sensortype = sensor_get_type(entry->sensortype);
  
  /*
   * Try to lookup the actual sensor (proc 1 fan, +3.3 volt, etc)
   */
  slist = ctx->sensor_list;
  while(slist)
    {
    if (slist->rec_header.record_type == 0x01)
      {
      if (entry->sensornum==slist->sensor_type.type_01_sensor.sensor_number)
        {
        decode_string(slist->sensor_type.type_01_sensor.id_string_type_code,
                      0,
                      &slist->sensor_type.type_01_sensor.id_string[0],
                      &sensorname[0],
                      slist->sensor_type.type_01_sensor.id_string_num_bytes);
        break;        
        }
      } 
    else if (slist->rec_header.record_type == 0x02)
      {
      if (entry->sensornum==slist->sensor_type.type_02_sensor.sensor_number)
        {
        decode_string(slist->sensor_type.type_02_sensor.id_string_type_code,
                      0,
                      &slist->sensor_type.type_02_sensor.id_string[0],
                      &sensorname[0],
                      slist->sensor_type.type_02_sensor.id_string_num_bytes);
        break;        
        }
      } 
    slist = slist->next;
    }
  if (!slist)
    strcpy(sensorname, "Unknown Sensor Name");
    
/*      
   * If it's a post err, display the error code
   */
  if (entry->sensortype == 0x0f) /* POST error */
    {
    for (i=0; posterrlist[i].code0 != 0x00; i++)
      { 
      if ((posterrlist[i].code0==entry->event_data[1])&&
          (posterrlist[i].code1==entry->event_data[2]))
        { 
        strcpy(description,posterrlist[i].msg);
        break;
        }
      } 
    if (posterrlist[i].code0 == 0x00)
      {
      sprintf(description,"0x%.2x 0x%.2x",entry->event_data[1], entry->event_data[0]);
      }
    } 
  if ((!descp) && (description[0] == 0x0))
    {
    int sensor_class;
    
    /* There was no 'canned' description for this entry */
    sensor_class = sensor_get_class(entry->event_type); 
    if (sensor_class == SENSOR_CLASS_THRESHOLD)
      {
#if 1
      THRESHOLD_EVENT_DATA  *event;
			event = (THRESHOLD_EVENT_DATA *) entry->event_data;
      descp = sensor_get_trigger_description(entry->event_type,
                                            event->event_trigger_offset);
#else
      descp = NULL;                         
#endif                                      
      }
    else if (sensor_class == SENSOR_CLASS_DIGITAL)
      {
#if 1  
      DIGITAL_EVENT_DATA  *event;
			event = (DIGITAL_EVENT_DATA *) entry->event_data;
      descp = sensor_get_trigger_description(entry->event_type,
                                            event->digital_event_offset);
#else
      descp = NULL;                         
#endif                                      
      }
    else if (sensor_class == SENSOR_CLASS_DISCRETE)
      {
#if 1 
      DISCRETE_EVENT_DATA *event;
			event = (DISCRETE_EVENT_DATA *) entry->event_data;
      descp = sensor_get_trigger_description(entry->event_type,
                                            event->event_trigger_offset);
#else
      descp = NULL;                         
#endif                                      
      }
    } 
    
  if (!(buffer = (char *) malloc(512)))
    {
    sprintf(log_msg,
						"0:0:LOG:Thread %s unable to allocate memory for sel string buffer",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
    return(0);
    }
  sprintf(buffer,"SEL:%s:%s:%s:%s:%s:%s:%s:%s:%s",
          (ctx->fru_data.board_serial ? ctx->fru_data.board_serial :
					"Unknown Serial"),
          record_id,
          type,
          time,
          generator_id,
          emvrev,
          sensortype,
          (sensorname != NULL ? sensorname : "Unknown Sensor Name"),
          ((descp != (char *) NULL ? descp : 
           (description[0] != (char) NULL ? description : "No Description Available"))));
           
  return(buffer);
}

/*
 * Functions for the SEL logging facilities
 */

static FILE *open_sel(CONTEXT	*ctx)
{
	FILE	*Tfilep_ptr;
	char	logfile[255];
	char	log_msg[255];

	sprintf(logfile,
					"%s/vacm_emp_sel_log.%s",
					LOGDIR,
					ctx->node_id);
	if (!(Tfilep_ptr = fopen(logfile, "a+")))
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to open disk based SEL",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return(NULL);
		}
	return(Tfilep_ptr);
}

static void close_sel(FILE	*Tfilep_ptr)
{
	if (Tfilep_ptr)
		fclose(Tfilep_ptr);
	return;
}

/* Sensor event stuff */
struct sensor_event_descriptions
  {
  unsigned char code;
  unsigned char sensor_class;
  unsigned char offset;
  char          *desc;
  };

static struct sensor_event_descriptions event_desc_table[] = {
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x00, "Lower Non-critical - going low"      },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x01, "Lower Non-critical - going high"   },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x02, "Lower Critical - going low"          },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x03, "Lower Critical - going high"       },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x04, "Lower Non-recoverable - going low" },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x05, "Lower Non-recoverable - going high"  },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x06, "Upper Non-critical - going low"      },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x07, "Upper Non-critical - going high"   },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x08, "Upper Critical - going low"          },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x09, "Upper Critical - going high"       },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x0A, "Upper Non-recoverable - going low" },
  { 0x01, SENSOR_CLASS_THRESHOLD, 0x0B, "Upper Non-recoverable - going low" },

  { 0x02, SENSOR_CLASS_DISCRETE,  0x00, "Transition to Idle"    },
  { 0x02, SENSOR_CLASS_DISCRETE,  0x01, "Transition to Active"  },
  { 0x02, SENSOR_CLASS_DISCRETE,  0x02, "Transition to Busy"    },

  { 0x03, SENSOR_CLASS_DIGITAL, 0x00, "State Deasserted"  },
  { 0x03, SENSOR_CLASS_DIGITAL, 0x01, "State Asserted"    },

  { 0x04, SENSOR_CLASS_DIGITAL, 0x00, "Predictive Failure Deasserted" },
  { 0x04, SENSOR_CLASS_DIGITAL, 0x01, "Predictive Failure Asserted"   },

  { 0x05, SENSOR_CLASS_DIGITAL, 0x00, "Limit Not Exceeded"  },
  { 0x05, SENSOR_CLASS_DIGITAL, 0x01, "Limit Exceeded"      },

  { 0x06, SENSOR_CLASS_DIGITAL, 0x00, "Performance Met" },
  { 0x06, SENSOR_CLASS_DIGITAL, 0x01, "Performance Lags"  },

  { 0x07, SENSOR_CLASS_DISCRETE, 0x00, "Transition to OK"                                 },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x01, "Transition to Non-Critical from OK"             },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x02, "Transition to Critical from less severe"          },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x03, "Transition to Non-recoverable from less severe" },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x04, "Transition to Non-Critical from more severe"      },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x05, "Transition to Critical from Non-recoverable"      },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x06, "Transition to Non-recoverable"                    },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x07, "Monitor"                                        },
  { 0x07, SENSOR_CLASS_DISCRETE, 0x08, "Informational"                                  },

  { 0x08, SENSOR_CLASS_DIGITAL, 0x00, "Device Removed"    },
  { 0x08, SENSOR_CLASS_DIGITAL, 0x01, "Device Inserted" },

  { 0x09, SENSOR_CLASS_DIGITAL, 0x00, "Device Disabled" },
  { 0x09, SENSOR_CLASS_DIGITAL, 0x01, "Device Enabled"    },

  { 0x0A, SENSOR_CLASS_DISCRETE, 0x00, "Transition to Running"        },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x01, "Transition to In Test"        },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x02, "Transition to Power Off"      },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x03, "Transition to On Line"        },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x04, "Transition to Off Line"     },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x05, "Transition to Off Duty"     },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x06, "Transition to Degraded"     },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x07, "Transition to Power Save"   },
  { 0x0A, SENSOR_CLASS_DISCRETE, 0x08, "Transition to Install Error"  },

  { 0x0B, SENSOR_CLASS_DISCRETE, 0x00, "Redundancy Regained"  },
  { 0x0B, SENSOR_CLASS_DISCRETE, 0x01, "Redundancy Lost"      },
  { 0x0B, SENSOR_CLASS_DISCRETE, 0x02, "Redundancy Degraded"  },
  { 0x00, SENSOR_CLASS_UNKNOWN, 0x00, NULL}};


int sensor_get_class(unsigned char sensor_reading_type_code)
{
  int i = 0;

  while (event_desc_table[i].desc != NULL)
    {
    if (event_desc_table[i].code == sensor_reading_type_code)
      return(event_desc_table[i].sensor_class);
    i++;
    }
  return(SENSOR_CLASS_UNKNOWN);
}

char *sensor_get_trigger_description(unsigned char sensor_reading_type_code,
                                   unsigned char offset)
{
  int i = 0;

  while(event_desc_table[i].desc != NULL)
    {
    if ((event_desc_table[i].code == sensor_reading_type_code) &&
        (event_desc_table[i].offset == offset))
      return(event_desc_table[i].desc);
    i++;
    }
  return(NULL);
}

struct sensor_unit_type
  {
  unsigned char code;
  unsigned char *unit;
  };

static struct sensor_unit_type sensorunittypelist[] = {
  { 0, "unspecified"      },
  { 1, "degrees C"      },
  { 2, "degrees F"      },
  { 3, "degrees K"      },
  { 4, "Volts"      },
  { 5, "Amps"     },
  { 6, "Watts"      },
  { 7, "Joules"     },
  { 8, "Coulombs"     },
  { 9, "VA"     },
  { 10, "Nits"      },
  { 11, "lumen"     },
  { 12, "lux"     },
  { 13, "Candela"     },
  { 14, "kPa"     },
  { 15, "PSI"     },
  { 16, "Newton"      },
  { 17, "CFM"     },
  { 18, "RPM"     },
  { 19, "Hz"      },
  { 20, "microsecond"     },
  { 21, "millisecond"     },
  { 22, "second"      },
  { 23, "minute"      },
  { 24, "hour"      },
  { 25, "day"     },
  { 26, "week"      },
  { 27, "mil"     },
  { 28, "inches"      },
  { 29, "feet"      },
  { 30, "cu in"     },
  { 31, "cu feet"     },
  { 32, "mm"      },
  { 33, "cm"      },
  { 34, "m"     },
  { 35, "cu cm"     },
  { 36, "cu m"      },
  { 37, "liters"      },
  { 38, "fluid ounce"     },
  { 39, "radians"     },
  { 40, "steradians"      },
  { 41, "revolutions"     },
  { 42, "cycles"      },
  { 43, "gravities"     },
  { 44, "ounce"     },
  { 45, "pound"     },
  { 46, "ft-lb"     },
  { 47, "oz-in"     },
  { 48, "gauss"     },
  { 49, "gilberts"      },
  { 50, "henry"     },
  { 51, "millihenry"      },
  { 52, "farad"     },
  { 53, "microfarad"      },
  { 54, "ohms"      },
  { 55, "siemens"     },
  { 56, "mole"      },
  { 57, "becquerel"     },
  { 58, "PPM"     },
  { 59, "reserved"      },
  { 60, "Decibels"      },
  { 61, "DbA"     },
  { 62, "DbC"     },
  { 63, "gray"      },
  { 64, "sievert"     },
  { 65, "color temp deg K"      },
  { 66, "bit"     },
  { 67, "kilobit"     },
  { 68, "megabit"     },
  { 69, "gigabit"     },
  { 70, "byte"      },
  { 71, "kilobyte"      },
  { 72, "megabyte"      },
  { 73, "gigabyte"      },
  { 74, "word"      },
  { 75, "dword"     },
  { 76, "qword"     },
  { 77, "line"      },
  { 78, "hit"     },
  { 79, "miss"      },
  { 80, "retry"     },
  { 81, "reset"     },
  { 82, "overrun / overflow"      },
  { 83, "underrun"      },
  { 84, "collision"     },
  { 85, "packets"     },
  { 86, "messages"      },
  { 87, "characters"      },
  { 88, "error"     },
  { 89, "correctable error"     },
  { 90, "uncorrectable error"     },
  { 0, NULL}};

struct sensortype
  {
  unsigned char code;
  char          *name;
  };
char *sensor_get_unit_type(unsigned char sensor_unit_type_code)
{
  int i = 0;

  while(sensorunittypelist[i].unit != NULL)
    {
    if (sensor_unit_type_code == sensorunittypelist[i].code)
      return(sensorunittypelist[i].unit);
    i++;
    }
  return(NULL);
}
static struct sensortype sensortypelist[] = {
  { 0x01, "Temperature"              },
  { 0x02, "Voltage"                  },
  { 0x03, "Current"                  },
  { 0x04, "Fan"                      },
  { 0x05, "Physical Security"        },
  { 0x06, "Platform Security"        },
  { 0x07, "Processor"                },
  { 0x08, "Power Supply"             },
  { 0x09, "Power Unit"               },
  { 0x0a, "Cooling Device"           },
  { 0x0b, "Other Sensor"             },
  { 0x0c, "Memory"                   },
  { 0x0d, "Drive Slot (Bay)"         },
  { 0x0e, "POST Memory Resize"       },
  { 0x0f, "POST Error"               },
  { 0x10, "Event Logging Disabled"   },
  { 0x11, "Watchdog 1"               },
  { 0x12, "System Event"             },
  { 0x13, "Critical Interrupt"       },
  { 0x14, "Button"                   },
  { 0x15, "Module / Board"           },
  { 0x16, "Microcontroller / Coprocessor" },
  { 0x17, "Add-in Card"              },
  { 0x18, "Chasis"                   },
  { 0x19, "Chipset"                  },
  { 0x1a, "Other FRU"                },
  { 0x1b, "Cable / Interconnect"     },
  { 0x1c, "Terminator"               },
  { 0x1d, "System Boot"              },
  { 0x1e, "Boot Error"               },
  { 0x1f, "O/S Boot"                 },
  { 0x20, "O/S Critical Stop"        },
  { 0x21, "Slot / Connector"         },
  { 0x22, "System ACPI Power State"  },
  { 0x00, 0                         }};

char *sensor_get_type(unsigned char sensor_type_code)
{
  int i = 0;

  while(sensortypelist[i].name != NULL)
    {
    if (sensor_type_code == sensortypelist[i].code)
      return(sensortypelist[i].name);
    i++;
    }
  return(NULL);
}

struct eventdesc
  {
  unsigned char sensor_type_code;
  unsigned char sensor_specific_offset;
  char          *desc;
  };

static struct eventdesc sensor_event_list[] = {
  { 0x05, 0x00, "General Chasis Intrusion" },
  { 0x05, 0x01, "Drive Bay Intrusion"      },
  { 0x05, 0x02, "I/O Card Area Intrusion"  },
  { 0x05, 0x03, "Processor Area Intrusion" },
  { 0x05, 0x04, "LAN Leash Lost"           },
  { 0x05, 0x05, "Unauthorized Dock/Undock" },

  { 0x06, 0x00, "Secure Mode Violation attempt"                       },
  { 0x06, 0x01, "Pre-boot Password Violation - user password"         },
  { 0x06, 0x02, "Pre-boot Password Violation - sys password"          },
  { 0x06, 0x03, "Pre-boot Password Violation - network boot password" },
  { 0x06, 0x04, "Other Pre-boot Password Violation"                   },
  { 0x06, 0x05, "Out-of-band Access Password Violation"               },

  { 0x07, 0x00, "IERR"                                        },
  { 0x07, 0x01, "Thermal Trip"                                },
  { 0x07, 0x02, "FBR1/BIST failure"                           },
  { 0x07, 0x03, "FBR2/Hang in POST failure"                   },
  { 0x07, 0x04, "FBR3/Processor Startup / Initialize failure" },
  { 0x07, 0x05, "Configuration Error"                         },
  { 0x07, 0x06, "SM BIOS 'Uncorrectable CPU-complex' error"   },
  { 0x07, 0x07, "Processor Presence Detected"                 },
  { 0x07, 0x08, "Processor Disabled"                          },
  { 0x07, 0x09, "Terminator Presence Detected"                },

  { 0x08, 0x00, "Presence Detected"             },
  { 0x08, 0x01, "Power Supply Failure Detected" },
  { 0x08, 0x02, "Predictive Failure Asserted"   },

  { 0x09, 0x00, "Power Off / Power Down"      },
  { 0x09, 0x01, "Power Cycle"                 },
  { 0x09, 0x02, "240 UA Power Down"           },
  { 0x09, 0x03, "Interlock Power Down"        },
  { 0x09, 0x04, "A/C Lost"                    },
  { 0x09, 0x05, "Soft Power Control Failure"  },
  { 0x09, 0x06, "Power Unit Failure Detected" },

  { 0x0c, 0x00, "Correctable ECC"     },
  { 0x0c, 0x01, "Uncorrectable ECC"   },
  { 0x0c, 0x02, "Parity"              },
  { 0x0c, 0x03, "Memory Scrub Failed" },

  { 0x10, 0x00, "Correctable Memory Error Logging Disabled" },
  { 0x10, 0x01, "Event 'Type' Logging Disabled"             },
  { 0x10, 0x02, "Log Area Reset / Cleared"                  },
  { 0x10, 0x03, "All Event Logging Disabled"                },

  { 0x11, 0x00, "BIOS Watchdog Reset"               },
  { 0x11, 0x01, "OS Watchdog Reset"                 },
  { 0x11, 0x02, "OS Watchdog Shutdown"              },
  { 0x11, 0x03, "OS Watchdog Power Down"            },
  { 0x11, 0x04, "OS Watchdog Power Cycle"           },
  { 0x11, 0x05, "OS Watchdog NMI"                   },
  { 0x11, 0x06, "OS Watchdog Expired"               },
  { 0x11, 0x07, "OS Watchdog Pre-timeout Interrupt" },

  { 0x12, 0x00, "System Reconfigured"                  },
  { 0x12, 0x01, "OEM System Boot Event"                },
  { 0x12, 0x02, "Undetermined System Hardware Failure" },

  { 0x13, 0x00, "Front Panel NMI"             },
  { 0x13, 0x01, "Bus Timeout"                 },
  { 0x13, 0x02, "I/O Channel Check NMI"       },
  { 0x13, 0x03, "Software NMI"                },
  { 0x13, 0x04, "PCI PERR"                    },
  { 0x13, 0x05, "PCI SERR"                    },
  { 0x13, 0x06, "EISA Fail Safe Timeout"      },
  { 0x13, 0x07, "Fatal NMI (port 61h, bit 7)" },

  { 0x1d, 0x00, "Initiated By Power up"        },
  { 0x1d, 0x01, "Initiated By Hard Reset"      },
  { 0x1d, 0x02, "Initiated By Warm Reset"      },
  { 0x1d, 0x03, "User Requested PXE Boot"      },
  { 0x1d, 0x04, "Autumatic Boot To Diagnostic" },

  { 0x1e, 0x00, "No Bootable Media"                                 },
  { 0x1e, 0x01, "Non-Bootable Diskette Left In Drive"               },
  { 0x1e, 0x02, "PXE Server Not Found"                              },
  { 0x1e, 0x03, "Invalid Boot Sector"                               },
  { 0x1e, 0x04, "Timeout Waiting For User Selection Of Boot Source" },

  { 0x1f, 0x00, "A: Boot Completed"                          },
  { 0x1f, 0x01, "C: Boot Completed"                          },
  { 0x1f, 0x02, "PXE Boot Completed"                         },
  { 0x1f, 0x03, "Diagnostic Boot Completed"                  },
  { 0x1f, 0x04, "CD-ROM Boot Completed"                      },
  { 0x1f, 0x05, "ROM Boot Completed"                         },
  { 0x1f, 0x06, "Boot Completed - Boot Device Not Specified" },

  { 0x20, 0x00, "Stop During O/S Load / Initialization" },
  { 0x20, 0x01, "Run-time Stop"                         },

  { 0x21, 0x00, "Fault Status Asserted"                          },
  { 0x21, 0x01, "Identify Status Asserted"                       },
  { 0x21, 0x02, "Slot / Connector Device Installed"              },
  { 0x21, 0x03, "Slot / Connector Ready For Device Installation" },
  { 0x21, 0x04, "Slot / Connector Ready For Device Removal"      },
  { 0x21, 0x05, "Slot Power Is Off"                              },
  { 0x21, 0x06, "Slot / Connector Device Removal Request"        },
  { 0x21, 0x07, "Interlock Asserted"                             },

  { 0x22, 0x00, "S0 / G0 'working'"                                          },
  { 0x22, 0x01, "S1 Sleeping With System H/W & Processor Context Maintained" },
  { 0x22, 0x02, "S2 Sleeping, Processor Context Lost"                        },
  { 0x22, 0x03, "S3 Sleeping, Processor & H/W Context Lost, Mem Retained"    },
  { 0x22, 0x04, "S4 Non-volatile Sleep / Suspend To Disk"                    },
  { 0x22, 0x05, "S5 / G2 Soft-Off"                                           },
  { 0x22, 0x06, "S4 / S5 Soft-Off, Exact S4 / S5 State Cannot Be Determined" },
  { 0x22, 0x07, "G3 / Mechanical Off"                                        },
  { 0x22, 0x08, "Sleeping In An S1, S2, Or S3 State"                         },
  { 0x22, 0x09, "G1 Sleeping"                                                },
  { 0x00, 0x00, NULL}};

char *sensor_get_canned_event_desc(unsigned char sensor_type_code,
                            unsigned char sensor_specific_offset)
{
  int i = 0;

  while(sensor_event_list[i].desc != NULL)
    {
    if ((sensor_event_list[i].sensor_type_code == sensor_type_code) &&
        (sensor_event_list[i].sensor_specific_offset == sensor_specific_offset))
      return(sensor_event_list[i].desc);

    i++;
    }
  return(NULL);
}

void sensor_str_convert(CONTEXT				*ctx,
                        unsigned char	sensor_number,
                        unsigned char	*data,
                        char					*string_msg)
{
 /* DISCRETE_SENSOR_READING_RESPONSE  *discrete; */
  DIGITAL_SENSOR_READING_RESPONSE   *digital;
  ANALOG_SENSOR_READING_RESPONSE    *analog;
  SENSOR                       			*sensor;
	char															log_msg[255];

  /* Make heads or tails out of the data */
  sensor = ctx->sensor_list;
  while(sensor)
    {
		if (lookup_sensor_number(sensor) == sensor_number)
			break;
    sensor = sensor->next;
    }
  if (!sensor)
    {
    sprintf(string_msg, "JOB_ERROR:INTERNAL_ERROR");
		sprintf(log_msg,
						"0:0:LOG:Thread %s internal error, Sensor %d not found",
						ctx->node_id,sensor_number);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		return;
    }
  if (sensor->rec_header.record_type == 0x01)
    {
    if (sensor->sensor_type.type_01_sensor.sens_capable_reading_type == ANALOG)
      {
      double  final_value;

      analog = (ANALOG_SENSOR_READING_RESPONSE *) &data[0];
      final_value = linearize_sensor_reading(sensor, analog->reading);
      sprintf(string_msg,
              "SENSOR:%d:%.2f:%s",
              sensor_number,
              final_value,
              sensor_get_unit_type(sensor->sensor_type.type_01_sensor.sensor_units_2_base_unit));
      return;
      }
    else
      {
      sprintf(log_msg,
							"0:0:LOG:Thread %s sensor reading type 0x%x not supported.",
						 	ctx->node_id,
             	sensor->sensor_type.type_01_sensor.sens_capable_reading_type);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
      }
    }
  else if (sensor->rec_header.record_type == 0x02)
    {
    int sensor_class;

    /* This is a digital / discrete reading without thresholds */
    sensor_class = sensor_get_class(sensor->sensor_type.type_02_sensor.event_reading_type_code);
    if (sensor_class == SENSOR_CLASS_DIGITAL)
      {
      digital = (DIGITAL_SENSOR_READING_RESPONSE *) &data[0];
      sprintf(string_msg,
              "SENSOR:%d:%s",
              sensor_number,
              sensor_get_trigger_description(
								sensor->sensor_type.type_02_sensor.event_reading_type_code,
                digital->digital_status_offset));
      return;
      }
    else if (sensor_class == SENSOR_CLASS_DISCRETE)
      {
      sprintf(log_msg,
							"0:0:LOG:Thread %s discrete sensors not currently supported",
							ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
      }
		else if (sensor_class == SENSOR_CLASS_UNKNOWN)
			{
      sprintf(string_msg, "JOB_ERROR:SENSOR_NOT_SUPPORTED");
			return;
			}
    else
      {
      sprintf(log_msg,
							"0:0:LOG:Thread %s unknown type 02 sensor class 0x%x",
							ctx->node_id,
							sensor_class);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
      sprintf(string_msg, "JOB_ERROR:SENSOR_NOT_SUPPORTED");
      return;
      }
    } 
  else
    {
    sprintf(log_msg,
						"0:0:LOG:Thread %s unsupported sensor record type 0x%x",
					 ctx->node_id,
           sensor->rec_header.record_type);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
    sprintf(string_msg, "JOB_ERROR:SENSOR_NOT_SUPPORTED");
    return;
    }
}

static char *iana_lookup(unsigned int iana_number)
{
	unsigned int i;

	for (i=0; iana_list[i].number != 0; i++)
		{
		pthread_testcancel();
		if (iana_number == iana_list[i].number)
			break;
		}

	if (iana_list[i].number == 0)
		return(NULL);
	return(iana_list[i].name);
}
