
/*
 * This program is original work by Zac Sprackett <zsprackett@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2000, Zac Sprackett, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This Module is based heavily on the other loose modules by San Mehat
 * <nettwerk@valinux.com>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <rsh_module_harness.h>
#include <rsh_slave_context.h>
#include <prototypes.h>
#include <rsh_slave_context.h>
#include <pthread.h>

typedef struct        info_squirrel_t {
	int        job_id;
	int        ret_fd;
} InfoSquirrel;

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_rsh_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static void	set_agent(CONTEXT *ctx, char *agent);
static void	set_addr(CONTEXT *ctx, char *addr);
static void	set_user(CONTEXT *ctx, char *user);
static FILE	*do_rsh(char *addr, char *user, char *cmd);

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT		*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
#ifdef THREAD_DEBUG
	printf("Thread %s cleaning up\n",ctx->node_id);
#endif
#ifdef THREAD_DEBUG
	printf("Thread %s clean\n",ctx->node_id);
#endif
	ctx->refcount--;
}

void *
thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT	*ctx = (CONTEXT *) data;
	struct timeval	to;
	fd_set		read_fds;
	int		max, rc;
	char		log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,"0:0:LOG:Thread %s starting (pid %d)",ctx->node_id,
		getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while (1) {
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0) {
			sprintf(log_msg,
				"0:0:LOG:Thread %s select() failed (%s)",
				ctx->node_id,strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
		}
		if (!rc) {
			pthread_testcancel();
		}
		else if (FD_ISSET(ctx->thread_ipc_read, &read_fds)) {
			process_msg_from_slave(ctx);
		}
		pthread_testcancel();
	}
}

static void 
process_msg_from_slave(CONTEXT *ctx)
{
	char	*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char	*ipc_cmd, *ipc_args, *variable, *value;
	char	*new_name;
	int	len, rc, client_fd;
	__uint32_t	job_id;
	char	*last = (char *) 0x00;
	char	log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0) {
		sprintf(log_msg,
			"0:0:LOG:Thread %s unable to read msg from slave (%s)",
			ctx->node_id,strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
	}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",
				 ctx->node_id,
				 msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC")) {
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_rsh_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR")) {
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;
		if (!(value = strtok_r(NULL, ":", &last)))
			goto out;
		if (!strcasecmp(variable,"AGENT"))
			set_agent(ctx, value);
		if (!strcasecmp(variable,"USER"))
			set_user(ctx, value);
		else if (!strcasecmp(variable,"IP_ADDRESS"))
			set_addr(ctx, value);
	}
	else if (!strcasecmp(cmd, "CLIENT_DIS")) {
		char	*fd;
		int		client_fd;

		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
	}
	else if (!strcasecmp(cmd, "NODE_RENAME")) {
		char	logfile_old[255];
		char	logfile_new[255];

		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
		/* Rename the logfile */
		sprintf(logfile_old,"%s/vacm_rsh_alert_log.%s",LOGDIR,
			ctx->node_id);
		free(ctx->node_id);
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		sprintf(logfile_new,"%s/vacm_rsh_alert_log.%s",LOGDIR,
			ctx->node_id);
		if (rename(logfile_old, logfile_new) < 0) {
			sprintf(log_msg,"0:0:LOG:Thread %s unable to rename logfile (%m)",
				ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
		}
	}
	else	{
		sprintf(log_msg,"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
			ctx->node_id,cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
	}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Modify the rsh agent
 */
static void set_agent(CONTEXT *ctx, char *agent)
{
	if (ctx->agent)
		free(ctx->agent);
	ctx->agent = malloc((strlen(agent)+1));
	strcpy(ctx->agent, agent);
}

static void set_addr(CONTEXT *ctx, char *addr)
{
	if (ctx->addr)
		free(ctx->addr);
	ctx->addr = malloc((strlen(addr)+1));
	strcpy(ctx->addr, addr);
}

static void set_user(CONTEXT *ctx, char *user)
{
	if (ctx->user)
		free(ctx->user);
	ctx->user = malloc((strlen(user)+1));
	strcpy(ctx->user, user);
}

static int
getPingInfo(CONTEXT *ctx,FILE *inf)
{
	char	line[1024];
	int	lines = 0;
	char	*cp;
	KNOWN_UNUSED_PARAM(ctx);

	while (fgets(line,1023,inf) != NULL) {
		if (line[0] == '\n') continue; 
		if ((cp = strstr(line,"packets received")) != NULL) {
			/*
			 *	backup to the comma and try and find
			 *	number of successful packets.
			 */
			while (cp > line && *cp != ',') cp--;
			if (*cp == ',') {
				cp++; 
				while (*cp && !isdigit(*cp)) cp++;
				return(*cp == '0' ? 0 : 1);
			}
			lines++;
			continue;
		}
		else        {
		}
		lines++;
	}
	return(1);
}

/*
 * Process an IPC message
 */
static void 
process_rsh_ipc(CONTEXT *ctx,__uint32_t job_id,int client_fd,char *ipc_cmd,
		char *ipc_args)
{
	char	resp_msg[1024];
	char	*last = (char *) 0x00;
	FILE	*rem = NULL;
	char	buff[4096];

	if (!strcasecmp(ipc_cmd, "CONFIGURATION")) {
		char	*agent;
		char	*user;

		if (!(agent = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(user = strtok_r(NULL, ":", &last)))
			goto out_malformatted;

		if (!strcasecmp(agent,"RSH")) {
			set_agent(ctx, agent);
			sprintf(resp_msg,"0:0:SET:%s:RSH:AGENT:%s",ctx->node_id,agent);
			send_msg(ctx->thread_ipc_write,resp_msg);
			set_user(ctx, user);
			sprintf(resp_msg,"0:0:SET:%s:RSH:USER:%s",ctx->node_id,user);
			send_msg(ctx->thread_ipc_write,resp_msg);
		}
		else	{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,
				client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);	
			goto out_ok;
		}

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);	
		goto out_ok;
	}

	/* Default to user root if not configured */
	if (!ctx->user) {
		sprintf(resp_msg,
			"0:0:LOG:Thread %s setting RSH user to default of root", ctx->node_id);
		send_msg(ctx->thread_ipc_write, resp_msg);

		set_user(ctx, "root");
		sprintf(resp_msg,"0:0:SET:%s:RSH:USER:%s",ctx->node_id,"root");
		send_msg(ctx->thread_ipc_write,resp_msg);
	}
#ifdef	notdef
	if (!ctx->agent) {
		sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s:%s",
			job_id,client_fd,ctx->node_id,strerror(ENOENT));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
	}
#endif	/*notdef*/
	if (!ctx->addr) {
		sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s:%s",
			job_id,client_fd,ctx->node_id,strerror(EDESTADDRREQ));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
	}

	/* Begin real commands */
	if (!strcasecmp(ipc_cmd, "INQUIRY")) {		//Inquiry
		sprintf(resp_msg, "%d:%d:FOR:INQUIRY:%s:%s:%s",
			job_id, client_fd,ctx->node_id,ctx->addr,ctx->user);
		send_msg(ctx->thread_ipc_write, resp_msg);
	}
	else if (!strcasecmp(ipc_cmd, "NODE_INV_INFO")) { // Node Inventory
		char *os;
		char *version;

//		if (!strcasecmp(ctx->agent,"rsh")) {
			if (!(rem = do_rsh(ctx->addr,ctx->user,"uname -rs")))
				goto failed_connect;
//		}

		fgets(buff,sizeof(buff),rem);
		os = strtok_r(buff," ",&last);
		version = strtok_r(NULL,"\n",&last);
    
   		sprintf(resp_msg,"%d:%d:FOR:NODEINVINFO:%s:%s:",job_id,
			client_fd, os,version);
   		send_msg(ctx->thread_ipc_write, resp_msg);
	}
	else if (!strcasecmp(ipc_cmd, "LOAD_AVG")) {		// Load Average
    		float one, five, fifteen;
 
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"cat /proc/loadavg")))
				goto failed_connect;

		fgets(buff,sizeof(buff),rem);
		one = atof(strtok_r(buff," ",&last));
		five = atof(strtok_r(NULL," ",&last));
		fifteen = atof(strtok_r(NULL," ",&last));

		sprintf(resp_msg,"%d:%d:FOR:LOAD_AVG:%.2f:%.2f:%.2f",
      			job_id,client_fd,one,five,fifteen);
		send_msg(ctx->thread_ipc_write, resp_msg);
		fclose(rem);
	}
	else if (!strcasecmp(ipc_cmd, "WHO")) { // Who
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"who")))
				goto failed_connect;
		while(fgets(buff,sizeof(buff),rem)) {
			char name[16];
			char tty[16];

			buff[strlen(buff)-1] = 0;

			strcpy(name, strtok_r(buff, "\t ", &last));
			strcpy(tty, strtok_r(NULL, "\t ", &last));

			sprintf(resp_msg,"%d:%d:FOR:WHO:%s:%s",job_id,client_fd,name,tty);
			send_msg(ctx->thread_ipc_write, resp_msg);
		}
    		fclose(rem);
	}
	else if (!strcasecmp(ipc_cmd, "MEM")) {			// MEM
		char *bword;
		int total = 0, free = 0, shared = 0;
		int buffers = 0, cached = 0, swap_total = 0, swap_free = 0;
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"cat /proc/meminfo")))
				goto failed_connect;
		while (fgets(buff,sizeof(buff),rem)) {
			if (!(bword = strtok_r(buff,":",&last))) continue;
			if (!strcmp(bword, "MemTotal"))
				total = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "MemFree"))
				free = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "MemShared"))
				shared = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "Buffers"))
				buffers = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "Cached"))
				cached = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "SwapTotal"))
				swap_total = atoi(strtok_r(NULL,"\t ",&last));
			else if (!strcmp(bword, "SwapFree"))
				swap_free = atoi(strtok_r(NULL,"\t ",&last));
		}
   		sprintf(resp_msg,"%d:%d:FOR:MEM:%d:%d:%d:%d:%d:%d:%d",
      			job_id,client_fd,total,free,shared,buffers,cached,
			swap_total,swap_free);
		send_msg(ctx->thread_ipc_write, resp_msg);
	}
	else if (!strcasecmp(ipc_cmd, "PS")) {			// MEM
		char pid[8],user[16],cpu[6],mem[6],rss[16],state[6],tty[8],cmd[64];
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,
				"ps --no-headers -eo pid,user,%cpu,%mem,rss,state,tty,cmd")))
				goto failed_connect;
		while(fgets(buff,sizeof(buff),rem)) {
			buff[(strlen(buff)-1)]=0;

			strncpy(pid,strtok_r(buff,"\t ",&last),sizeof(pid));
			strncpy(user,strtok_r(NULL,"\t ",&last),sizeof(user));
			strncpy(cpu,strtok_r(NULL,"\t ",&last),sizeof(cpu));
			strncpy(mem,strtok_r(NULL,"\t ",&last),sizeof(mem));
			strncpy(rss,strtok_r(NULL,"\t ",&last),sizeof(rss));
			strncpy(state,strtok_r(NULL,"\t ",&last),sizeof(state));
			strncpy(tty,strtok_r(NULL,"\t ",&last),sizeof(tty));
			strncpy(cmd,strtok_r(NULL,"\t ",&last),sizeof(cmd));

			sprintf(resp_msg,"%d:%d:FOR:PS:%s:%s:%s:%s:%s:%s:%s:%s",
			job_id,client_fd,pid,user,cpu,mem,rss,state,tty,cmd);
			send_msg(ctx->thread_ipc_write, resp_msg);
		}
	}
	else if (!strcasecmp(ipc_cmd, "DOWNLOAD_LOG")) {	// Download Log
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"cat /var/log/messages")))
				goto failed_connect;

		while (fgets(buff,sizeof(buff),rem)) {
			buff[strlen(buff)-1] = '\0';
			sprintf(resp_msg,"%d:%d:FOR:DOWNLOAD_LOG:%s",job_id,client_fd,buff);
			send_msg(ctx->thread_ipc_write, resp_msg);
		}
	}
	else if (!strcasecmp(ipc_cmd, "SHUTDOWN")) {	// Shutdown -h now
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"/sbin/shutdown -h now")))
				goto failed_connect;
		fclose(rem);
	}
	else if (!strcasecmp(ipc_cmd, "RESTART")) {	//  Shutdown -r now
//		if (!strcasecmp(ctx->agent,"rsh"))
			if (!(rem = do_rsh(ctx->addr, ctx->user,"/sbin/shutdown -r now")))
				goto failed_connect;
			fclose(rem);
	}

	else if (!strcasecmp(ipc_cmd,"CHASSIS_STATUS")) {
//		int	done = 0;
		FILE	*inf;
		char	cmd[512];
//		char	*cp = NULL;
//		char	line[1024];
//		int	lines = 0;
//		int	timer_flag=1;
		InfoSquirrel *sq = (InfoSquirrel *)malloc(sizeof(InfoSquirrel));

		sq->job_id = job_id;
		sq->ret_fd = ctx->thread_ipc_write;

		if (ctx->addr == NULL)
			goto failed_connect;
		sprintf(cmd,"/bin/ping -c 1 %s",ctx->addr);
		sprintf(resp_msg,
			"0:0:LOG:Requesting chassis status for %s",ctx->addr);
		send_msg(ctx->thread_ipc_write, resp_msg);

		if ((inf = popen(cmd,"r")) == NULL) {
			fprintf(stderr,"couldn't popen()!\n");
			goto failed_connect;
		}
		sprintf(resp_msg,"%d:%d:FOR:CSTATUS:%s:",job_id,
			client_fd,getPingInfo(ctx,inf) ? "YES" : "NO");
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		fclose(inf);
		goto out_ok;
	}
	else	{
		sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,
			strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
	}
	sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
	send_msg(ctx->thread_ipc_write, resp_msg);	

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,
		strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;

failed_connect:
	sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,
		strerror(ECONNREFUSED));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}

/*
 * Receive a message. This function is threadsafe
 */
static int 
recv_msg(int fd, char **buffer, __uint32_t *len)
{
	__uint32_t	net_len;
	__uint32_t	brtr;
	char	*offset;
	int	rc;

	if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	else if (!rc)
		return(-1);
	*len = ntohl(net_len);

	if (!(*buffer = malloc(*len)))
		return(-ENOMEM);
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr) {
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
	}
	return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int 
send_msg(int fd, char *msg)
{
	__uint32_t	len;
	__uint32_t	net_len;
	int	rc;

	if (fd == -1)
		return(-EBADF);
	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	if ((rc = write (fd, (void *) msg, len)) < 0)
		return(rc);
	return(0);
}

static FILE *
do_rsh(char *addr, char *user, char *cmd)
{
 	char	tmp[64]; 
 	char	*address = tmp;
 	int	remote;

 	strcpy(tmp,addr);
 	remote = rcmd(&address,SHELL,"root",user,cmd,0);
 	if (remote<0)
		return(NULL);

 	return(fdopen(remote,"r"));
}
