/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2000, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 *  with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef _SERCON_SLAVE_CONTEXT_H
#define _SERCON_SLAVE_CONTEXT_H

#include <config.h>
#include <pthread.h>

typedef struct console_connection
	{
	int							   client_fd;       /* Client socket for commands */
	struct sockaddr_in client_addr;

	int							   sock_fd;         /* Client socket for term I/O */
	int							   listen_fd;       /* Initial socket */

	time_t					   connect_time;    /* Time the client connected to I/O */
	time_t             last_byte_time;	/* Time of last byte reception */
#define MODE_RO	1
#define MODE_RW	2
	int							   access_mode;
	char						   otpasswd[64];
  unsigned short     port;
	
#define STAGE_LISTENING       0
#define STAGE_PASSWORD_PROMPT	1
#define STAGE_PASSWORD_REPLY  2
#define STAGE_CONNECTED       3
#define STAGE_DESTROYED				99
	int                stage;
  int                login_retries;

	char							 conv_buffer[1024];
	int								 conv_pos;

	struct console_connection *next;
	struct console_connection *prev;
	} CONSOLE_CONNECTION;

typedef struct alert
	{
	int		alert_id;							// Internal ID
	char	trigger[64];		// Trigger
	int		lines_to_capture;			// duhh
	int		current_pos;					// Current position in trigger for comparison
	int		capturing;						// set to 1 when capturing
	int		linecount;						// number of lines currently captured
	char	line_buffer[1024];		// buffer for line bytes
	int		line_pos;							// position in line pos
	char	*capture_buffer[1024];// Full capture buffer
	int		cb_pos;								// Line number in capture buffer

	struct alert	*next;
	struct alert	*prev;
	} ALERT;

typedef struct context
	{
	pthread_t						thread;
	pthread_attr_t			attr;

	FILE								*words;
	int									con_sock;
	char								*node_id;
	char								*device_address;
	unsigned int				baudrate;
	
	// Cached copy of the Nexxus IP address for connection requests
	struct in_addr			module_ipaddr;

	int									thread_ipc_read;
	int									thread_ipc_write;
	int									slave_ipc_read;
	int									slave_ipc_write;

	CONSOLE_CONNECTION	*connections;
	ALERT								*alerts;
	int									next_alert_id;

	int									refcount;
	struct context			*next;
	struct context			*prev;
	} CONTEXT;

#endif
