/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * VACM Remote System Status Daemon
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/resource.h>
#include <crypt.h>
#include <vacm_sys_statd.h>
#include <config.h>

#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define SSL_HOME "/etc"
#define CERTF SSL_HOME "/vacm.cert"
#define KEYF SSL_HOME "/vacm.key"
#endif

static void process_cmd(NEXXUS	*nexxusp, char *buffer);
static void send_response(int fd, char *msg);
static int	do_mem(__uint32_t job_id, int	fd);
static int	do_cpu(__uint32_t job_id, int	fd);
static int	do_uptime(__uint32_t job_id, int	fd);
static int	do_fs(__uint32_t job_id, int	fd);
static int	do_who(__uint32_t job_id, int	fd);
static int	do_ps(__uint32_t job_id, int	fd);
static int	do_version(__uint32_t job_id, int	fd);
static int	do_apm(__uint32_t job_id, int	fd);
static int	do_int(__uint32_t job_id, int	fd);
static int	do_io(__uint32_t job_id, int	fd);
static int	do_dma(__uint32_t job_id, int	fd);
static int	do_swap(__uint32_t job_id, int	fd);
static int	get_password(void);
static int	set_password(void);
static void usage(char *name);
static void become_daemon(void);
static void signal_handler(int signal);
#ifdef OPENSSL
static SSL *lookup_ssl_by_fd(int fd);
#endif

#define TIMER_INTERVAL 1
#define GRACE_PERIOD   0

static int				comm_sock;
static NEXXUS			nexxus[255];
int								debug =0;
char							daemon_password[255];
int								foreground_mode = 0;
int								priority;
struct itimerval	itimer;
struct timeval		last_time;

#ifdef OPENSSL
static SSL_CTX		*ctx = NULL;
static SSL_METHOD	*meth = NULL;
#endif

int main(argc, argv)
int		argc;
char	**argv;
{
	struct sockaddr_in	sock;
	int									I_optval = 1;
	fd_set							read_fds;
	struct timeval			to;
	int									max, I_c1, len, rc;
	char								buffer[512];
#ifdef OPENSSL
	int									err;
#endif
 
	if (getuid() != 0)
		{
		printf("You must be root to run this\n");
		exit(-1);
		}

	printf("VACM System Status Daemon version 1.2 (nettwerk@valinux.com)\n");
	if ((priority = getpriority(PRIO_PROCESS, 0))<0)
		{
		perror("getpriority");
		exit(-1);
		}

	while ((rc = getopt(argc, argv, "pf"))!=EOF)
		{
		if ((char)rc == 'p')
			{
			if (!set_password())
				exit(-1);
			exit(0);
			}
		else if ((char)rc =='f')
			foreground_mode = 1;
		else
			{
			usage(argv[0]);
			exit(-1);
			}
		}
		
	if (!get_password())
		{
		printf("No nexxus access password set. Please set it\n");
		if (!set_password())
			{
			printf("Unable to set password.\n");
			exit(-1);
			}
		printf("Password set. Please rerun daemon\n");
		exit(0);
		}

	if (!foreground_mode)
		become_daemon();

	signal(SIGPIPE, SIG_IGN);
#ifdef OPENSSL
  SSL_load_error_strings();
  SSLeay_add_ssl_algorithms();
  meth = SSLv23_server_method();
  ctx = SSL_CTX_new(meth);
  if (!ctx)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (SSL_CTX_use_certificate_file(ctx, CERTF, SSL_FILETYPE_PEM) <=0)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (SSL_CTX_use_PrivateKey_file(ctx, KEYF, SSL_FILETYPE_PEM) < 0)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (!SSL_CTX_check_private_key(ctx))
    {
    printf("Private key does not match the certificate public key\n");
		exit(-1);
    }
	printf("OpenSSL encryption layer enabled\n");
#endif
	if ((comm_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
		{
		perror("socket");
		exit(1);
		}
	if (setsockopt(comm_sock,
								 SOL_SOCKET,
								 SO_REUSEADDR,
								 &I_optval,
								 sizeof(I_optval)) < 0)
		{
		perror("setsockopt");
		exit(1);
		}
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = htonl(INADDR_ANY);
	sock.sin_port = htons(NETWORK_PORT);

	if (bind(comm_sock,
					 (struct sockaddr *) &sock,
					 sizeof(sock))<0)
		{
		perror("bind");
		exit(1);
		}
	if (listen(comm_sock, 5)<0)
		{
		perror("listen");
		exit(1);
		}
	for (I_c1=0; I_c1<255;I_c1++)
		nexxus[I_c1].sock = -1;

	/* Setup the interval timer */
	signal(SIGALRM, signal_handler);
	memset(&itimer,0,sizeof(itimer));
	itimer.it_value.tv_sec=TIMER_INTERVAL;
	itimer.it_value.tv_usec=0;
	itimer.it_interval.tv_sec=TIMER_INTERVAL;
	itimer.it_interval.tv_usec=0;
	if (setitimer(ITIMER_REAL, &itimer,NULL)<0)
		{
		perror("setitimer");
		exit(-1);
		}
	/* Initial value for it */
	gettimeofday(&last_time, NULL);

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 10;
		to.tv_usec = 0;
		FD_SET(comm_sock, &read_fds);
		max = comm_sock;

		for (I_c1 = 0; I_c1 < 255; I_c1++)
			{
			if (nexxus[I_c1].sock != -1)
				{
				if (max < nexxus[I_c1].sock)
					max = nexxus[I_c1].sock;
				FD_SET(nexxus[I_c1].sock, &read_fds);
				}
			}
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to))<0)
			{
			if (errno == EINTR)
				continue;
			perror("select");
			continue;
			}
		else if (!rc)
			{
			continue;
			}
		else 
			{
			if (FD_ISSET(comm_sock, &read_fds))
				{
				/* New client connection */
				for(I_c1 =0; I_c1<255;I_c1++)
					if (nexxus[I_c1].sock == -1)
						break;
				if (I_c1 == 255)
					{
					printf("[VACM_SYS_STATD] Maximum number of Nexxus's connected\n");
					sleep(1);
					continue;
					}
				len = sizeof(struct sockaddr_in);
				if ((nexxus[I_c1].sock = accept(comm_sock,
																			 (struct sockaddr *) &sock,
																			 &len))<0)
					{
					perror("accept");
					sleep(1);
					continue;
					}
				if (debug)
					{
					printf("[VACM_SYS_STATD] Nexxus connection accepted (fd %d)\n",
								 nexxus[I_c1].sock);
					}
#ifdef OPENSSL
				nexxus[I_c1].ssl = SSL_new(ctx);
				SSL_set_fd(nexxus[I_c1].ssl, nexxus[I_c1].sock);
				err = SSL_accept(nexxus[I_c1].ssl);
    		if (!strcmp(SSL_get_cipher(nexxus[I_c1].ssl), "(NONE)"))
      		{
          printf("Non-SSL client detected. Disconnecting.\n");
      		SSL_free(nexxus[I_c1].ssl);
					close(nexxus[I_c1].sock);
					nexxus[I_c1].authenticated = 0;
					nexxus[I_c1].sock = -1;
					continue;
      		}
				if (debug)
					printf("SSL connection using %s\n",SSL_get_cipher(nexxus[I_c1].ssl));
				if (!(nexxus[I_c1].cert = SSL_get_peer_certificate(nexxus[I_c1].ssl)))
					{
					if (debug)
						printf("Note: Client does not have a certificate\n");
					}
#endif
				} /* end new client connection */
			for (I_c1 =0; I_c1<255;I_c1++)
				{
				if (nexxus[I_c1].sock != -1)
					{
					if (FD_ISSET(nexxus[I_c1].sock, &read_fds))
						{
						__uint32_t	len;
		
						if (debug)
							{
							printf("[VACM_SYS_STATD] Data ready on fd %d\n",
										 nexxus[I_c1].sock);
							}
#ifdef OPENSSL
						rc = SSL_read(nexxus[I_c1].ssl, (char *) &len, sizeof(len));
						if ((!rc) || (rc <0))
							{
							SSL_shutdown(nexxus[I_c1].ssl);
							SSL_free(nexxus[I_c1].ssl);
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
#else
						rc = read(nexxus[I_c1].sock, (char *) &len, sizeof(len));
						if (rc <= 0)
							{
							if (rc < 0)
								perror("read 1");
							if (debug)
								{
								printf("[VACM_SYS_STATD] Closing Nexxus fd %d\n",
										 	 nexxus[I_c1].sock);
								}
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
#endif
						len = ntohl(len);
#ifdef OPENSSL
						rc = SSL_read(nexxus[I_c1].ssl, (char *) buffer, len);
						if ((!rc) || (rc <0))
							{
							SSL_shutdown(nexxus[I_c1].ssl);
							SSL_free(nexxus[I_c1].ssl);
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
#else
						rc = read(nexxus[I_c1].sock, buffer, len);
						if (rc <= 0)
							{
							if (rc < 0)
								perror("read 2");
							printf("[VACM_SYS_STATD] Closing Nexxus fd %d\n",
										 nexxus[I_c1].sock);
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
#endif
						process_cmd(&nexxus[I_c1], buffer);
						/* IPC */
						}
					} /* end if valid */
				} /* end for */
			} /* end else */
		} /* end while */
}

static void process_cmd(NEXXUS *nexxusp, char *buffer)
{
  char buff[255];
	char	*type;
	char	*cmd;
  char  *id;
	int		rc = 0;
	int		fd;
  __uint32_t job_id;

	fd = nexxusp->sock;
	if (debug)
		printf("[VACM_SYS_STATD] Processing <%s>\n",buffer);
	if (!(type = strtok(buffer,":")))
		return;
	if (!nexxusp->authenticated)
		{
		if (!strncmp(type,"AUTH",4))
			{
			char	*password;
			char	*crypted;
			if (!(password = strtok(NULL,":")))
				{
				send_response(fd,"AUTH_ERROR");
				return;
				}
			crypted = crypt(password, "HI");
			if (strcmp(crypted, daemon_password))
				{
				send_response(fd,"AUTH_ERROR");
				return;
				}
			nexxusp->authenticated = 1;
			send_response(fd, "AUTH_OK");
			return;
			}
		return;
		}
	if (!(cmd = strtok(NULL,":")))
		return;
	if (strncmp(type,"REQUEST", 7))
		return;

  if (!(id = strtok(NULL,":")))
    return;

  job_id = (__uint32_t) strtoul(id, (char **) NULL, 10);

	if (!strncmp(cmd,"MEM",3))
		rc = do_mem(job_id, fd);
	else if (!strncmp(cmd,"CPU",3))
		rc = do_cpu(job_id, fd);
	else if (!strncmp(cmd,"UPTIME",6))
		rc = do_uptime(job_id, fd);
	else if (!strncmp(cmd,"FS",2))
		rc = do_fs(job_id, fd);
	else if (!strncmp(cmd,"WHO",3))
		rc = do_who(job_id, fd);
	else if (!strncmp(cmd,"PS",2))
		rc = do_ps(job_id, fd);
	else if (!strncmp(cmd,"VERSION",7))
		rc = do_version(job_id, fd);
	else if (!strncmp(cmd,"APM",3))
		rc = do_apm(job_id, fd);
	else if (!strncmp(cmd,"INT",3))
		rc = do_int(job_id, fd);
	else if (!strncmp(cmd,"IO",2))
		rc = do_io(job_id, fd);
	else if (!strncmp(cmd,"DMA",3))
		rc = do_dma(job_id, fd);
	else if (!strncmp(cmd,"SWAP",4))
		rc = do_swap(job_id, fd);

	if (rc < 0)
    sprintf(buff,"%d:ERROR",job_id);
	else
    sprintf(buff,"%d:DONE",job_id);

	send_response(fd, buff);
	if (debug)
		printf("Request processing complete.\n");
}

static int do_mem(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	line[255];
	char	msg[512];
	char	*bword;
	int		total=0,free=0,shared=0,buffers=0,cached=0,swap_total=0,swap_free=0;

	if (!(file = fopen("/proc/meminfo","r")))
		return(-1);
	while(fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;
		if (!(bword = strtok(line,":")))
			continue;
		if (!strcmp(bword, "MemTotal"))
			total = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "MemFree"))
			free = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "MemShared"))
			shared = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "Buffers"))
			buffers = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "Cached"))
			cached = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "SwapTotal"))
			swap_total = atoi(strtok(NULL,"\t "));
		else if (!strcmp(bword, "SwapFree"))
			swap_free = atoi(strtok(NULL,"\t "));
		}
	fclose(file);
	snprintf(msg,
          sizeof(msg), 
					"%d:MEM:%d:%d:%d:%d:%d:%d:%d",
           job_id,
					 total,
					 free,
					 shared,
					 buffers,
					 cached,
					 swap_total,
					 swap_free);
	send_response(fd, msg);
	return(0);
}

static int do_cpu(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	l1[10], l2[10], l3[10];

	if (!(file = fopen("/proc/loadavg","r")))
		return(-1);
	fgets(line,sizeof(line),file);
	strncpy(l1,strtok(line," "),sizeof(l1));
	strncpy(l2,strtok(NULL," "),sizeof(l2));
	strncpy(l3,strtok(NULL," "),sizeof(l3));
	fclose(file);
	snprintf(msg,sizeof(msg),"%d:CPU_LOAD:%s:%s:%s",job_id,l1,l2,l3);
	send_response(fd, msg);
	return(0);
}

static int do_uptime(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	l1[32];

	if (!(file = fopen("/proc/uptime","r")))
		return(-1);
	fgets(line,sizeof(line),file);
	strncpy(l1,strtok(line," "),sizeof(l1));
	fclose(file);
	snprintf(msg,sizeof(msg),"%d:UPTIME:%s",job_id,l1);
	send_response(fd, msg);
	return(0);
}

static int do_fs(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	fsname[16],blk_total[32],blk_used[32],blk_free[32],perc[6],mnt[64];

	if (!(file = popen("/bin/df -l -t ext2 -t ext3","r")))
		return(-1);
	
	fgets(line,sizeof(line),file);
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;

		strncpy(fsname, strtok(line, "\t "),sizeof(fsname));
		strncpy(blk_total, strtok(NULL, "\t "),sizeof(blk_total));
		strncpy(blk_used, strtok(NULL, "\t "),sizeof(blk_used));
		strncpy(blk_free, strtok(NULL, "\t "),sizeof(blk_free));
		strncpy(perc, strtok(NULL, "\t "),sizeof(perc));
		strncpy(mnt, strtok(NULL, "\t "),sizeof(mnt));

		snprintf(msg,
            sizeof(msg),
						"%d:FS:%s:%s:%s:%s:%s:%s",
            job_id,
						fsname,
						blk_total,
						blk_used,
						blk_free,
						perc,
						mnt);
		send_response(fd, msg);
		}
	pclose(file);
	return(0);
}

static int do_who(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	name[16], tty[16];

	if (!(file = popen("/usr/bin/who","r")))
		return(-1);
	
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;

		strncpy(name, strtok(line, "\t "),sizeof(name));
		strncpy(tty, strtok(NULL, "\t "),sizeof(tty));

		snprintf(msg,sizeof(msg),"%d:WHO:%s:%s",job_id,name, tty);
		send_response(fd, msg);
		}
	pclose(file);
	return(0);
}

static int do_ps(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[2048], line[1024];
	char	pid[8],user[16],cpu[6],mem[6],rss[16],state[6],tty[8],cmd[512];

	if (!(file = popen("/bin/ps --no-headers -eo pid,user,%cpu,%mem,rss,state,tty,cmd","r")))
		return(-1);
	
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;

		strncpy(pid, strtok(line, "\t "),sizeof(pid));
		strncpy(user, strtok(NULL, "\t "),sizeof(user));
		strncpy(cpu, strtok(NULL, "\t "),sizeof(cpu));
		strncpy(mem, strtok(NULL, "\t "),sizeof(mem));
		strncpy(rss, strtok(NULL, "\t "),sizeof(rss));
		strncpy(state, strtok(NULL, "\t "),sizeof(state));
		strncpy(tty, strtok(NULL, "\t "),sizeof(tty));
		strncpy(cmd, strtok(NULL, "\t "),sizeof(cmd));

		snprintf(msg,
            sizeof(msg),
						"%d:PS:%s:%s:%s:%s:%s:%s:%s:%s",
            job_id,
						pid,
						user,
						cpu,
						mem,
						rss,
						state,
						tty,
						cmd);
		send_response(fd, msg);
		}
	pclose(file);
	return(0);
}

static int do_version(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	l1[32];
	char	*bword;

	if (!(file = fopen("/proc/version","r")))
		return(-1);
	fgets(line,sizeof(line),file);
	bword = strtok(line," ");
	bword = strtok(NULL," ");
	strncpy(l1,strtok(NULL,"("),sizeof(l1));
	fclose(file);
	snprintf(msg,sizeof(msg),"%d:VERSION:%s",job_id,l1);
	send_response(fd, msg);
	return(0);
}

typedef struct apm_info {
   const char driver_version[10];
   int        apm_version_major;
   int        apm_version_minor;
   int        apm_flags;
   int        ac_line_status;
   int        battery_status;
   int        battery_flags;
   int        battery_percentage;
   int        battery_time;
   int        using_minutes;
} apm_info;   

static int do_apm(__uint32_t job_id, int	fd)
{
	FILE			*file;
	apm_info	i;
	char			units[10];
	char			tmp[10];
	char			msg[512], line[255];

	if (!(file = fopen("/proc/apm","r")))
		{
		sprintf(msg,"%d:APM:NA:NA:0:0:NA",job_id);
		send_response(fd, msg);
		return(0);
		}
	fgets(line,sizeof(line),file);
	line[(strlen(line)-1)]=0;
   sscanf(line, "%s %d.%d %x %x %x %x %d%% %d %s\n",
     (char *)i.driver_version,
     &i.apm_version_major,
     &i.apm_version_minor,
     &i.apm_flags,
     &i.ac_line_status,
     &i.battery_status,
     &i.battery_flags,
     &i.battery_percentage,
     &i.battery_time,
     units );

	sprintf(msg,"%d:APM:",job_id);
	switch (i.ac_line_status)
		{
		case 0: strcat(msg, "OFFLINE:"); break;
		case 1: strcat(msg, "ONLINE:"); break;
		case 2: strcat(msg, "BACKUP:"); break;
		default:strcat(msg, "UNKNOWN:"); break;
		}
	switch(i.battery_status)
		{
		case 0: strcat(msg, "HIGH:"); break;
		case 1: strcat(msg, "LOW:"); break;
		case 2: strcat(msg, "CRITICAL:"); break;
		case 3: strcat(msg, "CHARGING:"); break;
		default:strcat(msg, "UNKNOWN:"); break;
		}
	sprintf(tmp, "%d%%:",i.battery_percentage);
	strcat(msg,tmp);
	sprintf(tmp, "%d:", i.battery_time);
	strcat(msg,tmp);
	strcat(msg,units);
	send_response(fd, msg);
	fclose(file);
	return(0);
}

static int do_int(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	num[32], cnt[4][32], source[32], name[32];
	char	*bword;
	int		num_cpus, I_c1;

	if (!(file = fopen("/proc/interrupts","r")))
		return(-1);
	
	/* First figure out how manu CPU's we have */
	num_cpus = 1;
	fgets(line,sizeof(line),file);
	line[(strlen(line)-1)]=0;
	bword = strtok(line, "\t ");
	while((bword = strtok(NULL,"\t ")))
		num_cpus++;

	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;

		/* Get the interrupt number */
		strncpy(num, strtok(line, "\t "),sizeof(num));
		num[(strlen(num)-1)]=0;
		
		if (!strcmp(num,"NMI"))
			{
			fclose(file);
			return(0);
			}
		/* Get the counts */
		for (I_c1 = 0; I_c1 < num_cpus; I_c1 ++)
			strcpy(cnt[I_c1], strtok(NULL," "));

		/* Get the interrupt source */
		strncpy(source, strtok(NULL, "\t "),sizeof(source));

		/* Get the interrupt name */
		strncpy(name, strtok(NULL, "\t "),sizeof(name));

		snprintf(msg,
            sizeof(msg),
						"%d:INT:%s:%s:%s:%s:%s:%s:%s",
            job_id,
						num,
						source,
						name,
						cnt[0],
						(num_cpus > 1 ? cnt[1] : "NONE"),
						(num_cpus > 2 ? cnt[2] : "NONE"),
						(num_cpus > 3 ? cnt[3] : "NONE"));
		send_response(fd, msg);
		}
	fclose(file);
	return(0);
}

static int do_io(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	addr[16], function[64];
	char	*bword;

	if (!(file = fopen("/proc/ioports","r")))
		return(-1);
	
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;
		strncpy(addr, strtok(line, " "),sizeof(addr));
		bword = strtok(NULL," ");
		strncpy(function,strtok(NULL," "),sizeof(function));
		snprintf(msg,sizeof(msg),"%d:IO:%s:%s",job_id,addr,function);
		send_response(fd, msg);
		}
	fclose(file);
	return(0);
}

static int do_dma(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	chan[16], function[64];

	if (!(file = fopen("/proc/dma","r")))
		return(-1);
	
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;
		strncpy(chan,strtok(line," "),sizeof(chan));
		chan[(strlen(chan)-1)]=0;
		strncpy(function,strtok(NULL," "),sizeof(function));
		snprintf(msg,sizeof(msg),"%d:DMA:%s:%s",job_id,chan,function);
		send_response(fd, msg);
		}
	fclose(file);
	return(0);
}

static int do_swap(__uint32_t job_id, int	fd)
{
	FILE	*file;
	char	msg[512], line[255];
	char	filename[32], type[32], size[32], used[32], priority[6];

	if (!(file = fopen("/proc/swaps","r")))
		return(-1);
	
	fgets(line,sizeof(line),file);
	while (fgets(line,sizeof(line),file))
		{
		line[(strlen(line)-1)]=0;

		strncpy(filename,strtok(line,"\t "),sizeof(filename));
		strncpy(type,strtok(NULL,"\t "),sizeof(type));
		strncpy(size,strtok(NULL,"\t "),sizeof(size));
		strncpy(used,strtok(NULL,"\t "),sizeof(used));
		strncpy(priority,strtok(NULL,"\t "),sizeof(priority));

		snprintf(msg,
            sizeof(msg),
						"%d:SWAP:%s:%s:%s:%s:%s",
            job_id,
						filename,
						type,
						size,
						used,
						priority);
		send_response(fd, msg);
		}
	fclose(file);
	return(0);
}

static void send_response(int fd, char *msg)
{
	__uint32_t	len;
	__uint32_t	netlen;
#ifdef OPENSSL
	int					rc;
	SSL					*ssl;
#endif

	if (debug)
		printf("[VACM_SYS_STATD] Sending response <%s>\n",msg);
	len = strlen(msg)+1;
	netlen = htonl(len);

#ifdef OPENSSL
	ssl = lookup_ssl_by_fd(fd);
  rc = SSL_write(ssl, (char *) &netlen, sizeof(netlen));
	if ((!rc) || (rc <0))
		{
		ERR_print_errors_fp(stderr);
		return;
		}
  rc = SSL_write(ssl, msg, (int) len);
	if ((!rc) || (rc <0))
		{
		ERR_print_errors_fp(stderr);
		return;
		}
#else
	if (write(fd, (char *) &netlen, sizeof(netlen))<0)
		{
		perror("write length");
		return;
		}
	if (write(fd, (char *) msg, len)<0)
		{
		perror("write data");
		return;
		}
#endif
}

#ifdef OPENSSL
static SSL *lookup_ssl_by_fd(int fd)
{
	int	I_c1;

	for (I_c1=0; I_c1<255;I_c1++)
		if (nexxus[I_c1].sock == fd)
			return(nexxus[I_c1].ssl);
  return(NULL);
}
#endif

static int set_password()
{
	FILE						*Tfilep_ptr;
	struct termios	term;
	char						passwd1[255];
	char						passwd2[255];
	char						*crypted;

	unlink("/etc/vacm_sys_statd.passwd");
	if (!(Tfilep_ptr = fopen("/etc/vacm_sys_statd.passwd","w")))
		{
		printf("Unable to create /etc/vacm_sys_statd.passwd (%m)\n");
		return(0);
		}

	tcgetattr(fileno(stdin), &term);
	term.c_lflag &= ~ECHO;
	tcsetattr(fileno(stdin), TCSANOW, &term);

	fprintf(stdout,"Enter new password -=>");
	fflush(stdout);
	fgets(passwd1,sizeof(passwd1),stdin);
	passwd1[(strlen(passwd1)-1)]=0;
	fprintf(stdout,"\nRe-enter new password -=>");
	fgets(passwd2,sizeof(passwd2),stdin);
	passwd2[(strlen(passwd2)-1)]=0;
	term.c_lflag |= ECHO;
	tcsetattr(fileno(stdin), TCSANOW, &term);
	printf("\n");
	if (strcmp(passwd1, passwd2))
		{
		printf("Passwords do not match\n");
		fclose(Tfilep_ptr);
		return(0);
		}
	crypted = crypt(passwd1, "HI");
	memset(passwd1,0,sizeof(passwd1));
	memset(passwd2,0,sizeof(passwd2));
	fprintf(Tfilep_ptr,"%s\n",crypted);
	strcpy(daemon_password, crypted);
	fclose(Tfilep_ptr);
	return(1);
}

static int get_password()
{
	FILE	*Tfilep_ptr;
	
	if (!(Tfilep_ptr = fopen("/etc/vacm_sys_statd.passwd","r")))
		{
		printf("Unable to open /etc/vacm_sys_statd.passwd (%m)\n");
		return(0);
		}
	if (!fgets(daemon_password, sizeof(daemon_password), Tfilep_ptr))
		{
		fclose(Tfilep_ptr);
		return(0);
		}
	daemon_password[(strlen(daemon_password)-1)]=0;
	fclose(Tfilep_ptr);
	return(1);
}

static void usage(char *name)
{
	printf("Usage: %s [-p] [-f]\n",name);
	printf("       [-p] flag changes the nexxus access password\n");
	printf("       [-f] flag sets daemon to run in foreground\n");
}

static void become_daemon()
{ 
  int i;
  int fd;
  
  chdir("/");
  if (fork())
    _exit(0);
  setsid(); 
  close(0); 
  close(1); 
  close(2); 
  for (i=0; i < 3; i++)
    {
    fd = open("/dev/null", O_RDWR, 0);
    if (fd < 0) 
      fd = open("/dev/null", O_WRONLY,0);
    if (fd < 0)
      {
      fprintf(stderr,"Unable to detach\n");
      _exit(-1);
      }
    if (fd != i)
      {
      fprintf(stderr,"Unable to detach\n");
      _exit(-1);
      }
    } 
}   

int	ok_count = 0;
static void signal_handler(int signal)
{
	struct timeval	mytime;
	int							diff;
	int							diff_usec;

	if (signal == SIGALRM)
		{
		gettimeofday(&mytime, NULL);
		diff = mytime.tv_sec - last_time.tv_sec;
		diff_usec = abs(mytime.tv_usec - last_time.tv_usec);
		
		if (diff > (TIMER_INTERVAL+GRACE_PERIOD))
			{
			ok_count = 0;
			priority--;
			if (priority < -10)
				priority = -10;
			printf("updating priority to %d (diff = %d)\n",priority,diff);
			if (setpriority(PRIO_PROCESS, 0, priority)<0)
				perror("setpriority");
			}
		else
			{
			ok_count++;
			if (priority != 0)
				{
				if (ok_count > (TIMER_INTERVAL * 10))
					{
					priority++;
					printf("updating priority to %d (diff = %d)\n",priority,diff);
					if (setpriority(PRIO_PROCESS, 0, priority)<0)
					perror("setpriority");
					}
				}
			}
		memcpy(&last_time, &mytime,sizeof(last_time));
		}
}

