/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * VACM Remote User Administration Module
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <libloose.h>
#include <user_adm.h>
#include <config.h>
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#ifndef OPENSSL
#error Due to the nature of this module, OpenSSL encryption is REQUIRED.
#endif

static void			config_cb(__uint32_t job_id, int client_fd, char *msg);
static void			deinit_cb(__uint32_t job_id, int client_fd, char *msg);
static void			discon_cb(__uint32_t job_id, int client_fd, char *msg);
static void			ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *ipcmsg);
static int			do_request(char				*reqmsg,
													 NODELIST		*node,
													 __uint32_t	job_id,
													 int				client_fd);
static int			do_disconnect(NODELIST *node);
static int			do_connect(NODELIST	*node, __uint32_t job_id, int client_fd);
static char			*return_req_str(int	request);
static NODELIST	*node_create(char *node_id);
static NODELIST	*lookup_node(char *node_id);
static void			node_destroy(NODELIST	*node);
static int 			do_auth(NODELIST	*node);

extern int errno;

static NODELIST	*nodelist_head = NULL;
static NODELIST	*nodelist_tail = NULL;

static	SSL_CTX			*ctx = NULL;
static	SSL_METHOD	*meth = NULL;

int main(argc, argv)
int		argc;
char	**argv;
{
	KNOWN_UNUSED_PARAM(argc);
	KNOWN_UNUSED_PARAM(argv);

	lm_init();
	lm_nexxus_connect();
	if (lm_register("Useradm",
									"User Administration Daemon Module",
									"Module to interface with a remote VACM User Administration Daemon",
									"San Mehat (nettwerk@valinux.com)",
									"USER_ADM",
									0,
									1) < 0)
		{
		printf("Unable to register module with Nexxus\n");
		exit(-1);
		}

	SSLeay_add_ssl_algorithms();
  meth = SSLv2_client_method();
  SSL_load_error_strings();
  ctx = SSL_CTX_new(meth);
	if (!ctx)
		{
		ERR_print_errors_fp(stderr);
		exit(0);
		}

	lm_main_loop(config_cb, deinit_cb, discon_cb, ipc_cb);
	exit(0);
}

static void config_cb(__uint32_t job_id, int client_fd, char *msg)
{
	NODELIST	*node;
	char			*cmd_module;
	char			*node_id;
	char			logmsg[255];

	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);

	cmd_module = strtok(msg, ":");
	node_id = strtok(NULL, ":");

	node = lookup_node(node_id);

	if (!strcasecmp(cmd_module, "NODE"))
		node = node_create(node_id);
	else if (!strcasecmp(cmd_module, "END_NODE"))
		{
		}
	else if (!strcasecmp(cmd_module, "ADDITION"))
		{
		node = node_create(node_id);
		return;
		}
	else if (!strcasecmp(cmd_module, "DELETION"))
		{
		node_destroy(node);
		return;
		}
	else if (!strcasecmp(cmd_module, "RENAME"))
		{
		char	*new_name;
		
		if (!(new_name = strtok(NULL,":")))
			{
			printf("[USER_ADM] Malformatted cfg msg\n");
			return;
			}
		strcpy(node->node_id, new_name);
		}
	else if (!strcasecmp(cmd_module, "GLOBAL"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			printf("[USER_ADM] Malformatted cfg msg\n");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			printf("[USER_ADM] Malformatted cfg msg\n");
			return;
			}
		/* Any global variabls we care about here... ie: */
		if (!strcasecmp(variable, "IP_ADDRESS"))
			{
			node->ip_address = inet_addr(value);
			}
		}
	else if (!strcasecmp(cmd_module, "USER_ADM"))
		{
		char	*variable;
		char	*value;
		
		if (!(variable = strtok(NULL,":")))
			{
			printf("[USER_ADM] Malformatted cfg msg\n");
			return;
			}
		if (!(value = strtok(NULL,":")))
			{
			printf("[USER_ADM] Malformatted cfg msg\n");
			return;
			}
		/* Process any local variables here */
		if (!strcasecmp(variable,"AUTH_PASSWORD"))
			{
			strncpy(node->password,value,40);
			return;
			}
		sprintf(logmsg,"Received unknown local module variable '%s'\n",
					variable);
		lm_log(logmsg);
		}
	else
		{
		sprintf(logmsg,"Received unknown config event <%s>",cmd_module);
		lm_log(logmsg);
		}
	if ((node->ip_address) && (node->password[0]))
		{
		struct in_addr	addr;
		addr.s_addr = node->ip_address;

		sprintf(logmsg,"Node '%s' (ip %s) ready",
					 node_id,inet_ntoa(addr));
		lm_log(logmsg);
		}
}

static void deinit_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
	KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);

	lm_nexxus_disconnect();
	exit(0);
}

static void discon_cb(__uint32_t job_id, int client_fd, char *msg)
{
	KNOWN_UNUSED_PARAM(job_id);
  KNOWN_UNUSED_PARAM(client_fd);
	KNOWN_UNUSED_PARAM(msg);
}

#define GET_NEXT_TOKEN(dst, job_id, client_fd) \
	{ \
	char	*tmp_token; \
	char	err_msg[255]; \
	if (!(tmp_token = strtok(NULL,":"))) \
		{ \
		sprintf(err_msg, \
					"%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE", \
					job_id, \
					client_fd); \
		lm_send_to_nexxus(err_msg); \
		return; \
		} \
	strncpy(dst, tmp_token, sizeof(dst)); \
	}
static void ipc_cb(char *node_id, __uint32_t job_id, int client_fd, char *ipcmsg)
{
	char			*cmd;
	char			toss[255];
	char			msg[255];
	NODELIST	*node;
	int				rc = 0;

	if (!(cmd = strtok(ipcmsg, ":")))
		goto out_malformatted;

	GET_NEXT_TOKEN(toss, job_id, client_fd);
	if (!(node = lookup_node(node_id)))
		{
		sprintf(msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_FOUND",
						job_id,
						client_fd);
		lm_send_to_nexxus(msg);
		return;
		}
	/* Process IPC here */
	memset(&msg[0], 0, sizeof(msg));
	if (!strcasecmp(cmd, "CONFIGURATION"))
		{
		char	daemon_password[80];
		int		i;

		GET_NEXT_TOKEN(daemon_password, job_id, client_fd);

		if ((int)strlen(daemon_password) > 40)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:PASSWORD_TOO_LONG",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return;
			}
		for (i=0; i < (int)strlen(daemon_password); i++)
			if (!isprint((int)daemon_password[i]))
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:BAD_PASSWORD",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				return;
				}
		sprintf(msg,
						"0:0:SET:%s:USER_ADM:AUTH_PASSWORD:%s",
						node_id,
						daemon_password);
		lm_send_to_nexxus(msg);
		sprintf(msg,
						"%d:%d:FOR:JOB_COMPLETED",
						job_id,
						client_fd);
		lm_send_to_nexxus(msg);
    strcpy(node->password,daemon_password);
    if ((node->ip_address) && (node->password[0]))
      {
      struct in_addr  addr;
      addr.s_addr = node->ip_address;
      
      sprintf(msg,"Node '%s' (ip %s) ready",
              node_id,inet_ntoa(addr));
      lm_log(msg);
      }
		return;
		}

	if ((!node->ip_address) || (!node->password[0]))
		{
		sprintf(msg,
						"%d:%d:FOR:JOB_ERROR:NODE_NOT_CONFIGURED",
						job_id,
						client_fd);
		lm_send_to_nexxus(msg);
		return;
		}
	if (!strcasecmp(cmd, "USER_ADD"))
		{
		char	user_name[16];
		char	user_id[16];
		char	primary_group[32];
		char	home_dir[255];
		char	shell_path[255];
		char	password[255];
		char	expire_date[16];
		char	inactive_days[8];
		char	comment[42];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(user_id, job_id, client_fd);
		GET_NEXT_TOKEN(primary_group, job_id, client_fd);
		GET_NEXT_TOKEN(home_dir, job_id, client_fd);
		GET_NEXT_TOKEN(shell_path, job_id, client_fd);
		GET_NEXT_TOKEN(password, job_id, client_fd);
		GET_NEXT_TOKEN(expire_date, job_id, client_fd);
		GET_NEXT_TOKEN(inactive_days, job_id, client_fd);
		GET_NEXT_TOKEN(comment, job_id, client_fd);
		sprintf(msg,
						"REQUEST:USER_ADD:%s:%s:%s:%s:%s:%s:%s:%s:%s",
						user_name,
						user_id,
						primary_group,
						home_dir,
						shell_path,
						password,
						expire_date,
						inactive_days,
						comment);
		}
	else if (!strcasecmp(cmd, "USER_REMOVE"))
		{
		char	user_name[16];
		char	delete_flag[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(delete_flag, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_REMOVE:%s:%s",user_name,delete_flag);
		}
	else if (!strcasecmp(cmd, "USER_GRP_ADD"))
		{
		char	user_name[16];
		char	group_name[32];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(group_name, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_GRP_ADD:%s:%s",user_name,group_name);
		}
	else if (!strcasecmp(cmd, "USER_GRP_REM"))
		{
		char	user_name[16];
		char	group_name[32];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(group_name, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_GRP_REM:%s:%s",user_name,group_name);
		}
	else if (!strcasecmp(cmd, "USER_CHG_GRP"))
		{
		char	user_name[16];
		char	group_name[32];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(group_name, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_GRP:%s:%s",user_name,group_name);
		}
	else if (!strcasecmp(cmd, "USER_CHG_DIR"))
		{
		char	user_name[16];
		char	home_dir[255];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(home_dir, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_DIR:%s:%s",user_name,home_dir);
		}
	else if (!strcasecmp(cmd, "USER_CHG_SHELL"))
		{
		char	user_name[16];
		char	shell_path[255];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(shell_path, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_SHELL:%s:%s",user_name,shell_path);
		}
	else if (!strcasecmp(cmd, "USER_CHG_UID"))
		{
		char	user_name[16];
		char	user_id[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(user_id, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_UID:%s:%s",user_name,user_id);
		}
	else if (!strcasecmp(cmd, "USER_CHG_EXP"))
		{
		char	user_name[16];
		char	expire_date[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(expire_date, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_EXP:%s:%s",user_name,expire_date);
		} 
	else if (!strcasecmp(cmd, "USER_CHG_INA"))
		{
		char	user_name[16];
		char	inactive_days[8];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(inactive_days, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_INA:%s:%s",user_name,inactive_days);
		}
	else if (!strcasecmp(cmd, "USER_CHG_COMMENT"))
		{
		char	user_name[16];
		char	comment[42];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(comment, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_COMMENT:%s:%s",user_name,comment);
		}
	else if (!strcasecmp(cmd, "USER_CHG_PASSWD"))
		{
		char	user_name[16];
		char	old_pass[255];
		char	new_pass[255];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		GET_NEXT_TOKEN(old_pass, job_id, client_fd);
		GET_NEXT_TOKEN(new_pass, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_CHG_PASSWD:%s:%s:%s",user_name,old_pass,new_pass);
		}
	else if (!strcasecmp(cmd, "USER_LIST"))
		strcpy(msg,"REQUEST:USER_LIST");
	else if (!strcasecmp(cmd, "GROUP_LIST"))
		strcpy(msg,"REQUEST:GROUP_LIST");
	else if (!strcasecmp(cmd, "LOCK_USER"))
		{
		char	user_name[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		sprintf(msg,"REQUEST:LOCK_USER:%s",user_name);
		}
	else if (!strcasecmp(cmd, "UNLOCK_USER"))
		{
		char	user_name[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		sprintf(msg,"REQUEST:UNLOCK_USER:%s",user_name);
		}
	else if (!strcasecmp(cmd, "USER_GRP_LIST"))
		{
		char	user_name[16];

		GET_NEXT_TOKEN(user_name, job_id, client_fd);
		sprintf(msg,"REQUEST:USER_GRP_LIST:%s",user_name);
		}
	else
		{
		sprintf(msg,
						"%d:%d:FOR:JOB_ERROR:UNSUPPORTED_MESSAGE",
						job_id,
						client_fd);
		lm_send_to_nexxus(msg);
		return;
		}
	rc = do_request(msg, node, job_id, client_fd);
	if (!rc)
		{
		/* Send job completed message */
		sprintf(msg,
						"%d:%d:FOR:JOB_COMPLETED",
						job_id,
						client_fd);
		lm_send_to_nexxus(msg);
		}
	return;
out_malformatted:
	sprintf(msg,
					"%d:%d:FOR:JOB_ERROR:MALFORMATTED_MESSAGE",
					job_id,
					client_fd);
	lm_send_to_nexxus(msg);
	return;
}

static int do_auth(NODELIST	*node)
{
	__uint32_t			netlen;
	__uint32_t			reqlen;
	char						msg[255];
	int							rc;
	char						rspstr[512];
	fd_set					read_fds;
	struct timeval	to;

	snprintf(msg, sizeof(msg), "AUTH:%s",node->password);
	reqlen = strlen(msg)+1;
	netlen = htonl(reqlen);
	rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
	if ((!rc) || (rc <0))
		return(-1);
	rc = SSL_write(node->ssl, msg, reqlen);
	if ((!rc) || (rc <0))
		return(-1);
	while(1)
		{
		FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
		to.tv_sec = 5;
		to.tv_usec= 0;
		if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
			return(-1);
		else if (!rc)
			return(-1);
		if (FD_ISSET(node->comm_sock, &read_fds))
			{
			rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
			if ((!rc) || (rc <0))
				{
				return(-1);
				}
			reqlen = ntohl(netlen);
			rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
			if ((!rc) || (rc <0))
				return(-1);
			if (!strncmp(rspstr,"AUTH_ERROR",10))
				{
        char  nexxus_msg[255];

        sprintf(nexxus_msg,
                "Authentication to node '%s' daemon failed",
                node->node_id);
        lm_log(nexxus_msg);

				return(-2);
				}
			else if (!strncmp(rspstr,"AUTH_OK",7))
				return(0);
			return(-1);
			}
		else
			return(-1);
		} /* end while */
}

static int do_request(char				*req,
											NODELIST		*node,
											__uint32_t	job_id,
											int					client_fd)
{
	char						reqmsg[255];
	__uint32_t			netlen;
	__uint32_t			reqlen;
	char						msg[255];
	int							rc;
	char						rspstr[512];
	fd_set					read_fds;
	struct timeval	to;
	__uint32_t			remote_id;
	char						*bword;

	sprintf(reqmsg,"%s:%d",req,job_id);
	reqlen = strlen(reqmsg)+1;
	netlen = htonl(reqlen);
retry:
	if (node->comm_sock == -1)
		{
		if ((rc = do_connect(node, job_id, client_fd)) != 0)
			{
			if (rc == -2)
				return(-1);
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:CONNECTION_FAILURE:%s",
							job_id,
							client_fd,
							strerror(rc));
			lm_send_to_nexxus(msg);
			return(-1);
			}
		}
	rc = SSL_write(node->ssl, (char *) &netlen, sizeof(netlen));
	if ((!rc) || (rc <0))
		{
		do_disconnect(node);
		goto retry;
		}
	rc = SSL_write(node->ssl, reqmsg, reqlen);
	if ((!rc) || (rc <0))
		{
		do_disconnect(node);
		goto retry;
		}

	while(1)
		{
		FD_ZERO(&read_fds);
		FD_SET(node->comm_sock, &read_fds);
		to.tv_sec = 5;
		to.tv_usec= 0;
		if ((rc = select((node->comm_sock+1), &read_fds, NULL, NULL, &to) )<0)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:%m",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return(-1);
			}
		else if (!rc)
			{
			sprintf(msg,
							"%d:%d:FOR:JOB_ERROR:TIMED OUT",
							job_id,
							client_fd);
			lm_send_to_nexxus(msg);
			return(-1);
			}
		if (FD_ISSET(node->comm_sock, &read_fds))
			{
			rc = SSL_read(node->ssl, (char *) &netlen, sizeof(netlen));
			if ((!rc) || (rc <0))
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
			reqlen = ntohl(netlen);
			rc = SSL_read(node->ssl, (char *) rspstr, reqlen);
			if ((!rc) || (rc <0))
				{
				printf("rc = %d\n",rc);
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_LOST",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				do_disconnect(node);
				return(-1);
				}
			if (!(bword = strtok(rspstr,":")))
				return(-1);
			remote_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
			bword += strlen(bword) +1;
			if (remote_id == job_id)
				{
				if (!strncmp(bword,"ERROR",5))
					{
					lm_log("Daemon reported internal error");
					sprintf(msg,
									"%d:%d:FOR:JOB_ERROR:INTERNAL_ERROR",
									job_id,
									client_fd);
					lm_send_to_nexxus(msg);
					return(-1);
					}
				else if (!strncmp(bword,"DONE_ERR",8))
					return(-1);
				else if (!strncmp(bword,"DONE",4))
					return(0);

				/* Forward the response off */
				sprintf(msg,
								"%d:%d:FOR:%s",
								job_id,
								client_fd,
								bword);
				lm_send_to_nexxus(msg);
				}
			else
				{
				sprintf(msg,
								"%d:%d:FOR:JOB_ERROR:CONNECTION_FAILURE:MISC ERROR",
								job_id,
								client_fd);
				lm_send_to_nexxus(msg);
				return(-1);
				}
			}
		} /* end while */
}

/*
 * disconnect from the daemon 
 */
static int do_disconnect(NODELIST *node)
{
	if (node->comm_sock == -1)
		return(-1);
	close(node->comm_sock);
	SSL_shutdown(node->ssl);
	node->comm_sock = -1;
	return(0);
}

/*
 * connect() to the daemon on the socket
 */
static int do_connect(NODELIST	*node, __uint32_t job_id, int client_fd)
{
	struct sockaddr_in	sock;
	int									optval = 1;
	int									optlen;
	fd_set							write_fds;
	struct timeval			to;
	int									rc;
	char								auth_errmsg[255];

	if ((rc = socket(AF_INET, SOCK_STREAM, 0)) <0)
		return(errno);
  node->comm_sock = rc;

  if (setsockopt(node->comm_sock,
                 SOL_SOCKET,
                 SO_REUSEADDR,
                 &optval,
                 sizeof(optval)) <0)
		return(errno);

	if ((rc = fcntl(node->comm_sock, F_GETFL))<0)
		return(errno);
  node->sock_opts = rc;

	if (fcntl(node->comm_sock, F_SETFL, O_NONBLOCK)<0)
		return(errno);
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = node->ip_address;
	sock.sin_port = htons(NETWORK_PORT);

	if (connect(node->comm_sock,
							(struct sockaddr *) &sock,
							sizeof(struct sockaddr)) <0)
		{
		if (errno != EINPROGRESS)
			return(errno);
		}

	FD_ZERO(&write_fds);
	FD_SET(node->comm_sock, &write_fds);
	to.tv_usec=0;
	to.tv_sec=5;
	if ((rc = select((node->comm_sock+1), NULL, &write_fds, NULL, &to)) < 0)
		return(errno);
	if (rc == 0)
		return(ETIMEDOUT);
	if (!FD_ISSET(node->comm_sock, &write_fds))
		return(errno);
	optlen = sizeof(optval);
	if (getsockopt(node->comm_sock,
								 SOL_SOCKET,
								 SO_ERROR,
								 &optval,
								 &optlen) <0)
		{
		perror("getsockopt");
		return(errno);
		}

	/* Set the socket to be blocking again */
	if (fcntl(node->comm_sock, F_SETFL, node->sock_opts)<0)
		return(errno);

	if (!(node->ssl = SSL_new(ctx)))
		return(EFAULT);
	SSL_set_fd(node->ssl, node->comm_sock);
	if ((rc = SSL_connect(node->ssl))<0)
		{
		lm_log("Unable to establish encrypted connection to daemon");
		ERR_print_errors_fp(stderr);
		do_disconnect(node);
		return(ENETRESET);
		}
	
  if (!(node->cert=SSL_get_peer_certificate(node->ssl)))
		{
		do_disconnect(node);
		return(ENETRESET);
    }
  X509_free(node->cert);
	/* Authenticate to daemon */
	rc = do_auth(node);
	if (rc == -1)
		{
		snprintf(auth_errmsg,
						 sizeof(auth_errmsg),
						 "%d:%d:FOR:JOB_ERROR:CONNECTION_FAILURE:Unable to authenticate",
						 job_id,
						 client_fd);
		lm_send_to_nexxus(auth_errmsg);
		do_disconnect(node);
		return(-2);
		}
	else if (rc == -2)
		{
		snprintf(auth_errmsg,
						 sizeof(auth_errmsg),
						 "%d:%d:FOR:JOB_ERROR:BAD_PASSWORD",
						 job_id,
						 client_fd);
		lm_send_to_nexxus(auth_errmsg);
		do_disconnect(node);
		return(-2);
		}
	return(optval);
}

static NODELIST	*node_create(char *node_id)
{
	NODELIST	*new;

	if (!(new = (NODELIST	*) malloc(sizeof(NODELIST))))
		return(0);
	memset(new, 0, sizeof(NODELIST));
	if (!nodelist_head)
		nodelist_head = nodelist_tail = new;
	else
		{
		nodelist_tail->next = new;
		new->prev = nodelist_tail;
		nodelist_tail = new;
		}
	new->comm_sock = -1;
	strcpy(new->node_id, node_id);
	return(new);
}

static NODELIST	*lookup_node(char *node_id)
{
	NODELIST	*scan;

	if (!node_id)
		return(NULL);
	scan = nodelist_head;
	while(scan)
		{
		if (!strcmp(node_id, scan->node_id))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

static void	node_destroy(NODELIST	*node)
{
  if (node == nodelist_head)
    {
    nodelist_head = nodelist_head->next;
    if (nodelist_head)
      nodelist_head->prev = NULL;
    else
      nodelist_tail = NULL;
    } 
  else if (node == nodelist_tail)
    {
    nodelist_tail = nodelist_tail->prev;
    if (nodelist_tail)
      nodelist_tail->next = NULL;
    else
      nodelist_head = NULL;
    } 
  else
    {
    node->next->prev = node->prev;
    node->prev->next = node->next;
    }
  free(node);
}

